(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory CPF_Parser 
imports
  Proof_Checker
  Xml
begin

fun
  var :: "('a, string) term xmlt"
where
  "var x = (Xml.text ''var'' x \<guillemotright>= (return \<circ> Var))"
hide_const (open) var

function
  "term" :: "'a xmlt \<Rightarrow> ('a, string) term xmlt"
where
  "term xml2name x = 
     xml_options [(''var'', CPF_Parser.var), (''funapp'', xml1many_elements ''funapp'' xml2name (xml1element ''arg'' (term xml2name) id) Fun)] x"  
by (pat_completeness) auto
hide_const (open) "term"

function
  ctxt :: "'a xmlt \<Rightarrow> ('a, string) ctxt xmlt"
where
  "ctxt xml2name x = xml_options [
     (''box'', xml0elements ''box'' Hole),  
     (''funContext'', xml4elements ''funContext'' 
       xml2name 
       (xml_many_elements ''before'' (CPF_Parser.term xml2name) id)
       (ctxt xml2name)
       (xml_many_elements ''after'' (CPF_Parser.term xml2name) id)
       More)
   ] x"  
by (pat_completeness) auto
hide_const (open) ctxt

definition
  flat_contexts :: "'a xmlt \<Rightarrow> ('a, string) ctxt list xmlt"
where
  "flat_contexts xml2name = xml_many_elements ''flatContexts'' (CPF_Parser.ctxt xml2name) id"
hide_const (open) flat_contexts

definition
  subst :: "'a xmlt \<Rightarrow> ('a, string) substL xmlt"
where
  "subst xml2name = 
     xml_many_elements ''substitution''
       (xml2elements ''substEntry'' 
         (Xml.text ''var'')
         (CPF_Parser.term xml2name)
         Pair)
       id"
hide_const (open) subst

primrec
  list2position :: "nat list \<Rightarrow> pos"
where
  "list2position [] = \<epsilon>" |
  "list2position (n # ns) = n <# list2position ns"

definition
  position :: "nat xmlt"
where
  "position \<equiv> xml_change (Xml.nat ''position'') (\<lambda>n. n - 1)"
hide_const (open) position

definition
  pos :: "pos xmlt"
where
  "pos \<equiv> xml_many_elements ''positionInTerm'' CPF_Parser.position list2position"
hide_const (open) pos


definition
  afs :: "'a xmlt \<Rightarrow> 'a afs_list xmlt" 
where
  "afs xml2name \<equiv> xml_many_elements ''argumentFilter'' 
  (xml3elements ''argumentFilterEntry''
      xml2name
      (Xml.nat ''arity'')
      (xml_options [(''collapsing'',xml_change (Xml.nat ''collapsing'') (\<lambda> i. Collapse (i - 1))),
                   (''nonCollapsing'', xml_many_elements ''nonCollapsing'' CPF_Parser.position AFList)])
      (\<lambda> f a e. ((f,a), e))
  ) id"
hide_const (open) afs

definition
  proj :: "'a xmlt \<Rightarrow> 'a projL xmlt"
where
  "proj xml2name \<equiv> xml_change (CPF_Parser.afs xml2name) (\<lambda> afl. Projection 
         (map (\<lambda> ((f,a),e). case e of 
               Collapse i \<Rightarrow> Pair f i
             | _ \<Rightarrow> Pair f a)
          afl))" 

definition
  rule :: "'a xmlt \<Rightarrow> ('a, string) rule xmlt"
where
  "rule xml2name = xml2elements ''rule'' (xml1element ''lhs'' (CPF_Parser.term xml2name) id) (xml1element ''rhs'' (CPF_Parser.term xml2name) id) Pair"
hide_const (open) rule 

definition
  rstep :: "'a xmlt \<Rightarrow> (pos \<times> ('a, string) rule \<times> bool \<times> ('a, string) term) xmlt"
where
  "rstep xml2name = xml3to4elements ''rewriteStep'' CPF_Parser.pos (CPF_Parser.rule xml2name) (xml0elements ''relative'' ()) (CPF_Parser.term xml2name) (\<lambda>p r rel t. (p, r, rel = None, t))"
hide_const (open) rstep

definition
  rsteps :: "'a xmlt \<Rightarrow> (('a,string)term \<times> ('a,string)prseq) xmlt"
where "rsteps xml2name \<equiv> xml1many_elements ''rewriteSequence'' 
             (xml1element ''startTerm'' (CPF_Parser.term xml2name) id)
             (CPF_Parser.rstep xml2name)
             Pair"
hide_const (open) rsteps

definition 
  loop :: "'a xmlt \<Rightarrow> (('a,string)term \<times> ('a,string)prseq \<times> ('a,string)substL \<times> ('a,string)ctxt) xmlt"
where "loop xml2name \<equiv> xml3elements ''loop''
             (CPF_Parser.rsteps xml2name)
             (CPF_Parser.subst xml2name)
             (CPF_Parser.ctxt xml2name)
             (\<lambda> (s,rseq) \<sigma> c. (s,rseq, \<sigma>, c))"
hide_const (open) loop

definition
  start_term :: "'a xmlt \<Rightarrow> ('a, string) term \<Rightarrow> ('a, string) term xmlt"
where
  "start_term xml2name t = xml1element ''startTerm'' (\<lambda>xml. (
     CPF_Parser.term xml2name xml \<guillemotright>= (\<lambda>s. if s = t
       then return t
       else error (''<startTerm> does not match lhs'')
     )
   )) id"
hide_const (open) start_term

definition
  rseq :: "'a xmlt \<Rightarrow> 'a proj \<Rightarrow> ('a, string) rule \<Rightarrow> (('a, string) rule \<times> (pos \<times> ('a, string) rule \<times> ('a, string) term) list) xmlt"
where
  "rseq xml2name \<pi> r = xml1many_elements ''rewriteSequence''
     (CPF_Parser.start_term xml2name (proj_term \<pi> (fst r))) (CPF_Parser.rstep xml2name) (\<lambda>_ rseq. (r, map (\<lambda> (x,y,_,z). (x,y,z)) rseq))"
hide_const (open) rseq

fun
  projected_rseq :: "'a xmlt \<Rightarrow> 'a proj \<Rightarrow> (('a, string) rule \<times> (pos \<times> ('a, string) rule \<times> ('a, string) term) list) xmlt"
where
  "projected_rseq xml2name \<pi> (XML name atts cs text) = (
     let tag = ''projectedRewriteSequence'' in (if name = tag \<and> atts = [] \<and> text = None \<and> length cs = 2
       then (CPF_Parser.rule xml2name (cs ! 0) \<guillemotright>= (\<lambda>r. CPF_Parser.rseq xml2name \<pi> r (cs ! 1)))
       else no_transform tag (XML name atts cs text)
     )
   )"
hide_const (open) projected_rseq

definition
  rules :: "'a xmlt \<Rightarrow> ('a,string) rules xmlt"
where
  "rules xml2name = xml_many_elements ''rules'' (CPF_Parser.rule xml2name) id"
hide_const (open) rules

definition
  funcsym :: "(string \<times> nat) xmlt"
where
  "funcsym = xml2elements ''funcsym'' (Xml.text ''name'') (Xml.nat ''arity'') Pair"
hide_const (open) funcsym 

definition
  "signature" :: "(string \<times> nat) list xmlt"
where
  "signature = xml_many_elements ''signature'' CPF_Parser.funcsym id"
hide_const (open) "signature"


function
  arith_fun :: "arithFun xmlt"
where
  "arith_fun xml = xml1element ''arithFunction'' (xml_options [ 
     (''natural'', xml_change (Xml.nat ''natural'') Const),
     (''variable'', xml_change (Xml.nat ''variable'') (\<lambda>n::nat. Arg (n - 1))),
     (''sum'', xml_many_elements ''sum'' arith_fun Sum),
     (''product'', xml_many_elements ''product'' arith_fun Prod),
     (''min'', xml_many_elements ''min'' arith_fun Min),
     (''max'', xml_many_elements ''max'' arith_fun Max)
   ]) id  xml"
by (pat_completeness, auto)
hide_const (open) arith_fun

definition
  sl_inter :: "('f, 'l) lab xmlt \<Rightarrow> ((('f, 'l) sl_inter \<times> bool) + (('f,'l)lab option)) xmlt"
where
  "sl_inter xml2name \<equiv> xml1element ''model'' (xml_options [
     (''finiteModel'', xml1or2many_elements ''finiteModel''
         (xml_change (Xml.nat ''carrierSize'') (\<lambda> n :: nat. n - 1))
         (xml1element ''tupleOrder'' (xml0elements ''pointWise'' ()) id)
         (xml3elements ''interpret''
           xml2name
           (Xml.nat ''arity'')
           CPF_Parser.arith_fun
           (\<lambda>f a inter. ((f, a), inter)))
         (\<lambda>n model xs. Inl (SL_Inter n xs, model = None))),
     (''rootLabeling'',xml0to1elements ''rootLabeling'' xml2name Inr)
   ]) id"
hide_const (open) sl_inter


fun
  sum_lpoly :: "('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> 'a list"
where
  "sum_lpoly _ [] ys = ys" |
  "sum_lpoly _ xs [] = xs" |
  "sum_lpoly pl (x # xs) (y # ys) = pl x y # sum_lpoly pl xs ys"

fun
  lpoly_of :: "'a \<Rightarrow> 'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> (nat,'a)tpoly \<Rightarrow> (string + 'a \<times> 'a list)"
where "lpoly_of _ _ _ _ (PNum i) = return (i,[])"
 | "lpoly_of ze on pl ti (PVar x) = return (ze, replicate x ze @ [on])"
 | "lpoly_of ze on pl ti (PSum []) = return (ze,[])"
 | "lpoly_of ze on pl ti (PSum (p # ps)) = (
      lpoly_of ze on pl ti p 
  \<guillemotright>= (\<lambda> (cp,ncp). lpoly_of ze on pl ti (PSum ps)
  \<guillemotright>= (\<lambda> (cq,ncq). return (pl cp cq,sum_lpoly pl ncp ncq))))"
 | "lpoly_of ze on pl ti (PMult []) = return (on,[])"
 | "lpoly_of ze on pl ti (PMult (p # ps)) = (
      lpoly_of ze on pl ti p
  \<guillemotright>= (\<lambda> (cp,ncp). lpoly_of ze on pl ti (PMult ps)
  \<guillemotright>= (\<lambda> (cq,ncq). if (list_all (op = ze) ncp) then return (ti cp cq, map (ti cp) ncq)
             else (if (list_all (op = ze) ncq) then return (ti cp cq, map (\<lambda> x. ti x cq) ncp)
             else error ''cannot transform non-linear polynomial to linear polynomial''))))"


function
  polynomial :: "'a xmlt \<Rightarrow> (nat,'a)tpoly xmlt"
where
  "polynomial xml2coeff x = xml1element ''polynomial'' 
     (xml_options [
       (''coefficient'',xml1element ''coefficient'' xml2coeff PNum), 
       (''variable'', xml_change (Xml.nat ''variable'') (\<lambda>n. PVar (n - 1))),
       (''sum'', xml_many_elements ''sum'' (polynomial xml2coeff) PSum),
       (''product'', xml_many_elements ''product'' (polynomial xml2coeff) PMult)
     ]) id x"
by pat_completeness auto
hide_const (open) polynomial

definition
  int_coeff :: "int xmlt"
where
  "int_coeff \<equiv> Xml.int ''integer''"
hide_const (open) int_coeff

definition
  vec_coeff :: "'a xmlt \<Rightarrow> 'a vec xmlt"
where
  "vec_coeff xml2coeff \<equiv> xml_many_elements ''vector''
     (xml1element ''coefficient'' xml2coeff id) id"
hide_const (open) vec_coeff

(* allow vectors and convert them to matrices by filling in zeros *)
definition
  mat_coeff :: "'a \<Rightarrow> 'a xmlt \<Rightarrow> 'a mat xmlt"
where
  "mat_coeff ze xml2coeff \<equiv> xml_options [
     (''matrix'', xml_many_elements ''matrix'' (CPF_Parser.vec_coeff xml2coeff) id),
     (''vector'', xml_change (CPF_Parser.vec_coeff xml2coeff) (\<lambda>v. v # replicate (length v - 1) (replicate (length v) ze)))
   ]"
hide_const (open) mat_coeff

definition
  arctic_coeff :: "arctic xmlt"
where
  "arctic_coeff \<equiv> xml_options [
     (''integer'', xml_change (Xml.int ''integer'') Num_arc),
     (''minusInfinity'', xml0elements ''minusInfinity'' MinInfty)
   ]"
hide_const (open) arctic_coeff

definition
  arctic_rat_coeff :: "arctic_rat xmlt"
where
  "arctic_rat_coeff \<equiv> xml_options [
     (''integer'', xml_change Xml.rat Num_arc_rat),
     (''rational'', xml_change Xml.rat Num_arc_rat),
     (''minusInfinity'', xml0elements ''minusInfinity'' MinInfty_rat)
   ]"
hide_const (open) arctic_rat_coeff


datatype Domain = Natural nat | Arctic | Arctic_rat rat | Int_mat nat nat | Arc_mat nat | Arc_rat_mat rat nat | Rational rat nat | Rat_mat rat nat nat

definition
  rat_domain :: "rat xmlt"
where
  "rat_domain \<equiv> xml1element ''rationals'' (xml1element ''delta'' Xml.rat id) id"
hide_const (open) rat_domain

definition
  basic_domain :: "(nat \<Rightarrow> Domain) xmlt"
where
  "basic_domain \<equiv> xml_options [
     (''naturals'', xml0elements ''naturals'' Natural),
     (''arctic'', xml1element ''arctic'' 
       (xml1element ''domain'' (xml_options [
         (''naturals'', xml0elements ''naturals'' (\<lambda> d. Arctic)),
         (''integers'', xml0elements ''integers'' (\<lambda> d. Arctic)),
         (''rationals'', xml_change CPF_Parser.rat_domain (\<lambda> r d. Arctic_rat r))
       ]) id) id),
     (''rationals'', xml_change CPF_Parser.rat_domain Rational)]"
hide_const (open) basic_domain

definition
  interpretation_type :: "Domain xmlt"
where
  "interpretation_type \<equiv> (xml1element ''type'' 
     (xml_options [
       (''polynomial'', (xml2elements ''polynomial'' 
       (xml1element ''domain'' 
       (xml_options [
         (''naturals'', CPF_Parser.basic_domain),
         (''arctic'', CPF_Parser.basic_domain),
         (''rationals'', CPF_Parser.basic_domain),
         (''matrices'',xml3elements ''matrices''
           (Xml.nat ''dimension'') 
           (Xml.nat ''strictDimension'') 
           (xml1element ''domain'' CPF_Parser.basic_domain id)
           (\<lambda> di sd domain d. case domain d of
             Natural _ \<Rightarrow> Int_mat di sd
           | Rational delta _ \<Rightarrow> Rat_mat delta di sd
           | Arctic \<Rightarrow> Arc_mat di
           | Arctic_rat delta \<Rightarrow> Arc_rat_mat delta di
         ))
       ]) 
     id)
     (Xml.nat ''degree'') 
     (\<lambda>type d . type d))),
     (''matrixInterpretation'',xml3elements ''matrixInterpretation''
       (xml1element ''domain'' CPF_Parser.basic_domain id)
       (Xml.nat ''dimension'') 
       (Xml.nat ''strictDimension'') 
       (\<lambda>domain di sd. case domain 0 of
         Natural _ \<Rightarrow> Int_mat di sd
       | Rational delta _  \<Rightarrow> Rat_mat delta di sd
       | Arctic \<Rightarrow> Arc_mat di
       | Arctic_rat delta \<Rightarrow> Arc_rat_mat delta di
     ))
   ])
 id)"
hide_const (open) interpretation_type

datatype ('a,'b,'c,'d,'e,'f,'g,'h,'i,'j)alter = Alter1 'a | Alter2 'b | Alter3 'c | Alter4 'd | Alter5 'e | Alter6 'f | Alter7 'g | Alter8 'h | Alter9 'i | Alter10 'j
fun alter1 where "alter1 (Alter1 x) = x"
fun alter2 where "alter2 (Alter2 x) = x"
fun alter3 where "alter3 (Alter3 x) = x"
fun alter4 where "alter4 (Alter4 x) = x"
fun alter5 where "alter5 (Alter5 x) = x"
fun alter6 where "alter6 (Alter6 x) = x"
fun alter7 where "alter7 (Alter7 x) = x"
fun alter8 where "alter8 (Alter8 x) = x"
fun alter9 where "alter9 (Alter9 x) = x"
fun alter10 where "alter10 (Alter10 x) = x"

fun
  fit_length :: "'a \<Rightarrow> nat \<Rightarrow> 'a list \<Rightarrow> 'a list"
where
  "fit_length _ 0 _ = []" |
  "fit_length a n [] = replicate n a" |
  "fit_length a (Suc n) (b # as) = b # (fit_length a n as)"

definition
  "interpretation" :: "'a xmlt \<Rightarrow> 'a redpair xmlt"
where
  "interpretation xml2name \<equiv> 
      (xml1many_elements_gen ''interpretation'' 
       CPF_Parser.interpretation_type
       (\<lambda>type. case type of
          Natural deg \<Rightarrow> if deg \<le> 1 then (\<lambda>c. (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial CPF_Parser.int_coeff x \<guillemotright>= lpoly_of (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c))
                        (\<lambda>f a poly.  (Alter1 ((f,a),(fst poly, fit_length (get_comp_0 c) a (snd poly)))))))
                   int_components else (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial CPF_Parser.int_coeff x \<guillemotright>= (\<lambda> p. return (poly_of p)))
                        (\<lambda>f a poly.  (Alter9 ((f,a),poly))))
        | Rational d deg \<Rightarrow> if deg \<le> 1 then (\<lambda> c. (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial Xml.rat x \<guillemotright>= lpoly_of (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c))
                        (\<lambda>f a poly.  (Alter2 ((f,a),(fst poly, fit_length (get_comp_0 c) a (snd poly))))))) 
                    (rat_components d) else (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial Xml.rat x \<guillemotright>= (\<lambda> p. return (poly_of p)))
                        (\<lambda>f a poly.  (Alter10 ((f,a),poly))))
        | Arctic \<Rightarrow> (\<lambda> c. (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial CPF_Parser.arctic_coeff x \<guillemotright>= lpoly_of (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c))
                        (\<lambda>f a poly.  (Alter3 ((f,a),(fst poly, fit_length (get_comp_0 c) a (snd poly))))))) 
                    arctic_components
        | Arctic_rat d \<Rightarrow> (\<lambda> c. (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial CPF_Parser.arctic_rat_coeff x \<guillemotright>= lpoly_of (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c))
                        (\<lambda>f a poly.  (Alter7 ((f,a),(fst poly, fit_length (get_comp_0 c) a (snd poly))))))) 
                    (arctic_rat_components d)
        | Int_mat n sd \<Rightarrow> (\<lambda> c. (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial (CPF_Parser.mat_coeff 0 CPF_Parser.int_coeff) x \<guillemotright>= lpoly_of (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c))
                        (\<lambda>f a poly.  (Alter4 ((f,a),(fst poly, fit_length (get_comp_0 c) a (snd poly))))))) 
                    (mat_components n sd int_components)
        | Rat_mat d n sd \<Rightarrow> (\<lambda>c. (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial (CPF_Parser.mat_coeff 0 Xml.rat) x \<guillemotright>= lpoly_of (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c))
                        (\<lambda>f a poly.  (Alter5 ((f,a),(fst poly, fit_length (get_comp_0 c) a (snd poly))))))) 
                    (mat_components n sd (rat_components d))
        | Arc_mat n \<Rightarrow> (\<lambda>c. (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial (CPF_Parser.mat_coeff MinInfty CPF_Parser.arctic_coeff) x \<guillemotright>= lpoly_of (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c))
                        (\<lambda> f a poly. (Alter6 ((f,a),(fst poly, fit_length (get_comp_0 c) a (snd poly))))))) 
                    (mat_arctic_components n arctic_components)
        | Arc_rat_mat d n \<Rightarrow> (\<lambda>c. (xml3elements ''interpret''
                        xml2name
                        (Xml.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial (CPF_Parser.mat_coeff MinInfty_rat CPF_Parser.arctic_rat_coeff) x \<guillemotright>= lpoly_of (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c))
                        (\<lambda>f a poly. (Alter8 ((f,a),(fst poly, fit_length (get_comp_0 c) a (snd poly))))))) 
                    (mat_arctic_components n (arctic_rat_components d))
        )
      (\<lambda> type pi. case type of 
            Natural deg \<Rightarrow> if (deg \<le> 1) then Int_carrier (map alter1 pi) else Int_nl_carrier (map alter9 pi)
          | Rational d deg \<Rightarrow> if (deg \<le> 1) then Rat_carrier d (map alter2 pi) else Rat_nl_carrier d (map alter10 pi)
          | Arctic \<Rightarrow> Arctic_carrier (map alter3 pi)
          | Arctic_rat d \<Rightarrow> Arctic_rat_carrier d (map alter7 pi)
          | Int_mat n sd \<Rightarrow> Int_mat_carrier n sd (map alter4 pi)
          | Rat_mat d n sd \<Rightarrow> Rat_mat_carrier d n sd (map alter5 pi)
          | Arc_mat n \<Rightarrow> Arc_mat_carrier n (map alter6 pi)
          | Arc_rat_mat d n \<Rightarrow> Arc_rat_mat_carrier d n (map alter8 pi)
      ))
    "
hide_const (open) "interpretation"

datatype path_order_kind = Kind_lpo | Kind_mpo | Kind_rpo

fun kind_of_path_order :: "path_order_kind list \<Rightarrow> path_order_kind" where
  "kind_of_path_order [k] = k"
| "kind_of_path_order (k#ks) = (if kind_of_path_order ks = k then k else Kind_rpo)"

(*if all list entries are the same, we have either LPO or MPO,
otherwise, we have RPO*)
lemma "(\<forall>x\<in>set (s#ss). x = Kind_lpo) \<longleftrightarrow> kind_of_path_order (s#ss) = Kind_lpo"
  by (induct ss arbitrary: s) auto

lemma "(\<forall>x\<in>set (s#ss). x = Kind_mpo) \<longleftrightarrow> kind_of_path_order (s#ss) = Kind_mpo"
  by (induct ss arbitrary: s) auto

      
definition
  precedence :: "'a xmlt \<Rightarrow> ('a prec_repr \<times> path_order_kind) xmlt"
where
  "precedence xml2name \<equiv> xml_many_elements ''statusPrecedence'' 
  (xml4elements ''statusPrecedenceEntry''
      xml2name
      (Xml.nat ''arity'')
      (Xml.nat ''precedence'')
      (xml_options [
        (''lex'', xml0elements ''lex'' Kind_lpo),
        (''mul'', xml0elements ''mul'' Kind_mpo)])
      (\<lambda>f a p s. (((f, a), p), s))
  ) (\<lambda>xs. (map fst xs, kind_of_path_order (map snd xs)))"
hide_const (open) precedence

definition
  path_order :: "'a xmlt \<Rightarrow> 'a redpair xmlt"
where
  "path_order xml2name \<equiv> xml1to2elements ''pathOrder''
    (\<lambda>xml. do {
      (pr, kind) \<leftarrow> CPF_Parser.precedence xml2name xml;
      case kind of
        Kind_lpo \<Rightarrow> return (LPO pr)
      | Kind_mpo \<Rightarrow> return (MPO pr)
      | Kind_rpo \<Rightarrow> error ''full RPO not supported''
    })
    (CPF_Parser.afs xml2name)
    (\<lambda>redpair afo. redpair (option_case [] id afo))"
hide_const (open) path_order

definition
  redpair :: "'a xmlt \<Rightarrow> 'a redpair xmlt"
where
  "redpair xml2name = 
     xml1element ''redPair'' 
     (xml_options [(''interpretation'', CPF_Parser.interpretation xml2name), (''pathOrder'', CPF_Parser.path_order xml2name)]) id"
hide_const (open) redpair

definition
  ordering_constraint_proof :: "'a xmlt \<Rightarrow> 'a redpair xmlt"
where
  "ordering_constraint_proof xml2name = 
     xml1element ''orderingConstraintProof'' (CPF_Parser.redpair xml2name) id"
hide_const (open) ordering_constraint_proof

definition
  id_labeling :: "(('f,'a list)lab,'a,'a list + ('f,'a list)lab list)label" 
where
  "id_labeling _ cs = Inl cs" 

definition
  scg_position :: "nat xmlt"
where
  "scg_position \<equiv> Xml.nat ''position''"
hide_const (open) scg_position

definition
  scg :: "'f xmlt \<Rightarrow> (('f,string)rule \<times> ((nat \<times> nat)list \<times> (nat \<times> nat)list)) xmlt"
where
  "scg xml2name \<equiv> xml1many_elements ''sizeChangeGraph'' 
    (CPF_Parser.rule xml2name)
    (xml3elements ''edge''
       CPF_Parser.scg_position
       (Xml.bool ''strict'')
       CPF_Parser.scg_position
       (\<lambda>p s q. (s,(p,q)))
    )
    (\<lambda>lr edges. (lr,(map snd (filter fst edges), map snd (filter (\<lambda> x. \<not> fst x) edges))))"
hide_const (open) scg

definition uncurry_info :: "'f xmlt \<Rightarrow> ('f,string) uncurry_info xmlt"
where "uncurry_info xml2name \<equiv> xml4elements ''uncurryInformation''
         xml2name
         (xml_many_elements ''uncurriedSymbols''
            (xml2many_elements ''uncurriedSymbolEntry''
                xml2name
                (Xml.nat ''arity'')
                xml2name
                (\<lambda> f n fs. ((f,n),fs)))
            id)
         (xml1element ''uncurryRules'' (CPF_Parser.rules xml2name) id) 
         (xml1element ''etaRules'' (CPF_Parser.rules xml2name) id) 
         (\<lambda> a sml U E. (a,sml,U,E))"

abbreviation (input) (* do not use for output. would be very confusing. *)
  dummy_var :: string where "dummy_var \<equiv> []"

function
  xml2dpProof :: "('f, label_type) lab xmlt \<Rightarrow> ('f, label_type, string) DPProof xmlt"
where
  "xml2dpProof xml2name x = xml1element ''dpProof'' (xml_options [ 
  (''pIsEmpty'',xml0elements ''pIsEmpty'' 
     PisEmpty),
  (''depGraphProc'',xml_many_elements ''depGraphProc''
     (xml2nd_choice ''component'' 
         (xml1element ''dps'' (CPF_Parser.rules xml2name) id)
         ''realScc''
         (xml2dpProof xml2name)
         (\<lambda> dps prfOpt. (prfOpt, dps)))
     DepGraphProc),
  (''redPairProc'',xml3elements ''redPairProc'' 
     (CPF_Parser.ordering_constraint_proof xml2name) 
     (xml1element ''dps'' (CPF_Parser.rules xml2name) id) 
     (xml2dpProof xml2name) 
     RedPairProc),
  (''redPairUrProc'',xml4elements ''redPairUrProc'' 
     (CPF_Parser.ordering_constraint_proof xml2name) 
     (xml1element ''dps'' (CPF_Parser.rules xml2name) id) 
     (xml1element ''usableRules'' (CPF_Parser.rules xml2name) id) 
     (xml2dpProof xml2name)
     RedPairUrProc),
  (''monoRedPairProc'',xml4elements ''monoRedPairProc'' 
     (CPF_Parser.ordering_constraint_proof xml2name) 
     (xml1element ''dps'' (CPF_Parser.rules xml2name) id) 
     (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
     (xml2dpProof xml2name)
     MonoRedPairProc),
  (''monoRedPairUrProc'',xml5elements ''monoRedPairUrProc'' 
     (CPF_Parser.ordering_constraint_proof xml2name) 
     (xml1element ''dps'' (CPF_Parser.rules xml2name) id) 
     (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
     (xml1element ''usableRules'' (CPF_Parser.rules xml2name) id) 
     (xml2dpProof xml2name)
     MonoRedPairUrProc),
  (''semlabProc'',xml4elements ''semlabProc'' 
     (CPF_Parser.sl_inter xml2name) 
     (xml1element ''dps'' (CPF_Parser.rules xml2name) id) 
     (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
     (xml2dpProof xml2name)
     (\<lambda>i. case i of Inl (i,True) \<Rightarrow> SemlabProc i (Inl id_labeling)
                  | Inl (i,False) \<Rightarrow> SemlabProc i (Inr (dummy_var,True))
                  | Inr delt_opt  \<Rightarrow> RootlabProc delt_opt)),
  (''subtermProc'',xml1many2elements_gen ''subtermProc''
     (CPF_Parser.proj xml2name)
     (\<lambda>\<pi>. CPF_Parser.projected_rseq xml2name (create_proj \<pi>))
     (xml1element ''dps'' (CPF_Parser.rules xml2name) id)
     (xml2dpProof xml2name)
     SubtermCriterionProc),
  (''flatContextClosureProc'', xml5elements ''flatContextClosureProc''
    (xml1element ''freshSymbol'' xml2name id)
    (CPF_Parser.flat_contexts xml2name)
    (xml1element ''dps'' (CPF_Parser.rules xml2name) id)
    (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
    (xml2dpProof xml2name)
    FccProc),
  (''uncurryProc'',xml_choice ''uncurryProc'' 
   [xml4elements ''uncurryProc'' 
     (uncurry_info xml2name) 
     (xml1element ''dps'' (CPF_Parser.rules xml2name) id) 
     (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
     (xml2dpProof xml2name)
     (UncurryProc None),
    xml5elements ''uncurryProc'' 
     (Xml.nat ''applicativeTop'')
     (uncurry_info xml2name) 
     (xml1element ''dps'' (CPF_Parser.rules xml2name) id) 
     (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
     (xml2dpProof xml2name)
     (\<lambda> n. UncurryProc (Some n))]),
  (''unlabProc'',xml3elements ''unlabProc''
     (xml1element ''dps'' (CPF_Parser.rules xml2name) id) 
     (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
     (xml2dpProof xml2name)
     UnlabProc),
  (''sizeChangeProc'', xml1many_elements ''sizeChangeProc''
     (xml_options [
         (''subtermCriterion'', xml0elements ''subtermCriterion'' (None)),
         (''reductionPair'', xml2elements ''reductionPair'' 
              (CPF_Parser.ordering_constraint_proof xml2name) 
              (xml1element ''usableRules'' (CPF_Parser.rules xml2name) id) 
              (\<lambda>redp ur. Some (redp,ur)))
     ])
     (CPF_Parser.scg xml2name)
     (\<lambda>version. case version of 
         None \<Rightarrow> SizeChangeSubTerm
       | Some redp_ur \<Rightarrow> SizeChangeRedPair (fst redp_ur) (snd redp_ur))
  )    
  ]) id x"
by (pat_completeness) auto


definition
  xmlIgnore :: "string \<Rightarrow> unit xmlt"
where
  "xmlIgnore name xml \<equiv> if name = (Xml.get_name xml) then return () else error (''expected '' @ name @ '' element, but found '' @ to_string xml)"


definition bounds_type :: "boundstype xmlt" where
  "bounds_type \<equiv> xml1element ''type'' 
     (xml_options [
         (''roof'', xml0elements ''roof'' Roof),
         (''match'', xml0elements ''match'' Match)
     ])
     id"

definition bounds_bound :: "nat xmlt" where
  "bounds_bound \<equiv> Xml.nat ''bound''"

definition state :: "string xmlt" where
  "state = Xml.text ''state''"

definition final_states :: "string list xmlt" where
  "final_states = xml_many_elements ''finalStates'' state id"

definition transition_lhs :: "'a xmlt \<Rightarrow> (('a \<times> nat) \<times> string list + string) xmlt" where
  "transition_lhs xml2name x = (if (case x of (XML _ _ cs _) \<Rightarrow> length cs = 1) then 
    (xml1element ''lhs'' state Inr) x
   else
    xml2many_elements ''lhs''
    (xml2name)
    (Xml.nat ''height'')
    (state)
    (\<lambda>f h qs. Inl ((f, h), qs)) x)"

definition transition :: "'a xmlt \<Rightarrow> ((string, 'a \<times> nat) ta_rule + (string \<times> string)) xmlt" where
  "transition xml2name = xml2elements ''transition''
    (transition_lhs xml2name)
    (xml1element ''rhs'' state id)
    (\<lambda> lhs q. case lhs of Inl (f, qs) \<Rightarrow> Inl (TA_rule f qs q) | Inr q' \<Rightarrow> Inr (q',q))"

definition transitions :: "'a xmlt \<Rightarrow> (((string, 'a \<times> nat) ta_rule list) \<times> ((string \<times> string)list)) xmlt" where
  "transitions xml2name = xml_many_elements ''transitions'' (transition xml2name)
     (\<lambda> rls. let (rules, eps) = partition (\<lambda> rl. case rl of Inl _ \<Rightarrow> True | Inr _ \<Rightarrow> False) rls;
                  ruls = map (\<lambda> rl. case rl of Inl r \<Rightarrow> r) rules;
                  ep   = map (\<lambda> rl. case rl of Inr e \<Rightarrow> e) eps
                in (ruls,ep))"


definition bounds_tree_automaton :: "'a xmlt \<Rightarrow> (string, 'a \<times> nat) tree_automaton xmlt" where
  "bounds_tree_automaton xml2name \<equiv> xml2elements ''treeAutomaton''
    final_states
    (transitions xml2name)
    (\<lambda> f (t, eps). Tree_Automaton f t eps)"

definition bounds_info :: "'a xmlt \<Rightarrow> ('a, string) bounds_info xmlt" where
  "bounds_info xml2name \<equiv> xml4elements ''bounds''
     (bounds_type)
     (bounds_bound)
     final_states
     (bounds_tree_automaton xml2name)
     (Bounds_Info)"

function
  xml2trsProof ::
    "('f, label_type) lab xmlt 
     \<Rightarrow> ('f, label_type, string) TRSProof xmlt"
where "xml2trsProof xml2name x = xml1element ''trsTerminationProof'' (xml_options [
   (''rIsEmpty'',xml0elements ''rIsEmpty''
      RisEmpty),
   (''dpTrans'',xml3elements ''dpTrans'' 
      (xml1element ''dps'' (CPF_Parser.rules xml2name) id)
      (Xml.bool ''markedSymbols'')
      (xml2dpProof xml2name)
      (\<lambda> dps _ prf. DPTrans dps prf)),
   (''ruleRemoval'',xml3elements ''ruleRemoval''
      (CPF_Parser.ordering_constraint_proof xml2name) 
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
      (xml2trsProof xml2name)
      RuleRemoval),
  (''unlab'',xml2elements ''unlab''
     (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
     (xml2trsProof xml2name)
     Unlab),
  (''stringReversal'',xml2elements ''stringReversal''
     (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
     (xml2trsProof xml2name)
     StringReversal),
  (''semlab'',xml3elements ''semlab'' 
     (CPF_Parser.sl_inter xml2name) 
     (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
     (xml2trsProof xml2name)
     (\<lambda>i. case i of Inl (i,True)  \<Rightarrow> Semlab i (Inl id_labeling)
                  | Inl (i,False) \<Rightarrow> Semlab i (Inr dummy_var)
                  | Inr _  \<Rightarrow> Rootlab)),
  (''bounds'', (\<lambda>xml. bounds_info xml2name xml \<guillemotright>= (return \<circ> Bounds))),
  (''uncurry'',xml3elements ''uncurry'' 
     (uncurry_info xml2name)
     (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
     (xml2trsProof xml2name)
     Uncurry),
  (''flatContextClosure'',xml3elements ''flatContextClosure''
    (CPF_Parser.flat_contexts xml2name)
    (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
    (xml2trsProof xml2name)
    Fcc
  )
  ]) id x"
by (pat_completeness, auto)

function
  xml2relTrsProof ::
    "('f, label_type) lab xmlt 
     \<Rightarrow> ('f, label_type, string) RelTRSProof xmlt"
where "xml2relTrsProof xml2name x = xml1element ''relativeTerminationProof'' (xml_options [
   (''rIsEmpty'',xml0elements ''rIsEmpty''
      RelRisEmpty),
   (''sIsEmpty'',xml1element ''sIsEmpty''
      (xml2trsProof xml2name)
      RelSisEmpty),
   (''unlab'',xml3elements ''unlab''
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
      (xml2relTrsProof xml2name)
      RelUnlab),
   (''semlab'',xml4elements ''semlab'' 
      (CPF_Parser.sl_inter xml2name) 
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
      (xml2relTrsProof xml2name)
      (\<lambda>i. case i of Inl (i,True)  \<Rightarrow> RelSemlab i (Inl id_labeling)
                   | Inl (i,False) \<Rightarrow> RelSemlab i (Inr dummy_var)
                   | Inr _  \<Rightarrow> RelRootlab)),
   (''ruleRemoval'',xml4elements ''ruleRemoval''
      (CPF_Parser.ordering_constraint_proof xml2name) 
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
      (xml2relTrsProof xml2name)
      RelRuleRemoval),
  (''uncurry'',xml4elements ''uncurry'' 
     (uncurry_info xml2name)
     (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
     (xml1element ''trs'' (CPF_Parser.rules xml2name) id) 
     (xml2relTrsProof xml2name)
     RelUncurry)
  ]) id x"
by (pat_completeness, auto)

function
  xml2dpDisproof :: "'f xmlt \<Rightarrow> ('f, string) DPDisproof xmlt"
where
  "xml2dpDisproof xml2name x = xml1element ''dpNonterminationProof'' (xml_options [
   (''dpRuleRemoval'',xml3elements ''dpRuleRemoval''
      (xml1element ''dps'' (CPF_Parser.rules xml2name) id)
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
      (xml2dpDisproof xml2name)
      DPRuleRemoval),
   (''loop'', xml_change (CPF_Parser.loop xml2name)
        (\<lambda> (s, rseq, \<sigma>, C). DPLoop s rseq \<sigma> C))
   ]) id x"
by pat_completeness auto

function
  xml2trsDisproof :: "'f xmlt \<Rightarrow> ('f, string) TRSDisproof xmlt"
where
  "xml2trsDisproof xml2name x = xml1element ''trsNonterminationProof'' (xml_options [
   (''variableConditionViolated'',xml0elements ''variableConditionViolated''
      NotWellFormed),
   (''loop'', xml_change (CPF_Parser.loop xml2name)
        (\<lambda> (s, rseq, \<sigma>, C). Loop s (map (\<lambda> (x,y,_,z). (x,y,z)) rseq) \<sigma> C)),
   (''ruleRemoval'',xml2elements ''ruleRemoval''
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
      (xml2trsDisproof xml2name)
      NTRuleRemoval),
   (''dpTrans'',xml3elements ''dpTrans''
      (xml1element ''dps'' (CPF_Parser.rules xml2name) id)
      (Xml.bool ''markedSymbols'')
      (xml2dpDisproof xml2name)
      (\<lambda> p _ prf. NTDPTrans p prf)),
   (''stringReversal'',xml2elements ''stringReversal''
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
      (xml2trsDisproof xml2name)
      NTStringReversal)
   ]) id x"
by pat_completeness auto

function
  xml2relTrsDisproof :: "'f xmlt \<Rightarrow> ('f, string) RelTRSDisproof xmlt"
where
  "xml2relTrsDisproof xml2name x = xml1element ''relativeNonterminationProof'' (xml_options [
   (''variableConditionViolated'',xml0elements ''variableConditionViolated''
      RelNotWellFormed),
   (''loop'', xml_change (CPF_Parser.loop xml2name)
        (\<lambda> (s, rseq, \<sigma>, C). RelLoop s rseq \<sigma> C)),
   (''trsNonterminationProof'', xml_change (xml2trsDisproof xml2name) RNotSN),
   (''ruleRemoval'',xml3elements ''ruleRemoval''
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
      (xml2relTrsDisproof xml2name)
      NTRelRuleRemoval)
   ]) id x"
by pat_completeness auto

definition
  xml2trsInput :: "'f xmlt \<Rightarrow> (('f,string) rules \<times> ('f,string)rules option) xmlt"
where
  "xml2trsInput xml2name \<equiv> xml1element ''input''
   (xml1to2elements ''trsInput''
      (xml1element ''trs'' (CPF_Parser.rules xml2name) id)
      (xml1element ''relativeRules'' (CPF_Parser.rules xml2name) id)
      (\<lambda> r s. (r,s)))
   id"

fun
  xml2certProblem :: "
  'f xmlt \<Rightarrow> 
  ('f,label_type)lab xmlt \<Rightarrow> 
  ('f, label_type, string) certificationProblem xmlt"
where "xml2certProblem xml2uname xml2name (XML name egal cs text) = (if name = ''certificationProblem'' \<and> text = None \<and> length cs \<ge> 3 then 
  xml2trsInput xml2uname (cs ! 0) 
      \<guillemotright>= (\<lambda> (r,s_opt). (Xml.text ''cpfVersion'') (cs ! 1)
      \<guillemotright>= (\<lambda> version. (case s_opt of 
                  None \<Rightarrow> 
         xml1element ''proof'' (
            xml_options [
              (''trsTerminationProof'', xml_change (xml2trsProof xml2name) (\<lambda> p. TRSProof r p)),
              (''trsNonterminationProof'', xml_change (xml2trsDisproof xml2name) (\<lambda> p. TRSDisproof r p))
                       ]
         ) id (cs ! 2) 
                | Some s \<Rightarrow> 
         xml1element ''proof'' (
            xml_options [
              (''relativeTerminationProof'', xml_change (xml2relTrsProof xml2name) (\<lambda> p. RelTRSProof (r,s) p)),
              (''relativeNonterminationProof'', xml_change (xml2relTrsDisproof xml2name) (\<lambda> p. RelTRSDisproof (r,s) p))                       ]
         ) id (cs ! 2)
         ) 
      ))
  else error ''could not parse root element'')"

type_synonym lsymbol = "(string, label_type) lab"

definition
  label :: "'a xmlt \<Rightarrow> (label_type + 'a list) xmlt"
where
  "label xml2name \<equiv> xml_options [
     (''numberLabel'', xml_many_elements ''numberLabel'' (Xml.nat ''number'') Inl),
     (''symbolLabel'', xml_many_elements ''symbolLabel'' xml2name Inr)
   ]"
hide_const (open) label 

function
  plain_name :: "lsymbol xmlt"
where
  "plain_name x = xml_options [
     (''name'', xml_change (Xml.text ''name'') UnLab),
     (''sharp'', xml1element ''sharp'' plain_name Sharp),
     (''labeledSymbol'', xml2elements ''labeledSymbol'' plain_name (CPF_Parser.label plain_name)
     (\<lambda>f l. case l of
       Inl ls  \<Rightarrow> Lab f ls
     | Inr ls \<Rightarrow> FunLab f ls))
   ] x"
by pat_completeness auto
hide_const (open) plain_name


primrec
  xmldoc2certProblem :: "xmldoc \<Rightarrow> string + (string,label_type,string)certificationProblem"
where
  "xmldoc2certProblem (XMLDOC header xml) = debug ''0'' ''parsing xml to certification problem'' (
     xml2certProblem (Xml.text ''name'') CPF_Parser.plain_name xml)"

fun
  parseCertProblem :: "string \<Rightarrow> string + (string,label_type,string)certificationProblem"
where
  "parseCertProblem s = (doc_of_string s \<guillemotright>= xmldoc2certProblem)"

declare parseCertProblem.simps[simp del]

(* ... to xml *)

class xml =  
  fixes xml :: "'a \<Rightarrow> xml"

instantiation "term" :: ("show", "show") xml
begin
fun xml_term 
  where "xml_term (Var x) = XML ''var'' [] [] (Some (to_string x))" 
      | "xml_term (Fun f ts) = XML ''funapp'' [] 
              (XML ''name'' [] [] (Some (to_string f)) # 
              (map (\<lambda> t. XML ''arg'' [] [xml_term t] None) ts)) 
           None"
instance ..
end


instantiation certificationProblem :: ("show", type, "show") xml
begin
fun xml_rules :: "string \<Rightarrow> ('a,'c) rules \<Rightarrow> xml"
where "xml_rules tag rls = XML tag Nil [XML ''rules'' [] (map (\<lambda> (l,r). XML ''rule'' [] [
         XML ''lhs'' [] [xml l] None,         
         XML ''rhs'' [] [xml r] None
         ] None) rls) None] None"
fun xml_certificationProblem :: "('a,'b,'c)certificationProblem \<Rightarrow> xml"
where "xml_certificationProblem (TRSProof trs _) = XML ''trsInput'' Nil [xml_rules ''trs'' trs] None"
  |   "xml_certificationProblem (TRSDisproof trs _) = XML ''trsInput'' Nil [xml_rules ''trs'' trs] None"
  |   "xml_certificationProblem (RelTRSProof (r,s) _) = XML ''trsInput'' Nil [xml_rules ''trs'' r, xml_rules ''relativeRules'' s] None"
  |   "xml_certificationProblem (RelTRSDisproof (r,s) _) = XML ''trsInput'' Nil [xml_rules ''trs'' r, xml_rules ''relativeRules'' s] None"
instance ..
end

end
