(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory DP_Transformation_Impl
imports Trs_Impl DP_Framework Substitution
begin

type_synonym ('f, 'v) dp = "('f, 'v) rule" 
type_synonym ('f, 'v) dpL = "('f, 'v) dp list"


definition
  check_DP :: "('f \<Rightarrow> 'f) \<Rightarrow> ('f::show,'v::show)rules \<Rightarrow> ('f,'v)rule_map \<Rightarrow> ('f,'v)dpL \<Rightarrow> shows check"
where
  "check_DP shp r rm P \<equiv> do {
     check_wf_trs r;
     let D = defs_list r;
     check_allm (\<lambda> (f,n). check ((shp f,n) \<notin> set D) (shows_string ''sharping '' +@+ shows f +@+ shows_string '' yields the defined symbol '' +@+ shows (shp f))) D;
     check_subseteq (DP_list shp r rm) P 
       <+? (\<lambda>dp. ''the DP '' +#+ shows_rule dp
         +@+ '' does not appear in the DP problem'' +#+ shows_nl)
   } <+? (\<lambda>s. ''the DP-transformation is not applied correctly.'' +#+ shows_nl +@+ s)"

lemma check_DP:
  assumes check: "isOK(check_DP shp r (computeRuleMap r) P)"
  shows "finite_dpp (set P, set r) \<Longrightarrow> SN (rstep (set r))"
proof -
  assume fin: "finite_dpp (set P, set r)"
  show "SN (rstep (set r))"
  proof -
    from check have wf: "wf_trs (set r)" 
      unfolding check_DP_def by auto
    with computeRuleMapSound2[of r] have "computeRuleMap r \<noteq> None" unfolding wf_trs_def by force
    then obtain rm where rm: "computeRuleMap r = Some rm" by (cases "computeRuleMap r", auto)
    from check have dp: "Trs.DP shp (set r) \<subseteq> set P"
      using DP_list[OF rm] unfolding check_DP_def Let_def by auto
    let ?DP = "Trs.DP shp (set r)"
    from fin have "finite_dpp (set P, set r)"  by simp
    hence fin2: "finite_dpp (set P \<inter> ?DP, set r)" by (unfold finite_dpp_def, auto)
    from dp have "?DP = set P \<inter> ?DP" by auto
    with fin2 have fin: "finite_dpp (DP shp (set r), set r)" by simp
    show "?thesis" 
    proof (rule finite_initial_dpp_imp_SN[OF wf _ fin])
      fix f n
      assume "defined (set r) (f,n)"
      thus "\<not> defined (set r) (shp f,n)"
        using check unfolding check_DP_def Let_def defs_list_sound by auto
    qed
  qed
qed

  


end
