theory RPO_Impl
imports RPO Multiset_Extension_Impl
begin

function
  mpo_impl :: "('f \<Rightarrow> nat \<Rightarrow> 'f \<Rightarrow> nat \<Rightarrow> bool \<times> bool) \<times> ('f \<Rightarrow> bool) \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool \<times> bool"
where
  "mpo_impl _  (Var x) (Var y) = (False, x = y)" |
  "mpo_impl pr  (Var x) (Fun g ts) = (False, ts = [] \<and> (snd pr) g)" |
  "mpo_impl pr  (Fun f ss) (Var y) = (let con = (\<exists> s \<in> set ss. snd (mpo_impl pr s (Var y))) in (con,con))" |
  "mpo_impl pr (Fun f ss) (Fun g ts) = (
    if (\<exists> s \<in> set ss. snd (mpo_impl pr s (Fun g ts)))
       then (True,True)
       else (let (prs,prns) = (fst pr) f (length ss) g (length ts) in 
         if prns \<and> (\<forall> t \<in> set ts. fst (mpo_impl pr (Fun f ss) t))
          then if prs then (True,True) 
            else mul_ext_impl (mpo_impl pr) ss ts
          else (False,False)))"
by pat_completeness auto

termination 
proof (relation "measures [\<lambda> (pr,s,t). size s + size t]")
  fix "pr" and f :: 'f and ss :: "('f,'v)term list" and g :: 'f and ts :: "('f,'v)term list" and x xa y xb x'
  assume s: "xb \<in> set ss" and t: "x' \<in> set ts"
  from s have s1: "size xb \<le> list_size size ss" by (induct ss, auto)
  from t have t1: "size x' \<le> list_size size ts" by (induct ts, auto)
  from s1 t1 s t show "((pr, xb, x'), pr, Fun f ss, Fun g ts) \<in> measures [\<lambda> (pr,s,t). size s + size t]"
    by auto
qed (auto simp: size_simps)

lemma mpo_impl: "mpo pr n s t = mpo_impl pr s t"
proof (induct rule: mpo_impl.induct)
case 1 show ?case by simp
next
case 2 show ?case by simp
next
case 3 thus ?case by auto
next
case (4 p f ss g ts)
  show ?case
  proof (cases "\<exists> s \<in> set ss. snd (mpo_impl p s (Fun g ts))")
  case True
    then obtain s where
      s_in_ss: "s \<in> set ss" and
      snd_mpo_impl: "snd (mpo_impl p s (Fun g ts))"
    by fast
    with 4(1)[OF s_in_ss] have "\<exists>s \<in> set ss. snd (mpo p n s (Fun g ts))"
      by metis
    with True show ?thesis by simp
  next
  case False note Hf = this
    hence Hrewi: "mpo_impl p (Fun f ss) (Fun g ts) = (
      let (prs,prns) = (fst p) f (length ss) g (length ts) in 
      if prns \<and> (\<forall> t \<in> set ts. fst (mpo_impl p (Fun f ss) t))
      then if prs then (True,True) 
           else mul_ext_impl (mpo_impl p) ss ts
      else (False,False))" by simp
    from False and 4(1)
     have " \<not> (\<exists>s\<in>set ss. snd (mpo p n s (Fun g ts)))" by simp
    hence Hrew: "mpo p n (Fun f ss) (Fun g ts) = (
      let (prs,prns) = (fst p) f (length ss) g (length ts) in 
      if prns \<and> (\<forall> t \<in> set ts. fst (mpo p n (Fun f ss) t))
      then if prs then (True,True) 
           else mul_ext (mpo p n) ss ts
      else (False,False))" by simp
    show ?thesis
    proof (cases "(fst p) f (length ss) g (length ts)")
    case (Pair prs prns)
      with Hrewi have Hrewi: "mpo_impl p (Fun f ss) (Fun g ts) = (
       if prns \<and> (\<forall> t \<in> set ts. fst (mpo_impl p (Fun f ss) t))
       then if prs then (True,True) 
            else mul_ext_impl (mpo_impl p) ss ts
       else (False,False))" by simp
      from Pair and Hrew have Hrew: "mpo p n (Fun f ss) (Fun g ts) = (
       if prns \<and> (\<forall> t \<in> set ts. fst (mpo p n (Fun f ss) t))
       then if prs then (True,True) 
            else mul_ext (mpo p n) ss ts
       else (False,False))" by simp
      from 4(2)[OF False _ Pair[symmetric]]
       have Hequiv: "(prns \<and> (\<forall> t \<in> set ts. fst (mpo p n (Fun f ss) t)))
        = (prns \<and> (\<forall> t \<in> set ts. fst (mpo_impl p (Fun f ss) t)))" by simp
      show ?thesis
      proof (cases "prns \<and> (\<forall> t \<in> set ts. fst (mpo p n (Fun f ss) t))")
      case False with Hrewi and Hrew and Hequiv show ?thesis by auto
      next
      case True
        note Hcond1 = this
        with Hequiv have Hcond2: "prns \<and> (\<forall>t\<in>set ts. fst (mpo_impl p (Fun f ss) t))" by simp
        show ?thesis
        proof (cases prs)
        case True with Hrew and Hrewi and Hequiv and Hcond1 show ?thesis by auto
        next
        case False
          with Hrewi and Hequiv and Hcond1
           have Hrewi: "mpo_impl p (Fun f ss) (Fun g ts) = mul_ext_impl (mpo_impl p) ss ts"
           by simp
          from False and Hrew and Hcond1
           have Hrew: "mpo p n (Fun f ss) (Fun g ts) = mul_ext (mpo p n) ss ts" by simp
          from 4(3)[OF Hf _ Pair[symmetric] Hcond2 False] and mul_ext_impl_cong[of ss ts]
           have "mul_ext_impl (mpo_impl p) ss ts = mul_ext_impl (mpo p n) ss ts" by metis
          also have "... = mul_ext (mpo p n) ss ts"
           using mul_ext_impl_sound and mpo_stri_imp_nstri by fast
          finally show ?thesis using Hrew and Hrewi by simp
        qed
      qed
    qed
  qed
qed
         
lemma [code]: "mpo = (\<lambda> p n. mpo_impl p)"
  by (intro ext, simp add: mpo_impl)

end