(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Trs_Impl
imports Trs Term_Impl Substitution
begin

definition
  vars_rule_list' :: "('f, 'v) rule \<Rightarrow> 'v list"
where
  "vars_rule_list' r \<equiv> vars_term_list' (fst r) @ vars_term_list' (snd r)"

lemma vars_rule_list'_sound[simp]:
  "set (vars_rule_list' r) = vars_rule r"
unfolding vars_rule_list'_def vars_rule_def by (cases r) simp

definition
  vars_rule_list :: "('f, 'v) rule \<Rightarrow> 'v list"
where
  "vars_rule_list r \<equiv> remdups (vars_rule_list' r)"



lemma vars_rule_list_sound[simp]:
  "set (vars_rule_list r) = vars_rule r"
unfolding vars_rule_list_def by simp

definition
  vars_trs_list :: "('f, 'v) rule list \<Rightarrow> 'v list"
where
  "vars_trs_list trs \<equiv> remdups (concat (map vars_rule_list' trs))"

lemma vars_trs_list_Nil[simp]: "vars_trs_list [] = []" unfolding vars_trs_list_def by simp

lemma vars_trs_list_sound[simp]:
  "set (vars_trs_list trs) = vars_trs (set trs)"
by (induct trs) (auto simp: vars_trs_def vars_trs_list_def)

lemmas vars_list_defs = vars_trs_list_def vars_rule_list_def

definition
  funs_rule_list' :: "('f, 'v) rule \<Rightarrow> 'f list"
where
  "funs_rule_list' r \<equiv> funs_term_list' (fst r) @ funs_term_list' (snd r)"

lemma funs_rule_list'_sound[simp]:
  "set (funs_rule_list' r) = funs_rule r"
unfolding funs_rule_list'_def funs_rule_def by (cases r) simp

definition
  funs_rule_list :: "('f, 'v) rule \<Rightarrow> 'f list"
where
  "funs_rule_list r \<equiv> remdups (funs_rule_list' r)"

lemma funs_rule_list_sound[simp]:
  "set (funs_rule_list r) = funs_rule r"
unfolding funs_rule_list_def by simp

definition
  funas_rule_list' :: "('f, 'v) rule \<Rightarrow> ('f \<times> nat) list"
where
  "funas_rule_list' r \<equiv> funas_term_list' (fst r) @ funas_term_list' (snd r)"

lemma funas_rule_list'_sound[simp]:
  "set (funas_rule_list' r) = funas_rule r"
unfolding funas_rule_list'_def funas_rule_def by (cases r) simp

definition
  funas_rule_list :: "('f, 'v) rule \<Rightarrow> ('f \<times> nat) list"
where
  "funas_rule_list r \<equiv> remdups (funas_rule_list' r)"

lemma funas_rule_list_sound[simp]:
  "set (funas_rule_list r) = funas_rule r"
unfolding funas_rule_list_def by simp

definition
  funs_trs_list :: "('f, 'v) rule list \<Rightarrow> 'f list"
where
  "funs_trs_list R \<equiv> remdups (concat (map funs_rule_list' R))"

lemma funs_trs_list_sound[simp]:
  "set (funs_trs_list R) = funs_trs (set R)"
by (induct R) (auto simp: funs_trs_def funs_trs_list_def)

definition
  funas_trs_list :: "('f, 'v) rule list \<Rightarrow> ('f \<times> nat) list"
where
  "funas_trs_list R \<equiv> remdups (concat (map funas_rule_list' R))"

lemma funas_trs_list_sound[simp]:
  "set (funas_trs_list R) = funas_trs (set R)"
by (induct R) (auto simp: funas_trs_def funas_trs_list_def)



lemmas funs_list_defs = funs_trs_list_def funs_rule_list_def
lemmas funas_list_defs = funas_trs_list_def funas_rule_list_def

fun
  defs_list :: "('f, 'v) rule list \<Rightarrow> ('f \<times> nat) list"
where
  "defs_list [] = []" |
  "defs_list ((Fun f ss, r) # ts) = (let fs = defs_list ts; n = length ss in if (f,n) \<in> set fs then fs else (f,n) # fs)" |
  "defs_list ((Var x, r) # ts) = defs_list ts"

lemma defs_list_sound:
  "defined (set R) (f, n) = ((f,n) \<in> set (defs_list R))"
proof (induct R, simp add: defined_def)
  case (Cons lr R)
  show ?case
  proof (cases lr)
    case (Pair l r)
    show ?thesis
    proof (cases l)
      case (Var x)
      with Pair Cons show ?thesis by (auto simp: defined_def)
    next
      case (Fun g ss)
      with Pair
      have "defined (set (lr # R)) (f, n) = defined (set ((Fun g ss,r) # R)) (f, n)" by simp
      also have "\<dots> =  (\<exists>l ra. (l, ra) \<in> set ((Fun g ss, r) # R) \<and> root l = Some f \<and> num_args l = n)" (is "_ = ?P (set ((Fun g ss,r) # R))") by (simp add: defined_def)
      also have "\<dots> = (?P (set R) \<or> ?P {(Fun g ss, r)})" by auto
      also have "\<dots> = ((f,n) \<in> set (defs_list R) \<or> ?P {(Fun g ss, r)})"  using Cons by (unfold defined_def, auto)
      also have "\<dots> = ((f,n) \<in> set (defs_list ((Fun g ss,r) # R)))" (is "?l = ?r")
      proof (cases "(g,n) \<in> set (defs_list R)")
	case False
	thus "?l = ?r" by (auto simp: Let_def)
      next
	case True
	thus "?l = ?r" by (simp add: Let_def, blast)
      qed		
      finally show ?thesis using Fun Pair Cons by simp
    qed
  qed
qed


lemma distinct_funas_trs_list:
  fixes trs::"('f,'v)rules" shows "distinct(funas_trs_list trs)"
  unfolding funas_trs_list_def by simp


definition
  check_left_linear_trs :: "('f::show, 'v :: show) rules \<Rightarrow> shows check"
where
  "check_left_linear_trs trs \<equiv>
     check_all (\<lambda>r. linear_term (fst r)) trs
       <+? (\<lambda> _. shows trs +@+ shows_nl +@+ ''is not left-linear'' +#+ shows_nl)"

lemma check_left_linear_trs[simp]: "isOK(check_left_linear_trs R) = (left_linear_trs (set R))"
  unfolding check_left_linear_trs_def left_linear_trs_def
  by auto

definition
  check_wf_trs :: "('f::show, 'v::show) rules \<Rightarrow> shows check"
where
  "check_wf_trs R \<equiv> (
     check_allm (\<lambda>rule. do {
       check (\<not> is_var (fst rule)) (shows_string ''variable left-hand side in'');
       check_subseteq (vars_term_list (snd rule)) (vars_term_list (fst rule))
         <+? (\<lambda>x. ''free variable '' +#+ shows x
               +@+ shows_string '' in right-hand side of'')
     } <+? (\<lambda>s. s +@+ '' rule '' +#+ shows_rule rule +@+ shows_nl)
     ) R
   <+? (\<lambda>e. ''the TRS is not well-formed'' +#+ shows_nl +@+ e)
   )"

lemma check_wf_trs_conf[simp]:
  shows "isOK (check_wf_trs R) = wf_trs (set R)"
proof
  assume "isOK (check_wf_trs R)"
  hence "\<forall>(l, r)\<in>set R. \<not> is_var l" unfolding check_wf_trs_def by auto
  hence "\<forall>(l, r)\<in>set R. \<exists>f ts. l = Fun f ts" by best
  moreover have "\<forall>(l, r)\<in>set R. vars_term r \<subseteq> vars_term l"
  proof (intro ballI2)
    fix l r assume "(l, r) \<in> set R"
    from `isOK (check_wf_trs R)`
      have "isOK(check_allm (\<lambda>(l, r). check_subseteq (vars_term_list r) (vars_term_list l)) R)"
      unfolding check_wf_trs_def by auto
    hence "\<forall>(l, r)\<in>set R. isOK (check_subseteq (vars_term_list r) (vars_term_list l))"
      unfolding split_def by simp
    with `(l, r) \<in> set R` show "vars_term r \<subseteq> vars_term l" by auto
  qed
  ultimately show "wf_trs (set R)" unfolding wf_trs_def by auto
next
  assume "wf_trs (set R)" thus "isOK (check_wf_trs R)"
    unfolding wf_trs_def check_wf_trs_def by force
qed


definition
  check_not_wf_trs :: "('f::show, 'v::show) rules \<Rightarrow> shows check"
where
  "check_not_wf_trs R \<equiv>
     check (\<not> isOK (check_wf_trs R)) (''The TRS is well formed'' +#+ shows_nl)"

lemma check_not_wf_trs:
  assumes "isOK(check_not_wf_trs R)"
  shows "\<not> SN (rstep (set R))"
proof -
  from assms have "\<not> wf_trs (set R)" unfolding check_not_wf_trs_def by auto
  with SN_rstep_imp_wf_trs show ?thesis by auto
qed

lemma [code]: "instance_rule lr st = (match_list [(fst lr, fst st), (snd lr, snd st)] \<noteq> None)" (is "?l = (match_list ?list \<noteq> None)")
proof
  assume ?l
  then obtain \<sigma> where fst: "fst lr = fst st \<cdot> \<sigma>" and snd: "snd lr = snd st \<cdot> \<sigma>" unfolding instance_rule_def by auto
  have "\<And> t l. (t, l) \<in> set ?list \<Longrightarrow> l \<cdot> \<sigma> = t" by (auto simp del: match_list.simps simp: fst snd)
  from match_list_complete[of ?list \<sigma>, OF this]
  obtain \<sigma>' where "match_list ?list = Some \<sigma>'" by blast
  thus "match_list ?list \<noteq> None" by simp
next
  assume "match_list ?list \<noteq> None"
  then obtain \<sigma> where "match_list ?list = Some \<sigma>" by (cases "match_list ?list", auto)
  from match_list_sound[OF this] have "fst lr = fst st \<cdot> \<sigma>" and "snd lr = snd st \<cdot> \<sigma>" by (force+)
  thus ?l unfolding instance_rule_def by auto
qed


definition check_CS_subseteq :: "('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rule check"
where "check_CS_subseteq R S \<equiv> check_allm (\<lambda> (l,r). check (list_ex (instance_rule (l,r)) S) (l,r)) R"

lemma check_CS_subseteq[simp]: "isOK(check_CS_subseteq R S) = (CS (set R) \<subseteq> CS (set S))" (is "?l = ?r")
proof
  assume ?l
  show ?r
  proof
    fix x y
    assume "(x,y) \<in> CS (set R)"    
    thus "(x,y) \<in> CS (set S)"
    proof (induct)
      case (subst_closureI s t \<sigma>)
      with `?l`[unfolded check_CS_subseteq_def]
      obtain l r \<delta> where lr: "(l,r) \<in> set S" and s: "s = l \<cdot> \<delta>" and t: "t = r \<cdot> \<delta>"
        unfolding list_ex_iff by (force simp: instance_rule_def)
      show ?case unfolding s t
        using subst_closure.subst_closureI[OF lr, of "\<delta> \<circ>s \<sigma>"]
        by auto
    qed
  qed
next
  assume ?r
  show ?l
    unfolding check_CS_subseteq_def
  proof (simp, intro ballI)
    fix lr
    assume mem: "lr \<in> set R"
    obtain l r where lr: "lr = (l,r)" by (cases lr, auto)
    with mem have "(l,r) \<in> CS (set R)" using subset_subst_closure by auto
    with `?r` have "(l,r) \<in> CS (set S)" by auto
    thus "list_ex (instance_rule lr) S" unfolding lr
    proof (induct)
      case (subst_closureI s t \<sigma>)
      thus ?case unfolding list_ex_iff instance_rule_def by force
    qed
  qed
qed


instantiation pos :: "show" begin
  fun
    shows_pos :: "pos \<Rightarrow> shows"
  where
    "shows_pos (i <# \<epsilon>) = shows i" |
    "shows_pos (i <# ps) = (shows i +@+ shows_pos ps)" |
    "shows_pos \<epsilon> = shows_string ''e''"
  definition "shows_prec (d::nat) (p::pos) = shows_pos p"
  definition "shows_list_pos (ps::pos list) = shows_list_aux shows ps"

  lemma assoc_pos: "shows_prec d (p::pos) r @ s = shows_prec d p (r @ s)"
  unfolding shows_prec_pos_def
  proof (induct p)
    case (PCons i ps) thus ?case by (cases ps) simp_all
  qed simp

  lemma assoc_pos_elt: "\<forall>(p::pos)\<in>set ps. shows p r @ s = shows p (r @ s)"
  using assoc_pos by auto

  instance by
    (intro_classes, rule assoc_pos, unfold shows_list_pos_def)
    (rule shows_list_aux_assoc[OF assoc_pos_elt])
end

definition
  check_rstep ::
    "('f::show, 'v::show) rules \<Rightarrow> pos \<Rightarrow> ('f, 'v) rule \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term
     \<Rightarrow> shows check"
where
  "check_rstep R p r s t \<equiv> do {
     let sps = poss_list s;
     let tps = poss_list t;
     check (list_ex (\<lambda> rule. r =v rule) R)
       (shows_rule r +@+ '' is not a rule of'' +#+ shows_nl +@+ shows_trs R +@+ shows_nl);
     check (p \<in> set sps) (shows p +@+ '' is not a position of '' +#+ shows s +@+ shows_nl);
     check (p \<in> set tps) (shows p +@+ '' is not a position of '' +#+ shows t +@+ shows_nl);
     let C = ctxt_of_pos_term p s;
     let D = ctxt_of_pos_term p t;
     let u = subt_at s p;
     let v = subt_at t p;
     (case match_list [(u,fst r), (v, snd r)] of
       Some \<sigma> \<Rightarrow>
         check (C = D)
           (''the term '' +#+ shows t
             +@+ '' does not result from a proper application of rule ''
             +#+ shows_nl +@+ shows_rule r +@+ '' at position '' +#+ shows p +@+ shows_nl)
     | None \<Rightarrow>
       error (''the term '' +#+ shows t
             +@+ '' does not result from a proper application of rule ''
             +#+ shows_nl +@+ shows_rule r +@+ '' at position '' +#+ shows p +@+ shows_nl))
   }"

lemma check_rstep_sound:
   "isOK (check_rstep R p r s t) \<Longrightarrow> (\<exists> \<sigma> r. (s,t) \<in> rstep_r_p_s (set R) r p \<sigma>)"
proof -
  assume check: "isOK (check_rstep R p r s t)"  
  let ?C = "ctxt_of_pos_term p s"
  let ?D = "ctxt_of_pos_term p t"
  let ?mlist = "[(subt_at s p, fst r), (subt_at t p, snd r)]"
  from check[unfolded check_rstep_def] have "p \<in> poss s" and "p \<in> poss t" and "\<exists> rule \<in> set R. r =v rule" by (auto simp: list_ex_iff)
  then obtain rule where rule: "rule \<in> set R" and inst: "r =v rule" by auto
  from inst obtain \<delta> where fst: "fst r = fst rule \<cdot> \<delta>" and snd: "snd r = snd rule \<cdot> \<delta>" unfolding eq_rule_mod_vars_def instance_rule_def by auto
  from check[unfolded check_rstep_def]
  obtain \<sigma> where some: "match_list [(subt_at s p, fst r), (subt_at t p, snd r)] = Some \<sigma>" by force
  from match_list_sound[OF this] have "fst r \<cdot> \<sigma> = s |_ p" and "snd r \<cdot> \<sigma> = t |_ p" by auto 
  hence "s = ?C\<langle>fst rule \<cdot> \<delta> \<circ>s \<sigma>\<rangle>" and "t = ?D\<langle>snd rule \<cdot> \<delta> \<circ>s \<sigma>\<rangle>" using ctxt_supt_id[OF `p \<in> poss s`] ctxt_supt_id[OF `p \<in> poss t`] fst snd by simp+
  moreover from check[unfolded check_rstep_def, simplified] and some
    have "?C = ?D" by simp 
  ultimately have "(s,t) \<in> rstep_r_p_s (set R) rule p (\<delta> \<circ>s \<sigma>)" using `p \<in> poss s` and rule unfolding rstep_r_p_s_def some Let_def by force
  thus ?thesis by blast
qed

(* TODO: move *)
lemma hole_pos_ctxt_of_pos_term: assumes "p \<in> poss t"
  shows "hole_pos (ctxt_of_pos_term p t) = p"
  using assms
proof (induct t arbitrary: p, simp)
  case (Fun f ts)
  show ?case 
  proof (cases p)
    case Empty
    thus ?thesis using Fun by auto
  next
    case (PCons i q)
    with Fun(2) have i: "i < length ts" and q: "q \<in> poss (ts ! i)" by auto
    hence "ts ! i \<in> set ts" by auto
    from Fun(1)[OF this q] PCons i show ?thesis by simp
  qed
qed

(* TODO: move *)
lemma hole_pos_id_ctxt: assumes "C\<langle>s\<rangle> = t"
  shows "ctxt_of_pos_term (hole_pos C) t = C"
using assms 
proof (induct C arbitrary: t, simp)
  case (More f bef C aft)
  thus ?case
  proof (cases t, simp)
    case (Fun g ts)
    with More have [simp]: "g = f" by simp
    from Fun More have bef: "take (length bef) ts = bef" by auto    
    from Fun More have aft: "drop (Suc (length bef)) ts = aft" by auto
    from Fun More have Cs: "C\<langle>s\<rangle> = ts ! length bef" by auto
    from Fun More show ?thesis by (simp add: bef aft More(1)[OF Cs])
  qed
qed

lemma check_rstep_complete:
   "\<exists> \<sigma>. (s,t) \<in> rstep_r_p_s (set R) r p \<sigma> \<Longrightarrow> isOK (check_rstep R p r s t)"
proof -
  assume "\<exists> \<sigma>. (s,t) \<in> rstep_r_p_s (set R) r p \<sigma>"
  then obtain \<sigma> where step: "(s,t) \<in> rstep_r_p_s (set R) r p \<sigma>" ..
  let ?C = "ctxt_of_pos_term p s"
  let ?C' = "ctxt_of_pos_term p t"
  from step have p: "p \<in> poss s" and r: "r \<in> set R" and s: "?C\<langle>fst r \<cdot> \<sigma>\<rangle> = s" and t: "?C\<langle>snd r \<cdot> \<sigma>\<rangle> = t"
    unfolding rstep_r_p_s_def by (auto simp: Let_def)  
  from hole_pos_ctxt_of_pos_term[OF p] have Cp: "hole_pos ?C = p" .
  from hole_pos_id_ctxt[OF t, simplified Cp] have C: "?C = ?C'" by auto
  from t[symmetric] have p': "p \<in> poss t" using hole_pos_poss[of ?C, simplified Cp] by best
  from r have r: "r \<in> set R" by (simp)
  have rr: "r =v r" unfolding eq_rule_mod_vars_def instance_rule_def 
    by (simp, rule exI[of _ Var], simp)
  have "s |_ p = ?C\<langle>fst r \<cdot> \<sigma>\<rangle> |_ hole_pos ?C" by (simp only: s Cp)
  also have "\<dots> = fst r \<cdot> \<sigma>" by (rule subt_at_hole_pos)
  finally have proj: "s |_ p = fst r \<cdot> \<sigma>" .
  have "t |_ p = ?C\<langle>snd r \<cdot> \<sigma>\<rangle> |_ hole_pos ?C" by (simp only: t Cp)
  also have "\<dots> = snd r \<cdot> \<sigma>" by (simp only: C, rule subt_at_hole_pos)
  finally have tp: "t |_ p = snd r \<cdot> \<sigma>" .
  from match_list_complete proj tp obtain \<sigma> where match: "match_list [(s |_ p, fst r), (t |_ p, snd r)] = Some \<sigma>" by force
  from r rr have list_ex: "list_ex (\<lambda> rule. r =v rule) R" by (auto simp: list_ex_iff)
  show "isOK(check_rstep R p r s t)"
    by (simp add: check_rstep_def r rr p' p match list_ex C del: match_list.simps)
qed

lemma check_rstep_sound_2[simp]:
  assumes "isOK (check_rstep R p r s t)"
  shows "(s, t) \<in> rstep (set R)"
using check_rstep_sound[OF assms] unfolding rstep_iff_rstep_r_p_s by auto

definition check_rrstep ::
  "('f::show, 'v::show) rules \<Rightarrow> ('f, 'v) rule \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term
     \<Rightarrow> shows check"
where "check_rrstep R r s t \<equiv> check_rstep R \<epsilon> r s t"

lemma check_rrstep: "(\<exists> r \<in> set R. (isOK (check_rrstep R r s t))) = ((s, t) \<in> rrstep (set R))"
using check_rstep_sound[of R \<epsilon> _ s t] check_rstep_complete[of s t R _ \<epsilon>]
unfolding check_rrstep_def rrstep_def rstep_r_p_s_def 
by (auto, blast+)

lemma check_rrstep_sound: assumes "isOK(check_rrstep R r s t)" shows "(s,t) \<in> rrstep (set R)"
using check_rstep_sound[OF assms[unfolded check_rrstep_def]]
unfolding rrstep_def rstep_r_p_s_def
by auto

type_synonym  ('f, 'v) prseq = "(pos \<times> ('f, 'v) rule \<times> bool \<times> ('f, 'v) term) list"
type_synonym  ('f, 'v) rseq = "(pos \<times> ('f, 'v) rule \<times> ('f, 'v) term) list"




fun
  check_steps ::
    "('f::show, 'v::show) rules \<Rightarrow> ('f,'v) rules \<Rightarrow> ('f, 'v) prseq \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
where
  "check_steps P R [] s u = check (s = u) (
     ''the last term of the rewrite sequence'' +#+ shows_nl +@+ shows s +@+ shows_nl
     ''does not correspond to the goal term''  +#+ shows_nl +@+ shows u +@+ shows_nl
   )" |
  "check_steps P R ((_, r, True, t) # prts) s u = do {
     check_rrstep P r s t;
     check_steps P R prts t u
   }" |
  "check_steps P R ((p, r, False, t) # prts) s u = do {
     check_rstep R p r s t;
     check_steps P R prts t u
   }"

lemma check_steps_sound:
  assumes "isOK (check_steps P R rseq s t)"
  shows "(s, t) \<in> (rrstep (set P) \<union> rstep (set R))^^(length rseq)"
proof -
  obtain PR where PR: "PR = rrstep (set P) \<union> rstep (set R)" by auto
  from assms show ?thesis unfolding PR[symmetric] 
  proof (induct rseq arbitrary: s)
    case Nil thus ?case by simp
  next
    case (Cons prt rseq)
    obtain p r or u where prt: "prt = (p, r, or, u)" by (cases prt, blast) 
    from Cons[unfolded prt]
    have IH: "(u, t) \<in> PR ^^ (length rseq)" by (cases or, induct rseq) auto
    have "(s,u) \<in> PR"
    proof (cases or)
      case True
      with Cons[unfolded prt]
      have "isOK (check_rrstep P r s u)" by simp
      hence "(s,u) \<in> rrstep (set P)" by (rule check_rrstep_sound)
      thus ?thesis unfolding PR ..
    next
      case False
      with Cons[unfolded prt]
      have "isOK (check_rstep R p r s u)" by simp
      hence "(s,u) \<in> rstep (set R)" by (rule check_rstep_sound_2)
      thus ?thesis unfolding PR ..
    qed
    with IH show ?case unfolding prt o_def using rel_pow_Suc_I2[of s u PR] by simp
  qed
qed


definition
  check_rsteps ::
    "('f::show, 'v::show) rules \<Rightarrow> ('f, 'v) rseq \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
where
  "check_rsteps R prts s u \<equiv> check_steps [] R (map (\<lambda> (p,r,t). (p,r,False,t)) prts) s u"

lemma check_rsteps_sound[simp]:
  assumes "isOK (check_rsteps R rseq s t)"
  shows "(s, t) \<in> (rstep (set R))^^(length rseq)"
proof -
  have empty: "rrstep {} \<union> rstep (set R) = rstep (set R)" unfolding rrstep_def rstep_r_p_s_def by auto
  from check_steps_sound[OF assms[unfolded check_rsteps_def]] show ?thesis 
    by (auto simp: Let_def empty)
qed

type_synonym ('f,'v)rule_map = "(('f \<times> nat) \<Rightarrow> ('f,'v)rules)option"

fun computeRuleMapH :: "('f,'v)rules \<Rightarrow> (('f \<times> nat) \<times> ('f,'v)rules)list option"
where "computeRuleMapH [] = Some []"
    | "computeRuleMapH ((Fun f ts,r) # rules) = (let n = length ts in case computeRuleMapH rules of None \<Rightarrow> None | Some rm \<Rightarrow> 
            (case extract (\<lambda> (fa,rls). fa = (f,n)) rm of 
                None \<Rightarrow> Some (((f,n), [(Fun f ts,r)]) # rm)
              | Some (bef,(fa,rls),aft) \<Rightarrow>  Some  ((fa,(Fun f ts,r) # rls) # bef @ aft)))"
    | "computeRuleMapH ((Var _, _) # rules) = None"

definition computeRuleMap :: "('f,'v)rules \<Rightarrow> ('f,'v)rule_map"
 where "computeRuleMap rls \<equiv> case computeRuleMapH rls of
           None \<Rightarrow> None
         | Some rm \<Rightarrow> Some (\<lambda> f. case lookup f rm of
               None \<Rightarrow> []
            | Some rls \<Rightarrow> rls)"

lemma computeRuleMapHSound2: "(computeRuleMapH R = None) = (\<exists> (l, r) \<in> set R. root l = None)"
proof (induct R, simp)
  case (Cons rule rules)
  show ?case 
  proof (cases rule)
    case (Pair l r)
    show ?thesis 
    proof (cases l, simp add: Pair Cons)
      case (Fun f ts)
      show ?thesis
        using Pair Cons
      proof (cases "computeRuleMapH rules", simp add: Pair Fun Let_def)
        case (Some rm) note oSome = this
        let ?e = "extract (\<lambda> (fa,rls). fa = (f,length ts)) rm"
        from Some Fun Pair Cons show ?thesis 
        proof (cases ?e, simp)
          case (Some res)
          then obtain bef aft g rls where "?e = Some (bef, (g,rls), aft)" by (cases res, force) 
          with extract_Some[OF this] have rm: "rm = bef @ ((f, length ts),rls) # aft" and e: "?e = Some (bef, ((f,length ts),rls), aft)"
            by auto
          show ?thesis using Cons
            by (simp add: Pair Fun Let_def oSome e)
        qed
      qed
    qed
  qed
qed

lemma computeRuleMapSound2: "(computeRuleMap R = None) = (\<exists> (l, r) \<in> set R. root l = None)"
unfolding computeRuleMap_def
by (simp only: computeRuleMapHSound2[symmetric], cases "computeRuleMapH R", auto)


lemma computeRuleMapHSound: assumes "computeRuleMapH R = Some rm"
  shows "(\<lambda> (f,rls). (f,set rls)) ` set rm = {((f,n),{(l,r) | l r. (l,r) \<in> set R \<and> root l = Some f \<and> num_args l = n}) | f n. {(l,r) | l r. (l,r) \<in> set R \<and> root l = Some f \<and> num_args l = n} \<noteq> {}} \<and> distinct_eq (\<lambda> (f,rls) (g,rls'). f = g) rm"
using assms
proof (induct R arbitrary: rm, force)
  case (Cons rule rules)
  let ?setl = "\<lambda> rm. (\<lambda> (f,rls). (f,set rls)) ` set rm"
  let ?setr = "\<lambda> R. {((f,n),{(l,r) | l r. (l,r) \<in> set R \<and> root l = Some f \<and> num_args l = n}) | f n. {(l,r) | l r. (l,r) \<in> set R \<and> root l = Some f \<and> num_args l = n} \<noteq> {}}"
  show ?case
  proof (cases rule)
    case (Pair l r)
    show ?thesis
    proof (cases l)
      case (Var v)
      with Cons Pair show ?thesis by simp
    next
      case (Fun f ts)
      with Cons Pair show ?thesis 
      proof (cases "computeRuleMapH rules", simp)
        case (Some rrm) note oSome = this
        let ?dis = "distinct_eq (\<lambda> (f,rls) (g,rls'). f = g)"
        from Cons(1)[OF Some] have drrm: "?dis rrm" and srrm: "?setl rrm = ?setr rules" by auto
        show ?thesis
        proof (cases "extract (\<lambda> (fa,rls). fa = (f,length ts)) rrm")
          case None
          let ?e = "((f,length ts), [(Fun f ts,r)])"
          let ?e' = "((f,length ts), {(Fun f ts,r)})"
          from None Cons(2) have rm: "rm = ?e # rrm" by (simp add: Fun Pair Some None)
          from extract_None[OF None] have rrm: "\<And> g n rl. ((g,n),rl) \<in> set rrm \<Longrightarrow> (f,length ts) \<noteq> (g,n)" by auto
          hence rrm': "\<And> g n rl. ((g,n),rl) \<in> ?setr rules \<Longrightarrow> (f,length ts) \<noteq> (g,n)" by (simp only: srrm[symmetric], auto)
          hence  id: " {(Fun f ts, r)} = {(l, ra). (l = Fun f ts \<and> ra = r \<or> (l, ra) \<in> set rules) \<and> root l = Some f \<and> num_args l = length ts}"  by force
          from rrm have dis: "?dis rm" 
            by (simp add: rm drrm, auto)
          have "?setl rm = insert ?e' (?setl rrm)" by (simp add: rm)
          also have "\<dots> = insert ?e' (?setr rules)" by (simp add: srrm)
          also have "\<dots> = ?setr ( (Fun f ts,r) # rules)" 
          proof (rule set_eqI, clarify)
            fix g n rls
            show "(((g,n),rls) \<in> insert ?e' (?setr rules)) = (((g,n),rls) \<in> ?setr ((Fun f ts,r) # rules))"
            proof (cases "(g,n) = (f,length ts)")
              case False
              hence "(((g,n),rls) \<in> insert ?e' (?setr rules)) = (((g,n),rls) \<in> ?setr rules)" by auto
              also have "\<dots> = (((g,n),rls) \<in> ?setr ((Fun f ts,r) # rules))" using False by auto
              finally show ?thesis .
            next
              case True note oTrue = this
              show ?thesis 
              proof (cases "rls = {(Fun f ts, r)}")
                case True
                with oTrue show ?thesis by (simp add: id, force)
              next
                case False
                show ?thesis using rrm'[of g n rls] True False by (simp add: False True id, auto)
              qed
            qed
          qed
          finally show ?thesis
            by (simp only: dis drrm, simp add: Pair Fun)
        next
          case (Some res)
          obtain bef fg rls aft where "res = (bef,(fg,rls),aft)" by (cases res, force)
          from extract_Some[OF Some[simplified this]] Some[simplified this] have rrm: "rrm = bef @ ((f,length ts), rls) # aft" 
            and e: "extract (\<lambda> (fa, rls). fa = (f,length ts)) rrm = Some (bef, ((f,length ts),rls),aft)" by auto
          let ?e = "((f,length ts), (Fun f ts,r) # rls)"
          let ?e' = "((f,length ts), insert (Fun f ts,r) (set rls))"
          have "((f,length ts),set rls) \<in> ?setl rrm" unfolding rrm by auto
          hence rls: "set rls = {(l, r) |l r. (l, r) \<in> set rules \<and> root l = Some f \<and> num_args l = length ts}" using Cons(1)[OF oSome] by auto
          obtain ba where ba: "ba = bef @ aft" by auto
          from Cons(2) e ba have rm: "rm = ?e # ba" by (simp add: Fun Pair oSome e)
          from drrm[simplified rrm] have dis: "?dis ba"  unfolding distinct_eq_append ba by auto
          from drrm[simplified rrm] have dis: "?dis rm" unfolding rm distinct_eq_append ba by (auto simp: dis[simplified ba])
          from drrm[simplified rrm distinct_eq_append] 
          have diff: "(\<forall>x\<in>set ba. \<not> (\<lambda>(g, rls). (f,length ts) = g) x)" by (auto simp: ba)
          have "?setl [((f, length ts),rls)] \<union> ?setl ba = ?setl rrm" using rrm ba by auto
          also have "\<dots> = ?setr rules" by (rule srrm)
          finally have id: "?setl [((f, length ts),rls)] \<union> ?setl ba = ?setr rules" .          
          have "?setl rm = insert ?e' (?setl ba)" by (simp add: rm)
          also have "\<dots> = ?setr ((Fun f ts,r) # rules)" 
          proof (rule set_eqI, clarify)
            fix g n rl 
            show "(((g,n),rl) \<in> insert ?e' (?setl ba)) = (((g,n),rl) \<in> ?setr ((Fun f ts,r) # rules))"
            proof (cases "(g,n) = (f,length ts)")
              case False 
              hence "(((g,n),rl) \<in> insert ?e' (?setl ba)) = (((g,n),rl) \<in> ?setl ba)" by auto
              also have "\<dots> = (((g,n),rl) \<in> ?setr rules)" using False by (simp only: id[symmetric], auto)
              also have "\<dots> = (((g,n),rl) \<in> ?setr ((Fun f ts,r) # rules))" using False by auto
              finally show ?thesis .
            next
              case True note oTrue = this
              show ?thesis 
              proof (cases "rl = insert (Fun f ts, r) (set rls)")
                case True
                hence "(((g,n),rl) \<in> insert ?e' (?setl ba)) = True" using oTrue by auto
                also have "\<dots> = (((g,n),rl) \<in> ?setr ((Fun f ts,r) # rules))" unfolding True rls using oTrue by force
                finally show ?thesis .
              next
                case False
                hence "(((g,n),rl) \<in> insert ?e' (?setl ba)) = False" using diff by (simp add: True, auto)
                also have "\<dots> = (((g,n),rl) \<in> ?setr ((Fun f ts,r) # rules))" 
                proof (rule ccontr)
                  assume "\<not> ?thesis"
                  with True have "((f,length ts),rl) \<in> ?setr ((Fun f ts,r) # rules)" by simp
                  hence "rl = {(l, ra) |l ra. (l, ra) \<in> set ((Fun f ts, r) # rules) \<and> root l = Some f \<and> num_args l = length ts}" by simp
                  with False rls show False by auto
                qed                              
                finally show ?thesis .
              qed
            qed
          qed
          also have "\<dots> = ?setr (rule # rules)" by (simp add: Pair Fun)
          finally show ?thesis by (simp add: dis)
        qed
      qed
    qed
  qed
qed

lemma computeRuleMapSound: assumes "computeRuleMap R = Some rm" 
shows "(set (rm (f,n))) = {(l,r) | l r. (l,r) \<in> set R \<and> root l = Some f \<and> num_args l = n}"
proof (cases "computeRuleMapH R")
  case None
  thus ?thesis using assms unfolding computeRuleMap_def by auto
next
  case (Some rrm)
  note rrm = computeRuleMapHSound[OF this] 
  note rm = assms[unfolded computeRuleMap_def, simplified Some, simplified, symmetric]
  show ?thesis
  proof (cases "lookup (f,n) rrm")
    case (Some rls)
    from lookup_sound[OF this] have "((f,n),set rls) \<in> (\<lambda> (f,rls). (f, set rls)) ` set rrm"
      by auto
    hence "set rls = {(l,r) | l r. (l,r) \<in> set R \<and> root l = Some f \<and> num_args l = n}"
      by (simp only: rrm, simp) 
    thus ?thesis by (simp add: rm Some)
  next
    case None
    have id: "{(l, r) |l r. (l, r) \<in> set R \<and> root l = Some f \<and> num_args l = n} = {}" (is "?set = {}")
    proof (rule ccontr)
      assume "\<not> ?thesis"
      then obtain l r where "(l, r) \<in> set R \<and> root l = Some f \<and> num_args l = n" by auto
      with rrm have "(((f,n), ?set)) \<in> (\<lambda> (f,rls). (f, set rls)) ` set rrm" by auto
      with lookup_complete[OF None] show False by auto
    qed
    thus ?thesis
      by (simp only: rm None id, auto)
  qed
qed
    
    
  
lemma computeRuleMap_left_vars:
  shows "(computeRuleMap R \<noteq> None) = (\<forall> lr \<in> set R. \<forall> x. fst lr \<noteq> Var x)"
proof (cases "computeRuleMap R")
  case None
  from None computeRuleMapSound2 have "\<exists> (l,r) \<in> set R. root l = None" by auto
  from this obtain l r where "(l,r) \<in> set R \<and> root l = None" by auto
  from this have "(l,r) \<in> set R \<and> \<not> (\<forall> x. fst (l,r) \<noteq> Var x)" by (cases l, auto)
  with None show ?thesis by blast
next
  case (Some rm)
  hence left: "computeRuleMap R \<noteq> None" by auto
  from Some computeRuleMapSound2 have "\<forall> (l,r) \<in> set R. root l \<noteq> None" by force
  hence "\<forall> lr \<in> set R. \<forall> x. fst lr \<noteq> Var x" by auto
  with left show ?thesis by blast
qed

lemma computeRuleMap_defs: fixes R :: "('f,'v)rules"
  assumes "computeRuleMap R = Some rm"
  shows "(rm (f,n) = []) = ((f,n) \<notin> defs (set R))"
proof -
  from assms computeRuleMapSound have rm: "\<And> (f :: 'f) n. set (rm (f,n)) = {(l,r) | l r. (l, r) \<in> set R \<and> root l = Some f \<and> num_args l = n}" by force
  show ?thesis
  proof (cases "rm (f,n)")
    case Nil
    with rm have "(f,n) \<notin> defs (set R)" unfolding defs_def defined_def by force
    with Nil show ?thesis by blast
  next
    case (Cons lr RR)
    hence left: "rm (f,n) \<noteq> []" by auto
    from Cons rm[where f = f and n = n] have "(f,n) \<in> defs (set R)" unfolding defs_def defined_def by (cases lr, force)
    with left show ?thesis by blast
  qed
qed


definition
  DP_list :: "('f \<Rightarrow> 'f) \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f,'v)rule_map \<Rightarrow> ('f, 'v) rules"
where
  "DP_list shp R RM \<equiv>
    let rm = the RM in        
    concat (map (\<lambda>lr.
      let l = fst lr; s = sharp_term shp l in
      map (\<lambda>u. (s, sharp_term shp u)) (filter (\<lambda>u. \<not> l \<rhd> u \<and> \<not> is_var u \<and> rm (the (root u), num_args u) \<noteq> []) (supteq_list (snd lr)))) R)"

lemma DP_list[simp]: assumes rm: "computeRuleMap R = Some rm"
  shows "set (DP_list shp R (computeRuleMap R)) = DP shp (set R)" (is "?L = ?R")
proof -
  from computeRuleMap_defs[OF rm] have rm_def: "\<And> f n. defined (set R) (f,n) = (rm (f,n) \<noteq> [])" unfolding defs_def by auto
  show ?thesis
  proof
    show "?L \<subseteq> ?R"
    proof
      fix s t assume "(s, t) \<in> ?L" 
      then obtain l r u where "(l, r) \<in> set R" and "r \<unrhd> u" and "\<not> l \<rhd> u"
        and "s = sharp_term shp l" and "t = sharp_term shp u"
        and "\<not> is_var u" and "defined (set R) (the (root u), num_args u)" unfolding rm_def unfolding DP_list_def Let_def by (auto simp: rm)
      thus "(s, t) \<in> DP shp (set R)" unfolding DP_def by (cases u) auto      
    qed
  next
    show "?R \<subseteq> ?L"
    proof
      fix s t assume "(s, t) \<in> ?R"
      then obtain lr u where "lr \<in> set R" and "snd lr \<unrhd> u" and "\<not> fst lr \<rhd> u"
        and "s = sharp_term shp (fst lr)" and "t = sharp_term shp u"
        and "\<not> is_var u" and "defined (set R) (the (root u), num_args u)" unfolding DP_def by auto
      thus "(s, t) \<in> ?L" unfolding DP_list_def Let_def rm_def by (auto simp: rm)
    qed
  qed
qed

definition
  check_no_defined_root :: "('f::show \<times> nat \<Rightarrow> ('f,'v)rules) \<Rightarrow> ('f, 'v::show) term \<Rightarrow> shows check"
where
  "check_no_defined_root rm t = check (rm (the (root t), num_args t) = []) (
     ''the root of '' +#+ shows t +@+ shows_string '' is defined''
   )"

lemma check_no_defined_root_sound:
  assumes "computeRuleMap R = Some rm" 
  shows "isOK (check_no_defined_root rm t) = (\<not> defined (set R) (the (root t), num_args t))"
using computeRuleMap_defs[OF assms] unfolding defs_def check_no_defined_root_def by auto 

lemma computeRuleMap_None_not_SN:
  assumes "computeRuleMap R = None"
  shows "\<not> SN_elt (rstep (set R)) t"
proof -
  from assms computeRuleMap_left_vars[of R] obtain x r where "(Var x,r) \<in> set R" by auto
  from left_var_imp_not_SN[OF this] show ?thesis .
qed



end
