(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Uncurry
imports
  Semantic_Labeling_Framework
  Trs_Impl
  Edg
  Name
  Regexp_Method
begin

function unapp :: "'f \<Rightarrow> ('f,'v)term \<Rightarrow> (('f,'v)term \<times> (('f,'v)term) list)"
where "unapp a (Var x) = (Var x,[])"
   |  "unapp a (Fun f ss) = (if f = a \<and> length ss = 2 then 
               (case unapp a (ss ! 0) of (r,ts) \<Rightarrow> (r,ts @ [ss ! 1])) else (Fun f ss, []))"
  by pat_completeness auto

termination
proof
  show "wf (measures [\<lambda> (x,y). size y])" by simp
next
  fix a :: 'a  and f :: 'a and ss :: "('a,'b)term list"
  assume "f = a \<and> length ss = 2"
  then obtain s ts where "ss = s # ts" by (cases ss, auto)
  thus  "((a, ss ! 0), a, Fun f ss) \<in> measures [\<lambda> (x,y). size y]" by auto
qed



fun apply_args :: "'f \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term list \<Rightarrow> ('f,'v)term"
where "apply_args a t [] = t"
    | "apply_args a t (s # ss) = apply_args a (Fun a [t,s]) ss"

lemma apply_args_last: "apply_args a t (ss @ [s]) = Fun a [apply_args a t ss, s]"
  by (induct ss arbitrary: t, auto)

lemma apply_args_append: "apply_args a t (ss @ ts) = apply_args a (apply_args a t ss) ts"
  by (induct ss arbitrary: t, auto)


lemma apply_args_subst: "(apply_args a t ts) \<cdot> \<sigma> = apply_args a (t \<cdot> \<sigma>) (map (\<lambda> s. s \<cdot> \<sigma>) ts)"
  by (induct ts rule: List.rev_induct, auto simp: apply_args_last)

lemma apply_args_rsteps: assumes t: "(t,s) \<in> (rstep R)^*"
  and len: "length ts = length ss"
  and ts: "\<And> i. i < length ts \<Longrightarrow> (ts ! i, ss ! i) \<in> (rstep R)^*"
  shows "(apply_args a t ts, apply_args a s ss) \<in> (rstep R)^*"
using len ts t
proof (induct ts arbitrary: ss t s)
  case Nil
  thus ?case by auto
next
  case (Cons v ts)
  from Cons(2) obtain w ws where ss: "ss = w # ws" and len: "length ts = length ws" by (cases ss, auto)
  note Cons = Cons[unfolded ss]
  from Cons(3)[of 0] have step: "(v,w) \<in> (rstep R)^*" by simp
  { 
    fix i
    assume "i < length ts"
    hence "(ts ! i, ws ! i) \<in> (rstep R)^*"
      using Cons(3)[of "Suc i"] by auto
  } note steps = this
  have step: "(Fun a [t,v], Fun a [s,w]) \<in> (rstep R)^*"
  proof (rule args_rsteps_imp_rsteps, simp)
    show "\<forall> i < length [t,v]. ([t,v] ! i, [s,w] ! i) \<in> (rstep R)^*"
    proof(simp, intro allI impI)
      fix i
      assume "i < Suc (Suc 0)"
      hence "i = 0 \<or> i = 1" by arith
      thus "([t,v] ! i, [s,w] ! i) \<in> (rstep R)^*"
        using Cons(4) step by force
    qed
  qed    
  show ?case
    by (unfold ss, simp, rule Cons(1)[OF len steps step])
qed


lemma apply_args_root_rstep: assumes "(t,s) \<in> rstep R"
  shows "(apply_args a t ts, apply_args a s ts) \<in> rstep R"
using assms
proof (induct ts arbitrary: t s)
  case Nil
  thus ?case by auto
next
  case (Cons v ts)
  from Cons(1)[OF rstep.ctxt[OF Cons(2), of "More a [] \<box> [v]"]]
  show ?case by simp
qed


lemma apply_args_root_rsteps: assumes "(t,s) \<in> (rstep R)^+"
  shows "(apply_args a t ts, apply_args a s ts) \<in> (rstep R)^+"
  using assms
proof (induct)
  case (base z)
  from apply_args_root_rstep[OF this]
  show ?case by auto
next
  case (step v us)
  from step(3) apply_args_root_rstep[OF step(2), of a ts]
  show ?case by auto
qed


lemma obtain_length2: "length xs = 2 \<Longrightarrow> (\<And> x y. xs = [x,y] \<Longrightarrow> P) \<Longrightarrow> P"
  by (cases xs, simp, cases "tl xs", simp, cases "tl (tl xs)", auto)

lemma obtain_length2_conj: "Q \<and> length xs = 2 \<Longrightarrow> (\<And> x y. xs = [x,y] \<Longrightarrow> P) \<Longrightarrow> P"
  using obtain_length2 by auto



lemma unapp_apply_app: assumes unapps: "unapp a s = (r,ts)"
  shows "unapp a (Fun a [s,t]) = (r,ts @ [t])"
  by (simp add: unapps)

lemma unapp_apply_args: assumes "unapp a t = (r,ss)"
  shows "t = apply_args a r ss"
using assms
proof (induct t arbitrary: ss)
  case (Var y)
  thus ?case by auto
next
  case (Fun f ts)
  show ?case
  proof (cases "f = a \<and> length ts = 2")
    case False
    thus ?thesis using Fun(2) by auto
  next
    case True
    obtain s t where ts: "ts = [s,t]" by (rule obtain_length2_conj[OF True]) 
    obtain rr sss where us: "unapp a s = (rr,sss)" by (cases "unapp a s", auto)
    from Fun(2) have "unapp a (Fun f ts) = (rr, sss @ [t])"      
      by (simp add: True ts us)
    with Fun(2) have r: "r = rr" and ss: "ss = (sss @ [t])" by auto
    show ?thesis 
      unfolding ts r ss Fun(1)[unfolded ts r, OF _ us, simplified] apply_args_last
      using True by simp
  qed
qed



lemma unapp_apply_non_app: assumes napp: "\<And> t1 t2. t \<noteq> Fun a [t1,t2]"
  shows "unapp a (apply_args a t ss) = (t, ss)"
proof (induct ss rule: List.rev_induct)
  case Nil with napp show ?case 
  proof (cases t, simp)
    case (Fun f ts)
    show ?thesis 
    proof (cases "f = a \<and> length ts = 2")
      case False with Fun show ?thesis by auto
    next
      case True
      from obtain_length2_conj[OF True] Fun napp show ?thesis by auto
    qed
  qed
next
  case (snoc s ss)
  show ?case using snoc
    by (unfold apply_args_last, simp)
qed


lemma unapp_root: "unapp a t = (r,ss) \<Longrightarrow> r \<noteq> Fun a [u,v]"
proof (induct t arbitrary: ss)
  case (Var x) thus ?case by auto
next
  case (Fun f ts)
  show ?case  
  proof (cases "f = a \<and> length ts = 2")
    case False
    thus ?thesis using Fun by auto
  next
    case True
    obtain t1 t2 where ts: "ts = [t1,t2]" by (rule obtain_length2_conj[OF True])
    obtain rr sss where unapp2: "unapp a t1 = (rr,sss)" by (cases "unapp a t1", auto)
    with True Fun ts have "unapp a (Fun f ts) = (rr,sss @ [t2])" by auto
    with Fun have r: "r = rr" and ss: "ss = sss @ [t2]" by auto
    from Fun(1)[unfolded ts r, OF _ unapp2]
    show ?thesis unfolding r by auto
  qed
qed

lemma unapp_root_fun: assumes "unapp a t = (Fun f ss,ts)"
  shows "\<not> (f = a \<and> length ss = 2)"
proof 
  assume ass: "f = a \<and> length ss = 2"
  obtain x y where "ss = [x,y]" by (rule obtain_length2_conj[OF ass])
  with unapp_root[OF assms] ass show False by auto
qed

lemma unapp_apply_var: "unapp a (apply_args a (Var x) ss) = (Var x,ss)"
  by (rule unapp_apply_non_app, simp)

lemma unapp_apply_fun: assumes nonapp: "\<not> (f = a \<and> length ts = 2)" 
  shows "unapp a (apply_args a (Fun f ts) ss) = (Fun f ts,ss)"
proof (rule unapp_apply_non_app)
  fix t1 t2
  show "Fun f ts \<noteq> Fun a [t1,t2]"
  proof
    assume "Fun f ts = Fun a [t1,t2]"
    hence "f = a \<and> length ts = 2" by simp
    with nonapp show False .. 
  qed
qed

    

lemma unapp_size_root: assumes "unapp a t = (r,ts)"
  shows "size r \<le> size t"
using assms
proof (induct t arbitrary: ts )
  case (Var x) thus ?case by simp
next
  case (Fun f ss)
  show ?case 
  proof (cases "f = a \<and> length ss = 2")
    case False
    with Fun show ?thesis by auto
  next
    case True
    obtain u v where ss: "ss = [u,v]" by (rule obtain_length2_conj[OF True])
    obtain rr sss where unapp: "unapp a u = (rr,sss)" by (cases "unapp a u", auto)
    from Fun(2) have "unapp a (Fun f ss) = (rr, sss @ [v])"
      by (simp add: True unapp ss)
    with Fun(2) have r: "r = rr" and ts: "ts = sss @ [v]" by auto
    from Fun(1)[unfolded r ss, OF _ unapp, simplified] have "size r \<le> size u" 
      unfolding r by simp
    thus ?thesis unfolding ss by simp
  qed
qed


lemma unapp_size_arg: assumes "unapp a t = (r,ts)" and "s \<in> set ts"
  shows "size s < size t"
using assms
proof (induct t arbitrary: ts )
  case (Var x) thus ?case by simp
next
  case (Fun f ss)
  show ?case 
  proof (cases "f = a \<and> length ss = 2")
    case False
    with Fun show ?thesis by auto
  next
    case True
    obtain u v where ss: "ss = [u,v]" by (rule obtain_length2_conj[OF True])
    obtain rr sss where unapp: "unapp a u = (rr,sss)" by (cases "unapp a u", auto)
    from Fun(2) have "unapp a (Fun f ss) = (rr, sss @ [v])"
      by (simp add: True unapp ss)
    with Fun(2) have r: "r = rr" and ts: "ts = sss @ [v]" by auto
    from Fun(1)[unfolded r ss, OF _ unapp, simplified] have rec: "s \<in> set sss \<Longrightarrow> size s < size u"
      by simp
    from Fun(3)[unfolded ts] have "s \<in> set sss \<or> s = v" by auto
    thus ?thesis
    proof
      assume "s \<in> set sss"
      from rec[OF this] show ?thesis unfolding ss by simp
    next
      assume "s = v" thus ?thesis unfolding ss by simp
    qed
  qed
qed

lemma unapp_size_intro1: assumes "(x, ys) = unapp a t" and "x = Fun f xs" and mem: "y \<in> set (xs @ ys)"
       shows "size y < size t"
proof - 
  from assms have unapp: "unapp a t = (Fun f xs, ys)" by auto
  from mem have "y \<in> set xs \<or> y \<in> set ys" by auto
  thus ?thesis
  proof
    assume "y \<in> set xs"
    hence "size y < size (Fun f xs)" by (auto simp: size_simps)
    with unapp_size_root[OF unapp] 
    show ?thesis by simp
  next
    assume "y \<in> set ys"
    from unapp_size_arg[OF unapp this]
    show ?thesis by simp
  qed
qed


declare unapp_size_intro1[intro]

fun hvf_term :: "'f \<Rightarrow> ('f,'v)term \<Rightarrow> bool"        
where "hvf_term a t = (case unapp a t of 
            (Var _, ts) \<Rightarrow> ts = []
          | (Fun f us, ts) \<Rightarrow> list_all (hvf_term a) (us @ ts))
      "


lemma applicative_term_induct[case_names Var Fun]:
  assumes var_IH: "\<And> x ts. (\<And> t. t \<in> set ts \<Longrightarrow> P t) \<Longrightarrow> P (apply_args a (Var x) ts)"
    and   fun_IH: "\<And> f ss ts. \<lbrakk>\<And> t. t \<in> set ts \<Longrightarrow> P t; \<And> s. s \<in> set ss \<Longrightarrow> P s; \<not> (f = a \<and> length ss = 2)\<rbrakk> \<Longrightarrow>
                P (apply_args a (Fun f ss) ts)"
  shows "P t"
proof (induct rule: wf_induct[of _ P, OF wf_measures])
  fix t
  assume "\<forall> s. (s,t) \<in> measures [size] \<longrightarrow> P s"
  hence ind: "\<And> s. size s < size t \<Longrightarrow> P s" by auto
  obtain r ts where unapp: "unapp a t = (r,ts)" by (cases "unapp a t", auto)
  from ind[OF unapp_size_arg[OF unapp]] have indt: "\<And> t. t \<in> set ts \<Longrightarrow> P t" by auto
  from unapp_apply_args[OF unapp] have t: "t = apply_args a r ts" .
  show "P t" unfolding t
  proof (cases r)
    case (Var x)
    show "P (apply_args a r ts)"
      unfolding Var
      by (rule var_IH, rule indt)
  next
    case (Fun f ss)
    note unapp = unapp[unfolded Fun]
    have napp: "\<not> (f = a \<and> length ss = 2)"
      by (rule unapp_root_fun[OF unapp])
    from unapp_size_root[OF unapp] have s1: "size (Fun f ss) \<le> size t" .
    {
      fix s
      assume "s \<in> set ss"
      hence "size s < size (Fun f ss)" by (auto simp: size_simps)
      from this s1 have "size s < size t" by simp
      note ind[OF this]
    } note inds = this
    from fun_IH[OF indt inds napp]
    show "P (apply_args a r ts)" unfolding Fun .
  qed
qed



lemma hvf_term_induct[consumes 1,case_names Var Fun]: assumes app: "hvf_term a t"
  and ind_var: "\<And> x. P (Var x)"
  and fun_IH: "\<And> f ss ts. ((\<And> t. t \<in> set ts \<Longrightarrow> P t) \<Longrightarrow> (\<And> s. s \<in> set ss \<Longrightarrow> P s) \<Longrightarrow> (\<And> t. t \<in> set ts \<Longrightarrow> hvf_term a t) \<Longrightarrow> 
               (\<And> s. s \<in> set ss \<Longrightarrow> hvf_term a s) \<Longrightarrow> \<not> (f = a \<and> length ss = 2) \<Longrightarrow>
                P (apply_args a (Fun f ss) ts))"
  shows "P t"
proof -
  have "hvf_term a t \<longrightarrow> P t" (is "?P t")
  proof (induct rule: wf_induct[of _ ?P, OF wf_measures])
    fix t
    assume "\<forall> s. (s,t) \<in> measures [size] \<longrightarrow> ?P s"
    hence ind: "\<And> s. \<lbrakk>hvf_term a s; size s < size t\<rbrakk> \<Longrightarrow> P s" by auto
    obtain r ts where unapp: "unapp a t = (r,ts)" by (cases "unapp a t", auto)
    from unapp_apply_args[OF unapp] have t: "t = apply_args a r ts" .
    show "?P t" unfolding t
    proof
      assume hvf: "hvf_term a (apply_args a r ts)"
      note unapp = unapp[unfolded t]
      show "P (apply_args a r ts)"
      proof (cases r)
        case (Var x)
        show "P (apply_args a r ts)"
        proof (cases ts)
          case Nil thus ?thesis unfolding Var using ind_var by auto
        next
          case (Cons t tts)
          from hvf[simplified, simplified unapp, unfolded Var Cons] 
          show ?thesis by auto
        qed
      next
        case (Fun f ss)
        note unapp = unapp[unfolded Fun]
        note hvf = hvf[unfolded Fun]
        note ind = ind[unfolded t Fun]
        from hvf have hvfs: "\<And> s. s \<in> set ss \<Longrightarrow> hvf_term a s" by (auto simp: unapp list_all_iff)
        from hvf have hvft: "\<And> t. t \<in> set ts \<Longrightarrow> hvf_term a t" by (auto simp: unapp list_all_iff)
        from ind[OF hvft unapp_size_arg[OF unapp]] have indt: "\<And> t. t \<in> set ts \<Longrightarrow> P t" . 
        {
          fix s
          assume s: "s \<in> set ss"
          hence "size s < size (Fun f ss)" by (auto simp: size_simps)
          with unapp_size_root[OF unapp]
          have "size s < size (apply_args a (Fun f ss) ts)" by simp
          note ind[OF hvfs[OF s] this]
        } note inds = this
        have napp: "\<not> (f = a \<and> length ss = 2)"
          by (rule unapp_root_fun[OF unapp])        
        show ?thesis unfolding Fun
          by (rule fun_IH[OF indt inds hvft hvfs napp])
      qed
    qed
  qed
  with app show ?thesis by simp
qed

(* store for each original symbol f/n a non-empty list of symbols
   [f0,f1,..,f_aa] *)
type_synonym 'f sig_map = "'f \<Rightarrow> nat \<Rightarrow> 'f list"

definition get_symbol :: "'f sig_map \<Rightarrow> 'f \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> 'f"
  where "get_symbol sm f n i \<equiv> sm f n ! i"

definition aarity :: "'f sig_map \<Rightarrow> 'f \<Rightarrow> nat \<Rightarrow> nat"
  where "aarity sm f n \<equiv> length (sm f n) - 1"

lemma unapp_size_intro2: assumes "(x, ys) = unapp a t" and "x = Fun f xs" and mem: "y \<in> set xs"
       shows "size y < size t"
  using assms unapp_size_intro1  by auto

lemma unapp_size_intro3: assumes "(x, ys) = unapp a t" and mem: "y \<in> set ys"
       shows "size y < size t"
  using unapp_size_arg[OF assms(1)[symmetric] assms(2)] .

declare unapp_size_intro2[intro]
declare unapp_size_intro3[intro]


fun uncurry_term :: "'f \<Rightarrow> 'f sig_map \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term"
where "uncurry_term a sm t = (case unapp a t of
           (Var x,ts) \<Rightarrow> apply_args a (Var x) (map (uncurry_term a sm) ts)
         | (Fun f ss,ts) \<Rightarrow> (
          let n = length ss;
              uss = map (uncurry_term a sm) ss;
              uts = map (uncurry_term a sm) ts;
              aa  = aarity sm f n;
              m   = min (length ts) aa;
              fm  = get_symbol sm f n m
            in apply_args a (Fun fm (uss @ take m uts)) (drop m uts)))"

declare uncurry_term.simps[simp del]

abbreviation uncurry_subst where
"uncurry_subst a sm \<sigma> \<equiv> (\<lambda>x. uncurry_term a sm (\<sigma> x))"

lemma uncurry_apply_args_fun: assumes "\<not> (f = a \<and> length ss = 2)"
  shows "uncurry_term a sm (apply_args a (Fun f ss) ts) = 
  apply_args a (Fun (get_symbol sm f (length ss) (min (length ts) (aarity sm f (length ss)))) (map (uncurry_term a sm) ss @ take (min (length ts) (aarity sm f (length ss))) (map (uncurry_term a sm) ts))) (drop (min (length ts) (aarity sm f (length ss))) (map (uncurry_term a sm) ts))"
  by (unfold uncurry_term.simps[where t =  "apply_args a (Fun f ss) ts"] unapp_apply_fun[OF assms], simp add: Let_def)

lemma uncurry_apply_args_var: 
  "uncurry_term a sm (apply_args a (Var x) ts) = 
  apply_args a (Var x) (map (uncurry_term a sm) ts)"
  by (unfold uncurry_term.simps[where t =  "apply_args a (Var x) ts"] unapp_apply_var, simp)

lemma uncurry_subst_hvf: "hvf_term a t \<Longrightarrow> 
  uncurry_term a sm (t \<cdot> \<sigma>) =  uncurry_term a sm t \<cdot> uncurry_subst a sm \<sigma>"
proof (induct t rule: hvf_term_induct)
  case (Var x)
  thus ?case by (simp add: uncurry_term.simps)
next
  case (Fun f ss ts)
  show ?case (is "?l = ?r")
  proof -
    from Fun(5) have nappsig: "\<not> (f = a \<and> length (map (\<lambda> s. s \<cdot> \<sigma>) ss) = 2)" by simp 
    from Fun(2) have indss:  "(map (\<lambda>x. uncurry_term a sm (x \<cdot> \<sigma>)) ss) = (map (\<lambda> x. uncurry_term a sm x \<cdot> uncurry_subst a sm \<sigma>) ss)"
      by auto
    from Fun(1) have indts:  "(map (\<lambda>x. uncurry_term a sm (x \<cdot> \<sigma>)) ts) = (map (\<lambda> x. uncurry_term a sm x \<cdot> uncurry_subst a sm \<sigma>) ts)"
      by auto
    show ?thesis
      by (unfold uncurry_apply_args_fun[OF Fun(5)] apply_args_subst, simp, unfold uncurry_apply_args_fun[OF nappsig], 
        simp add: o_def indss indts take_map drop_map)
  qed
qed


definition generate_var :: "nat \<Rightarrow> string"
where "generate_var i \<equiv> (CHR ''x'') # to_string i"

lemma inj_generate_var: "inj generate_var" 
  unfolding inj_on_def generate_var_def using show_nat_inj[unfolded inj_on_def] by auto

definition generate_f_xs :: "'f \<Rightarrow> nat \<Rightarrow> ('f,string)term"
where "generate_f_xs f n \<equiv> Fun f (map (\<lambda> i. Var (generate_var i)) [0 ..< n])"

definition uncurry_of_sig :: "'f \<Rightarrow> 'f sig_map \<Rightarrow> ('f,string)trs"
where "uncurry_of_sig a sm \<equiv> {(Fun a [generate_f_xs (get_symbol sm f n i) (n + i), Var (generate_var (n + i))],generate_f_xs (get_symbol sm f n (Suc i)) (n + Suc i)) | f n i. i < aarity sm f n}"

lemma apply_arg_uncurry_of_sig:
  assumes n: "n < aarity sm f m"
  and len: "length ts = m + n"
  shows "(Fun a [Fun (get_symbol sm f m n) ts, t], Fun (get_symbol sm f m (Suc n)) (ts @ [t])) \<in> CS (uncurry_of_sig a sm)" (is "(?l, ?r) \<in> CS ?U") 
proof -
  let ?ll = "Fun a [generate_f_xs (get_symbol sm f m n) (m + n), Var (generate_var (m + n))]"
  let ?rr = "generate_f_xs (get_symbol sm f m (Suc n)) (m + Suc n)"
  let ?sigma = "(\<lambda>x. (ts @ [t]) ! (the_inv generate_var) x)"
  show ?thesis
  proof (rule subst_closureI2)
    show "(?ll, ?rr) \<in> uncurry_of_sig a sm" unfolding uncurry_of_sig_def using n by blast
  next
    show "?l = ?ll \<cdot> ?sigma"
      unfolding  generate_f_xs_def 
      by (simp add: o_def the_inv_f_f[OF inj_generate_var], unfold len[symmetric], simp, rule nth_equalityI, auto simp: nth_append)
  next
    show "?r = ?rr \<cdot> ?sigma"
      unfolding  generate_f_xs_def 
      by  (simp add: o_def the_inv_f_f[OF inj_generate_var], unfold len[symmetric], simp, rule nth_equalityI, auto simp: nth_append)
  qed
qed

lemma apply_args_uncurry_of_sig: assumes xs: "length ts = m + n"
  and aarity: "n + length us \<le> aarity sm f m"
  shows "(apply_args a (Fun (get_symbol sm f m n) ts) us, Fun (get_symbol sm f m (n + length us)) (ts @ us)) 
  \<in> (rstep (uncurry_of_sig a sm))^*" (is "_ \<in> ?U^*")
using xs aarity
proof (induct us arbitrary: ts n)
  case Nil
  thus ?case by auto
next
  case (Cons u us)
  from Cons(2) Cons(3) have n: "n < aarity sm f m"  by auto
  have step: "(Fun a [Fun (get_symbol sm f m n) ts, u], Fun (get_symbol sm f m (Suc n)) (ts @ [u])) \<in> ?U" (is "(Fun a [?l,_], ?r) \<in> _")
    using apply_arg_uncurry_of_sig[OF n Cons(2)] unfolding rstep_CC_CS
    using subset_ctxt_closure by auto
  have step: "(apply_args a ?l (u # us), apply_args a ?r us) \<in> ?U"
    by (simp, rule apply_args_root_rstep[OF step])
  have id1: "ts @ u # us = (ts @ [u]) @ us" by auto
  have id2: "n + length (u # us) = Suc n + length us" by auto
  show ?case
    by (rule converse_rtrancl_into_rtrancl[OF step], unfold id1 id2, rule Cons(1), insert Cons(2) Cons(3), auto)
qed


lemma apply_args_uncurry: "(apply_args a (uncurry_term a sm t) (map (uncurry_term a sm) ts), uncurry_term a sm (apply_args a t ts)) \<in> (rstep (uncurry_of_sig a sm))^*" (is "(_,?u (apply_args a t ts)) \<in> ?U")
proof (induct t rule: applicative_term_induct[where a = a])
  case (Var x us)
  show ?case
    by (unfold apply_args_append[symmetric] uncurry_apply_args_var, simp)
next
  case (Fun f ss us)
  obtain u where u: "?u = u" by auto
  obtain aa where aa: "aarity sm f (length ss) = aa" by auto
  obtain m where m: "min (length us + length ts) aa = m" by auto
  obtain n where n: "min (length us) aa = n" by auto
  obtain xs where xs: "map u ss @ take n (map u us) = xs" by simp
  from m n have nm: "n \<le> m" and nu: "n - length us = 0" by auto
  hence "m = n + (m - n)" by auto
  with m obtain m where m: "min (length us + length ts) aa = n + m" by auto
  hence nmut: "n + m \<le> length (map u us @ map u ts)" by auto
  obtain ys where ys: "drop (n + m - length us) ts = ys" by auto
  hence ys': "drop (n + m - length us) (map u ts) = map u ys" by (auto simp: drop_map)
  obtain zs where zs: "take (n + m - length us) ts = zs" by auto
  have ts: "ts = zs @ ys" unfolding ys[symmetric] zs[symmetric] by force
  have id1: "map u ss @ take n (map u us) @ take m (drop n (map u us) @ map u ts) = 
    xs @ take m (drop n (map u us) @ map u ts)" unfolding xs[symmetric] by simp
  have "map u ts = map u zs @ map u ys" unfolding ts
    by simp
  also have "\<dots> = map u (take (n + m - length us) ts) @ map u ys" unfolding zs ..
  finally have id2: "map u ts = map u (take (n + m - length us) ts) @ map u ys" by simp
  obtain g where g: "get_symbol sm f (length ss) = g" by auto
  show ?case
  proof (unfold apply_args_append[symmetric] uncurry_apply_args_fun[OF Fun(3)], simp add: u aa m n g take_map[symmetric] drop_append[symmetric]
      del: take_append, unfold take_add[OF nmut] take_append[of n], simp add: xs nu id1 ys' del: take_append map_append,
      unfold arg_cong[OF id2, of "\<lambda> t. apply_args a (Fun (g n) xs) (drop n (map u us) @ t)"],
      simp add: take_map[symmetric] del: take_append
    )
    show "(apply_args a (Fun (g n) xs)
      (drop n (map u us) @ take (n + m - length us) (map u ts) @ map u ys),
     apply_args a
      (Fun (g (n + m)) (xs @ take m (drop n (map u us) @ map u ts)))
      (drop (n + m) (map u us) @ map u ys))
    \<in> ?U" (is "(?l,?r) \<in> _")
    proof (cases "n = length us")
      case False
      with n have n: "n = aa" by simp
      with m have m: "m = 0" by auto
      from m nu have "?l = apply_args a (Fun (g n) xs)
         (drop n (map u us) @ map u ys)" by simp
      also have "\<dots> = ?r" using m by simp
      finally show ?thesis by simp
    next
      case True
      from True have l: "?l = apply_args a (apply_args a (Fun (g n) xs) (take m (map u ts))) (map u ys)" by (auto simp: apply_args_append)
      from True have r: "?r = apply_args a (Fun (g (n+m)) (xs @ take m (map u ts))) (map u ys)" by simp
      show ?thesis unfolding l r 
      proof (rule apply_args_rsteps)
        have len1: "length xs = length ss + n" unfolding xs[symmetric] using True by simp
        from m  have "m \<le> length ts" and "n + m \<le> aa"  unfolding True[symmetric] by auto
        hence len2: "n + length (take m (map u ts)) \<le> aarity sm f (length ss)" 
          and m: "n + m = n + length (take m (map u ts))" unfolding aa by auto
        show "(apply_args a (Fun (g n) xs) (take m (map u ts)), Fun (g (n+m)) (xs @ take m (map u ts))) \<in> ?U"
          unfolding g[symmetric] m
          by (rule apply_args_uncurry_of_sig[OF len1 len2])
      qed auto
    qed
  qed
qed
    


lemma uncurry_subst:
  fixes \<sigma> :: "('f, string) subst"
  shows "(uncurry_term a sm t \<cdot> uncurry_subst a sm \<sigma>, uncurry_term a sm (t \<cdot> \<sigma>)) \<in> (rstep (uncurry_of_sig a sm))^*" (is "(?u t \<cdot> ?us,_) \<in> ?U")
proof -
  let ?m = "map (\<lambda>x. ?u x \<cdot> ?us)"
  let ?n = "map (\<lambda>x. ?u (x \<cdot> \<sigma>))"
  let ?k = "\<lambda> ts. map ?u (map (\<lambda> x. x \<cdot> \<sigma>) ts)"
  show ?thesis
  proof (induct t rule: applicative_term_induct[where a = a])
    case (Fun f ss ts)
    from Fun(3) have napp: "\<not> (f = a \<and> length (map (\<lambda>t. t \<cdot> \<sigma>) ss) = 2)" by simp
    obtain g where g: " (get_symbol sm f (length ss)
      (min (length ts) (aarity sm f (length ss)))) = g" by auto
    obtain m where m: " (min (length ts) (aarity sm f (length ss))) = m" by auto
    show ?case
    proof (simp add: apply_args_subst, unfold uncurry_apply_args_fun[OF Fun(3)] uncurry_apply_args_fun[OF napp], simp add: o_def g, simp add: m,
        simp add: apply_args_subst, simp add: o_def, rule apply_args_rsteps, rule args_rsteps_imp_rsteps, simp_all, simp add: take_map o_def,
        intro allI impI)
      fix i
      assume "i < length ts - m"
      thus "(?u (ts ! (m+i)) \<cdot> ?us, ?u (ts ! (m+i) \<cdot> \<sigma>)) \<in> ?U" using Fun(1) by auto
    next
      fix i
      assume i: "i < length ss + min (length ts) m"
      show "((?m ss @ ?m (take m ts)) ! i, (?n ss @ ?n (take m ts)) ! i) \<in> ?U"
      proof (cases "i < length ss")
        case True
        thus ?thesis using Fun(2) by (auto simp: nth_append)
      next
        case False
        hence "i - length ss < min (length ts) m" using i by auto
        hence "i - length ss < length (take m ts)" by auto
        thus ?thesis using False Fun(1) by (auto simp: nth_append)
      qed
    qed
  next
    case (Var x ts)
    show ?case 
    proof (unfold apply_args_subst uncurry_apply_args_var, simp add: o_def, rule rtrancl_trans, rule apply_args_rsteps[OF rtrancl_refl])
      show "length (?m ts) = length (?k ts)" by simp
    next
      fix i
      assume "i < length (?m ts)"
      with Var show "(?m ts ! i, ?k ts ! i) \<in> ?U" by auto
    qed (rule apply_args_uncurry)
  qed
qed



definition uncurry_trs :: "'f \<Rightarrow> 'f sig_map \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs"
where "uncurry_trs a sm R \<equiv> (\<lambda>(l,r). (uncurry_term a sm l, uncurry_term a sm r)) ` R"



fun unapp_ctxt :: "'f \<Rightarrow> ('f,'v)ctxt \<Rightarrow> (('f,'v)ctxt \<times> (('f,'v)term)list)"
where "unapp_ctxt a Hole = (Hole,[])"
   |  "unapp_ctxt a (More f bef C aft) = (if f = a \<and> bef = [] \<and> length aft = 1 then 
               (case unapp_ctxt a C of (r,ts) \<Rightarrow> (r,ts @ aft)) else (More f bef C aft, []))"


fun apply_args_ctxt :: "'f \<Rightarrow> ('f,'v)ctxt \<Rightarrow> ('f,'v)term list \<Rightarrow> ('f,'v)ctxt"
where "apply_args_ctxt a C [] = C"
    | "apply_args_ctxt a C (s # ss) = apply_args_ctxt a (More a [] C [s]) ss"

lemma apply_args_ctxt_last: "apply_args_ctxt a C (ss @ [s]) = More a [] (apply_args_ctxt a C ss) [s]"
  by (induct ss arbitrary: C, auto)

lemma unapp_apply_args_ctxt: assumes "unapp_ctxt a C = (D,ss)"
  shows "C = apply_args_ctxt a D ss"
using assms
proof (induct C arbitrary: ss)
  case Hole
  thus ?case by auto
next
  case (More f bef E aft)
  show ?case
  proof (cases "f = a \<and> bef = [] \<and> length aft = 1")
    case False
    thus ?thesis using More(2) by auto
  next
    case True
    then obtain t where aft: "aft = [t]" by (cases aft, auto)
    obtain DD sss where us: "unapp_ctxt a E = (DD,sss)" by (cases "unapp_ctxt a E", auto)
    from More(2) have "unapp_ctxt a (More f bef E aft) = (DD, sss @ [t])"      
      by (simp add: True aft us)
    with More(2) have D: "D = DD" and ss: "ss = (sss @ [t])" by auto
    show ?thesis 
      unfolding aft D ss  More(1)[unfolded aft True D, OF us] apply_args_ctxt_last
      using True by simp
  qed
qed


lemma unapp_ctxt_root: "unapp_ctxt a C = (D,ss) \<Longrightarrow> D \<noteq> More a [] E [v]"
proof (induct C arbitrary: ss)
  case Hole thus ?case by auto
next
  case (More f bef F aft)
  show ?case  
  proof (cases "f = a \<and> bef = [] \<and> length aft = 1")
    case False
    thus ?thesis using More by auto
  next
    case True
    then obtain t where aft: "aft = [t]" by (cases aft, auto)
    obtain G sss where unapp2: "unapp_ctxt a F = (G,sss)" by (cases "unapp_ctxt a F", auto)
    with True More aft have "unapp_ctxt a (More f bef F aft) = (G,sss @ [t])" by auto
    with More have D: "D = G" and ss: "ss = sss @ [t]" by auto
    from More(1)[unfolded aft D, OF unapp2]
    show ?thesis unfolding D by auto
  qed
qed


lemma unapp_ctxt_size_root: assumes "unapp_ctxt a C = (D,ts)"
  shows "size D \<le> size C"
using assms
proof (induct C arbitrary: ts )
  case Hole thus ?case by simp
next
  case (More f bef C aft)
  show ?case 
  proof (cases "f = a \<and> bef = [] \<and> length aft = 1")
    case False
    with More show ?thesis by auto
  next
    case True
    then obtain u where aft: "aft = [u]" by (cases aft, auto)
    obtain DD sss where unapp: "unapp_ctxt a C = (DD,sss)" by (cases "unapp_ctxt a C", auto)
    from More(2) have "unapp_ctxt a (More f bef C aft) = (DD, sss @ [u])"
      by (simp add: True unapp aft)
    with More(2) have D: "D = DD" and ts: "ts = sss @ [u]" by auto
    from More(1)[unfolded D aft, OF unapp, simplified] 
    show ?thesis unfolding D by auto 
  qed
qed

lemma apply_args_ctxt_insert: "(apply_args_ctxt a C ss)\<langle>t\<rangle> = apply_args a (C\<langle>t\<rangle>) ss"
  by (induct ss arbitrary: C, auto)



definition aarity_term :: "'f \<Rightarrow> 'f sig_map \<Rightarrow> ('f,'v)term \<Rightarrow> nat option"
where "aarity_term a sm t \<equiv> case (unapp a t) of (Fun f ss,ts) \<Rightarrow> Some (aarity sm f (length ss) - (length ts)) | (Var _,_) \<Rightarrow> None"

lemma aarity_term_subst: assumes "aarity_term a sm t = Some i"
  shows "aarity_term a sm (t \<cdot> \<sigma>) = Some i"
proof -
  obtain r ts where unapp: "unapp a t = (r,ts)" by (cases "unapp a t", auto)
  show ?thesis 
  proof (cases r)
    case (Var x)
    with assms unapp show ?thesis unfolding aarity_term_def by auto
  next
    case (Fun f ss)
    note unapp = unapp[unfolded Fun]
    from unapp_root_fun[OF unapp]
    have napp: "\<And> g. \<not> (f = a \<and> length (map g ss) = 2)" by auto          
    show ?thesis unfolding aarity_term_def unapp_apply_args[OF unapp] apply_args_subst 
      by (simp add: unapp_apply_fun[OF napp], insert unapp assms, unfold aarity_term_def, simp)
  qed
qed
    
  
lemma uncurry_aarity: assumes "aarity_term a sm t \<in> {Some 0, None}"
  shows "uncurry_term a sm (apply_args a t ts) = apply_args a (uncurry_term a sm t) (map (uncurry_term a sm) ts)"
using assms
proof (induct t rule: applicative_term_induct[where a = a])
  case (Var x ss)
  show ?case unfolding uncurry_apply_args_var apply_args_append[symmetric] by auto
next
  case (Fun f ss us)
  obtain aa where aa: "aarity sm f (length ss) = aa" by auto
  from unapp_apply_fun[OF Fun(3), of us] Fun(4)[unfolded aarity_term_def] 
  have "aa - length us = 0" by (auto simp: aa)
  hence id1: "min (length us + length ts) aa = min (length us) aa" by auto 
  show ?case unfolding apply_args_append[symmetric] uncurry_apply_args_fun[OF Fun(3)]
    by (simp add: aa id1)
qed

lemma size_apply_args_ctxt: "size (apply_args_ctxt a t ss) = size t +  length ss"
  by (induct ss arbitrary: t, auto)

definition eta_closed :: "'f \<Rightarrow> 'f sig_map \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> bool"
where "eta_closed a sm R R' \<equiv> \<forall> l r aa. (l,r) \<in> R \<longrightarrow> aarity_term a sm l = Some aa \<longrightarrow> aa > 0 \<longrightarrow> (\<exists> ll rr y. instance_rule (l,r) (ll,rr) \<and> y \<notin> vars_rule (ll,rr) \<and> (Fun a [ll,Var y], Fun a [rr,Var y]) \<in> R')"

fun apply_args_as_ctxt :: "'f \<Rightarrow> ('f,'v)term list \<Rightarrow> ('f,'v)ctxt"
  where "apply_args_as_ctxt a [] = \<box>"
      | "apply_args_as_ctxt a (s # ss) = apply_args_as_ctxt a ss \<circ>c (More a [] \<box> [s])"

lemma apply_args_as_ctxt: "apply_args a t ss = (apply_args_as_ctxt a ss)\<langle>t\<rangle>"
  by (induct ss arbitrary: t, auto)

declare apply_args_as_ctxt.simps[simp del]
lemma apply_args_root_rel: assumes ctxt: "one_step_ctxt_closed R"
  and step: "(t,s) \<in> R"
  shows "(apply_args a t ts, apply_args a s ts) \<in> R"
  unfolding apply_args_as_ctxt 
  using one_step_ctxt_closed_imp_ctxt_closed[OF ctxt]
  unfolding ctxt_closed_def using step by auto


lemma apply_args_root_trancl: assumes ctxt: "one_step_ctxt_closed R"
  and steps: "(t,s) \<in> R^+"
  shows "(apply_args a t ts, apply_args a s ts) \<in>  R^+"
  by (rule apply_args_root_rel[OF trancl_one_step_ctxt_closed[OF ctxt] steps])

lemma apply_args_arg_rel: assumes ctxt: "one_step_ctxt_closed R"
  and steps: "(u,v) \<in> R"
  shows "(apply_args a t (bef @ u # aft), apply_args a t (bef @ v # aft)) \<in> R"
proof -
  show ?thesis
  proof (induct bef arbitrary: t)
    case Nil
    show ?case by (simp, 
      rule apply_args_root_rel[OF ctxt], insert steps ctxt[unfolded one_step_ctxt_closed_def, THEN spec[of _ a], THEN spec[of _ "[t]"]], auto)
  next
    case (Cons b bef t)
    show ?case
      by (simp, rule Cons)
  qed
qed

lemma apply_args_arg_rsteps: assumes steps: "(u,v) \<in> (rstep R)^+"
  shows "(apply_args a t (bef @ u # aft), apply_args a t (bef @ v # aft)) \<in> (rstep R)^+"
  by (rule apply_args_arg_rel[OF trancl_one_step_ctxt_closed[OF one_step_ctxt_closed_rstep] steps])

lemma uncurry_step: assumes hvfR: "\<And> l r. (l,r) \<in> R \<Longrightarrow> hvf_term a l"
  and eta: "eta_closed a sm R R"
  and trs: "\<And> x r. (Var x,r) \<notin> R" (* new condition in comparison to paper.
          E.g. x \<rightarrow> f x, then f x \<rightarrow> f f x, but f(x) \<rightarrow> f(f) x" does not hold for U(R) = x \<rightarrow> f(x) *)
  and step: "(s,t) \<in> rstep R"
  shows "(uncurry_term a sm s, uncurry_term a sm t) \<in> rel_rstep (uncurry_trs a sm R, uncurry_of_sig a sm)" (is "_ \<in> rel_rstep (?R,?S)")
proof -
  let ?u = "uncurry_term a sm"
  from step show ?thesis unfolding rel_rstep.simps rel_step.simps
  proof (induct)
    case (IH C \<sigma> l r)
    from IH
    show ?case (is "_ \<in> ?U")
    proof (induct C arbitrary: l r \<sigma> rule: wf_induct[OF wf_measures[of "[size]"]])
      case (1 C)
      have ctxtU: "one_step_ctxt_closed ?U"
        using comp_one_step_ctxt_closed one_step_ctxt_closed_rstep rtrancl_one_step_ctxt_closed by blast
      let ?us = "uncurry_subst a sm \<sigma>"
      let ?tCs = "\<lambda> t C s. C\<langle>t \<cdot> s\<rangle>"
      let ?tCsig = "\<lambda> t C. C\<langle>t \<cdot> \<sigma>\<rangle>"
      let ?lC = "?tCsig l"
      let ?rC = "?tCsig r"
      note lr = 1(2)
      from lr hvfR have hvfl: "hvf_term a l"by auto
      from 1(1) have ind_gen: "\<And> l r D s. (l,r) \<in> R \<Longrightarrow>  size D < size C \<Longrightarrow> (?u (?tCs l D s), ?u (?tCs r D s)) \<in> ?U" by auto
      note ind = ind_gen[OF lr, of _ \<sigma>]
      obtain D ss where unapp: "unapp_ctxt a C = (D,ss)" by (cases "unapp_ctxt a C", auto)
      from unapp_ctxt_root[OF unapp] have napp: "\<And> E t. D \<noteq> More a [] E [t]" .
      from unapp_apply_args_ctxt[OF unapp] have C: "C = apply_args_ctxt a D ss" (is "_ = ?C") by auto
      from unapp_ctxt_size_root[OF unapp] have size: "size D \<le> size C" by auto
      show ?case 
      proof (cases D)
        case (More f bef E aft)
        obtain u v where u: "?tCsig l E = u" and v: "?tCsig r E = v" by auto
        show ?thesis unfolding C More apply_args_ctxt_insert
        proof (simp add: u v)
          from size More have "size E < size C" by auto
          note ind = ind[OF this]
          note steps = ind[unfolded u v]
          show "(?u (apply_args a (Fun f (bef @ u # aft)) ss),
                 ?u (apply_args a (Fun f (bef @ v # aft)) ss)) \<in> ?U"
          proof (cases "f = a \<and> length bef + length aft = 1")
            case False
            hence napp: "\<And> x. \<not> (f = a \<and> length (bef @ x # aft) = 2)" by auto
            obtain g where g: "(get_symbol sm f (Suc (length bef + length aft))
             (min (length ss) (aarity sm f (Suc (length bef + length aft))))) = g" by auto
            obtain bbef where bbef: "map ?u bef = bbef" by auto
            obtain aaft where aaft: "map (uncurry_term a sm) aft @ take (min (length ss) (aarity sm f (Suc (length bef + length aft))))
                 (map (uncurry_term a sm) ss) = aaft" by auto
            show ?thesis 
              unfolding uncurry_apply_args_fun[OF napp] 
              by (simp, rule apply_args_root_rel[OF ctxtU], rule one_step_ctxt_closed_one[OF ctxtU steps])
          next
            case True
            show ?thesis
            proof (cases bef)
              case (Cons b bbef)
              with True have f: "f = a" and bef: "bef = [b]" and aft: "aft = []" by auto
              show ?thesis unfolding f bef aft
              proof (simp)
                show "(?u (apply_args a (Fun a [b,u]) ss), ?u (apply_args a (Fun a [b,v]) ss)) \<in> ?U"
                proof (cases "aarity_term a sm b \<in> {Some 0, None}")
                  case True
                  note ua = uncurry_aarity[OF True]                  
                  show ?thesis unfolding apply_args.simps(2)[symmetric]
                    unfolding ua[of "u # ss"]  ua[of "v # ss"]
                    using  apply_args_arg_rel[OF ctxtU steps, of a "?u b" "[]" "map ?u ss"] by auto
                next
                  case False
                  obtain r ts where unapp: "unapp a b = (r,ts)" by (cases "unapp a b", auto)              
                  from unapp_apply_args[OF unapp] have b: "b = apply_args a r ts" by auto
                  from False obtain g us where Fun: "r = Fun g us" unfolding b by (cases r, auto simp: aarity_term_def unapp_apply_var)
                  obtain h where h: "h = get_symbol sm g (length us) (min (Suc (length ts + length ss)) (aarity sm g (length us)))" by auto
                  obtain n where n: "n = min (length (ts @ t # ss)) (aarity sm g (length us))" by auto
                  obtain uus where uus: "uus = map (uncurry_term a sm) us" by auto
                  obtain uts where uts: "uts = map (uncurry_term a sm) ts" by auto
                  obtain uss where uss: "uss = map (uncurry_term a sm) ss" by auto
                  let ?t = "\<lambda> t. apply_args a (Fun h (uus @ take n  (uts @ ?u t # uss)))  (drop n (uts @ ?u t # uss))" 
                  have napp: "\<not> (g = a \<and> length us = 2)"
                  proof
                    assume ass: "g = a \<and> length us = 2"
                    from obtain_length2_conj[OF this] obtain y z where "us = [y,z]" by blast
                    with unapp_root[OF unapp, unfolded Fun] ass show False by auto
                  qed                  
                  {
                    fix t
                    have "?u (apply_args a (Fun a [b,t]) ss) = ?u (apply_args a (Fun a [apply_args a (Fun g us) ts,t]) ss)"
                      unfolding f bef aft b Fun by simp
                    also have "\<dots> = ?u (apply_args a (Fun g us) (ts @ t # ss))" unfolding apply_args_last[symmetric] apply_args_append[symmetric]
                      by simp
                    also have "\<dots> = ?t t"
                      unfolding uncurry_apply_args_fun[OF napp] by (simp add: h n uus uts uss)
                    finally have "?u (apply_args a (Fun a [b,t]) ss) = ?t t" .
                  } note id = this
                  show ?thesis 
                    unfolding id
                  proof (cases "n \<le> length uts")
                    case False
                    hence n: "n - length uts > 0 \<and> n = length uts + (n - length uts)" by auto
                    then obtain k where n: "n = length uts + Suc k" by (cases "n - length uts", auto)
                    {
                      fix t
                      have "?t t = apply_args a (Fun h (uus @ uts @ uncurry_term a sm t # take k uss)) (drop k uss)" unfolding n 
                        by (simp)
                    } note id = this
                    show "(?t u, ?t v) \<in> ?U" unfolding id
                      by (rule apply_args_root_rel[OF ctxtU], insert one_step_ctxt_closed_one[OF ctxtU steps, of h "uus @ uts"], auto) 
                  next
                    case True
                    {
                      fix t
                      have "?t t = apply_args a (Fun h (uus @ take n uts)) (drop n uts @ uncurry_term a sm t # uss)" using True by simp
                    } note id = this
                    show "(?t u, ?t v) \<in> ?U" unfolding id                    
                      by (rule apply_args_arg_rel[OF ctxtU steps])
                  qed
                qed
              qed
            next
              case Nil             
              with True obtain t where f: "f = a" and aft: "aft = [t]" by (cases aft, auto)
              thus ?thesis using napp[unfolded More] Nil by auto
            qed
          qed
        qed
      next
        case Hole
        thus ?thesis unfolding C Hole apply_args_ctxt_insert
        proof (simp)
          show "(?u (apply_args a (l \<cdot> \<sigma>) ss), ?u (apply_args a (r \<cdot> \<sigma>) ss)) \<in> ?U"
          proof (cases ss)
            case Nil
            have "(?u (l \<cdot> \<sigma>), ?u r \<cdot> ?us) \<in> rstep ?R"
            proof
              show "(?u l, ?u r) \<in> ?R" unfolding uncurry_trs_def using lr by auto
            next
              show "?u (l \<cdot> \<sigma>) = \<box>\<langle>?u l \<cdot> ?us\<rangle>" by (simp add: uncurry_subst_hvf[OF hvfl])
            next
              show "?u r \<cdot> ?us = \<box>\<langle>?u r \<cdot> ?us\<rangle>" by simp
            qed 
            moreover have "(?u r \<cdot> ?us, ?u (r \<cdot> \<sigma>)) \<in> (rstep ?S)^*"
              by (rule set_mp[OF rtrancl_mono[OF rstep_mono] uncurry_subst]) simp
            ultimately show ?thesis unfolding Nil by auto
          next
            case (Cons s sss)
            have rsig: "(apply_args a (?u r \<cdot> ?us) (map ?u ss), ?u (apply_args a (r \<cdot> \<sigma>) ss)) \<in> (rstep ?S)^*"
            proof (rule set_mp[OF rtrancl_mono[OF rstep_mono[of "uncurry_of_sig a sm"]]], simp, rule rtrancl_trans)
              show "(apply_args a (?u r \<cdot> ?us) (map ?u ss), apply_args a (?u (r \<cdot> \<sigma>)) (map ?u ss)) \<in> (rstep (uncurry_of_sig a sm))^*"
                by (rule apply_args_rsteps[OF uncurry_subst], auto)
            next
              show "(apply_args a (?u (r \<cdot> \<sigma>)) (map ?u ss), ?u (apply_args a (r \<cdot> \<sigma>) ss)) \<in> (rstep (uncurry_of_sig a sm))^*"
                by (rule apply_args_uncurry)
            qed            
            show ?thesis
            proof (cases "aarity_term a sm (l \<cdot> \<sigma>) \<in> {Some 0, None}")
              case True
              have "?u (apply_args a (l \<cdot> \<sigma>) ss) = apply_args a (?u (l \<cdot> \<sigma>)) (map ?u ss)"
                by (rule uncurry_aarity[OF True])
              also have "\<dots> = apply_args a (?u l \<cdot> ?us) (map ?u ss)"
                unfolding uncurry_subst_hvf[OF hvfl] ..
              finally have lsig: "?u (apply_args a (l \<cdot> \<sigma>) ss) = apply_args a (?u l \<cdot> ?us) (map ?u ss)" .
              have lrsig: "(apply_args a (?u l \<cdot> ?us) (map ?u ss), apply_args a (?u r \<cdot> ?us) (map ?u ss)) \<in> rstep ?R"
                by (rule apply_args_root_rel[OF one_step_ctxt_closed_rstep], rule rstepI[where C = \<box>], unfold uncurry_trs_def, insert lr, auto)
              show ?thesis 
                unfolding lsig using lrsig rsig by auto
            next
              case False
              show ?thesis 
              proof (cases "aarity_term a sm l")
                case (Some aa)
                from aarity_term_subst[OF Some, of "\<sigma>"] False
                have "aa > 0" by auto
                from eta[unfolded eta_closed_def] lr Some this obtain ll rr y where inst: "instance_rule (l,r) (ll,rr)" and newRule: "(Fun a [ll,Var y], Fun a [rr,Var y]) \<in> R" and y: "y \<notin> vars_rule (ll,rr)" by blast+
                from inst[unfolded instance_rule_def] obtain \<delta> where l: "l = ll \<cdot> \<delta>" and r: "r = rr \<cdot> \<delta>" by auto
                let ?C = "apply_args_ctxt a \<box> sss" 
                let ?sig = "subst_extend (\<delta> \<circ>s \<sigma>) (zip [y] [s])"
                have "size ?C < size C" unfolding C Hole Cons by (simp add: size_apply_args_ctxt)
                note ind = ind_gen[OF newRule this, of ?sig]
                {
                  fix t :: "('a,string)term"
                  assume y: "y \<notin> vars_term t"
                  hence "t \<cdot> ?sig = t \<cdot> (\<delta> \<circ>s \<sigma>)" 
                    using subst_extend_id[of "UNIV - {y}" "[y]" t "\<delta> \<circ>s \<sigma>" "[s]"] by auto
                  hence "?C\<langle>Fun a [t, Var y] \<cdot> ?sig\<rangle> = apply_args a (t \<cdot> (\<delta> \<circ>s \<sigma>)) ss"
                    unfolding Cons
                    by (simp add: apply_args_ctxt_insert)
                } note id = this
                from y have ly: "y \<notin> vars_term ll" and ry: "y \<notin> vars_term rr" unfolding vars_rule_def by auto
                from ind[unfolded id[OF ly] id[OF ry]]
                show ?thesis unfolding l r by auto
              next
                case None
                then obtain x vs where unappl: "unapp a l = (Var x,vs)" 
                  unfolding aarity_term_def by (cases "unapp a l", cases "fst (unapp a l)", auto)
                with hvfl have l: "l = Var x"
                proof (induct l rule: hvf_term_induct)
                  case (Fun g xs ys)
                  from Fun(6) unapp_apply_fun[OF Fun(5)] show ?case by auto
                qed simp
                show ?thesis using trs lr[unfolded l]  by blast
              qed
            qed
          qed
        qed
      qed
    qed
  qed
qed

lemma uncurry_step_trancl: assumes hvfR: "\<And> l r. (l,r) \<in> R \<Longrightarrow> hvf_term a l"
  and eta: "eta_closed a sm R R"
  and trs: "\<And> x r. (Var x,r) \<notin> R" (* new condition in comparison to paper.
          E.g. x \<rightarrow> f x, then f x \<rightarrow> f f x, but f(x) \<rightarrow> f(f) x" does not hold for U(R) = x \<rightarrow> f(x) *)
  and step: "(s,t) \<in> rstep R"
  shows "(uncurry_term a sm s, uncurry_term a sm t) \<in> (rstep (uncurry_trs a sm R \<union> uncurry_of_sig a sm))^+" (is "_ \<in> (rstep (?R \<union> ?S))^+")
proof -
  from uncurry_step[OF hvfR eta trs step]
  have step: "(uncurry_term a sm s, uncurry_term a sm t) \<in> rel_rstep (?R, ?S)" .
  show ?thesis
    by (rule set_mp[OF _ step], unfold rel_rstep.simps rel_step.simps rstep_union, regexp)
qed


lemma uncurry_steps: assumes hvfR: "\<And> l r. (l,r) \<in> R \<Longrightarrow> hvf_term a l"
  and eta: "eta_closed a sm R R"
  and trs: "\<And> x r. (Var x,r) \<notin> R" 
  and steps: "(s,t) \<in> (rstep R)^*"
  shows "(uncurry_term a sm s, uncurry_term a sm t) \<in> (rstep (uncurry_trs a sm R \<union> uncurry_of_sig a sm))^*" (is "_ \<in> (rstep ?R)^*")
  using steps
proof (induct)
  case (step t u)
  let ?u = "uncurry_term a sm"
  have "(?u t, ?u u) \<in> (rstep ?R)^*"
    by (rule set_mp[OF _ uncurry_step_trancl[OF hvfR eta trs step(2)]], regexp)
  with step(3) show ?case by auto
qed simp

lemma uncurry_rel_step: assumes hvfR: "\<And> l r. (l,r) \<in> R \<Longrightarrow> hvf_term a l"
  and hvfRw: "\<And> l r. (l,r) \<in> Rw \<Longrightarrow> hvf_term a l"
  and eta: "eta_closed a sm R R"
  and etaw: "eta_closed a sm Rw Rw"
  and trs: "\<And> x r. (Var x,r) \<notin> R" 
  and trsw: "\<And> x r. (Var x,r) \<notin> Rw" 
  and step: "(s,t) \<in> rel_step (rstep R,rstep Rw)"
  shows "(uncurry_term a sm s, uncurry_term a sm t) \<in> rel_rstep (uncurry_trs a sm R, uncurry_trs a sm Rw \<union> uncurry_of_sig a sm)" (is "_ \<in> rel_rstep (?R,?Rw \<union> ?S)")
proof -
  let ?u = "uncurry_term a sm"
  let ?NS = "(rstep (?Rw \<union> ?S))^*"
  let ?RS = "rel_rstep (?R, ?S)"
  from step[unfolded rel_step.simps] obtain u v where
    su: "(s,u) \<in> (rstep Rw)^*"
    and uv: "(u,v) \<in> rstep R"
    and vt: "(v,t) \<in> (rstep Rw)^*" by auto
  note steps = uncurry_steps[OF hvfRw etaw trsw]
  from steps[OF _ su] have "(?u s, ?u u) \<in> ?NS" by auto 
  moreover from uncurry_step[OF hvfR eta trs uv] have "(?u u, ?u v) \<in> ?RS" by auto
  moreover from steps[OF _ vt] have "(?u v, ?u t) \<in> ?NS" by auto
  ultimately have steps: "(?u s, ?u t) \<in> ?NS O ?RS O ?NS" by auto
  show ?thesis
    by (rule set_mp[OF _ steps],
    simp add: rel_step.simps rstep_union, regexp)
qed

lemma uncurry_trs_union: "uncurry_trs a sm (R \<union> R') = uncurry_trs a sm R \<union> uncurry_trs a sm R'" unfolding uncurry_trs_def by auto

lemma rel_SN_mono': "R \<subseteq> R' \<Longrightarrow> S \<subseteq> R' \<union> S' \<Longrightarrow> rel_SN (R',S') \<Longrightarrow> rel_SN (R,S)"
  unfolding rel_SN_conv rel_SN_alt.simps by blast

lemma uncurrying_rel_sound: assumes 
       hvfR: "\<And> l r. (l,r) \<in> R_eta \<Longrightarrow> hvf_term a l"
  and hvfRw: "\<And> l r. (l,r) \<in> Rw_eta \<Longrightarrow> hvf_term a l"
  and eta2: "R \<subseteq> R_eta"
  and eta1: "eta_closed a sm R_eta R_eta"
  and eta2w: "Rw \<subseteq> R_eta \<union> Rw_eta"
  and eta1w: "eta_closed a sm (R_eta \<union> Rw_eta) (R_eta \<union> Rw_eta)"
  and trsw: "\<And> x r. (Var x,r) \<notin> Rw_eta" 
  and SN: "rel_SN (rstep (uncurry_trs a sm R_eta), rstep (uncurry_trs a sm Rw_eta \<union> uncurry_of_sig a sm))" (is "rel_SN (rstep ?UR,rstep (?URw \<union> ?US))")
  shows "rel_SN (rstep R,rstep Rw)"
proof (rule rel_SN_mono[OF rstep_mono[OF eta2] rstep_mono[OF eta2w]], unfold rel_SN_def, rule)
  fix f
  let ?Rw = "rstep (R_eta \<union> Rw_eta)"
  let ?R = "rstep R_eta"
  let ?URW = "uncurry_trs a sm (R_eta \<union> Rw_eta)"
  assume steps: "\<forall> i. (f i, f (Suc i)) \<in> rel_step (?R,?Rw)"
  let ?u = "uncurry_term a sm"
  show False
  proof (cases "\<exists> x r. (Var x, r) \<in> R_eta")
    case False
    obtain g where g: "g = (\<lambda> i. ?u (f i))" by auto
    {
      fix i   
      have "(g i, g (Suc i)) \<in> (rel_rstep (?UR,?URW \<union> ?US))"
        unfolding g
        by (rule uncurry_rel_step[OF hvfR _ eta1 eta1w _ _ steps[THEN spec, of i]], insert False trsw hvfR hvfRw, blast+) 
    }    
    hence "\<not> SN (rel_rstep (?UR,?URW \<union> ?US))" unfolding SN_defs by auto
    hence "\<not> rel_SN (rstep ?UR,rstep (?URW \<union> ?US))" by (simp add: rel_SN_def)
    moreover have "rel_SN (rstep ?UR,rstep (?URW \<union> ?US))"
      by (rule rel_SN_mono'[OF subset_refl _ SN], auto simp: rstep_union uncurry_trs_union)
    ultimately show False ..
  next
    case True
    then obtain x r where xr: "(Var x, r) \<in> R_eta" by auto
    hence "(?u (Var x), ?u r) \<in> ?UR" unfolding uncurry_trs_def by auto
    hence "(Var x, ?u r) \<in> ?UR" unfolding uncurry_term.simps by auto
    from lhs_var_imp_rstep_not_SN[OF this] have nSN: "\<not> (SN (rstep ?UR))" .
    from SN have SN: "SN (rel_step (rstep ?UR, rstep (?URw \<union> ?US)))"
      unfolding rel_SN_def by auto
    have "SN (rstep ?UR)" by (rule SN_subset[OF SN], auto simp: rel_step.simps)
    with nSN show False ..
  qed
qed 

lemma uncurrying_sound: assumes hvfR: "\<And> l r. (l,r) \<in> R_eta \<Longrightarrow> hvf_term a l"
  and eta2: "R \<subseteq> R_eta"
  and eta1: "eta_closed a sm R_eta R_eta"
  and SN: "SN (rstep (uncurry_trs a sm R_eta \<union> uncurry_of_sig a sm))" (is "SN (rstep (?U \<union> ?S))")
  shows "SN (rstep R)"
proof(rule SN_subset[OF _ rstep_mono[OF eta2]],rule) 
  fix f
  assume steps: "\<forall> i. (f i, f (Suc i)) \<in> rstep R_eta"
  let ?u = "uncurry_term a sm"
  show False
  proof (cases "\<exists> x r. (Var x, r) \<in> R_eta")
    case False
    obtain g where g: "g = (\<lambda> i. ?u (f i))" by auto
    {
      fix i      
      have step: "(g i, g (Suc i)) \<in> (rstep (?U \<union> ?S))^+"
        by (unfold g, rule uncurry_step_trancl[OF hvfR eta1 _ steps[THEN spec, of i]], insert False, auto)
    }
    hence "\<not> SN ((rstep (?U \<union> ?S))^+)" unfolding SN_defs by auto
    with SN_imp_SN_trancl[OF SN] show False by simp
  next
    case True
    then obtain x r where xr: "(Var x, r) \<in> R_eta" by auto
    hence "(?u (Var x), ?u r) \<in> ?U \<union> ?S" unfolding uncurry_trs_def by auto
    hence "(Var x, ?u r) \<in> ?U \<union> ?S" unfolding uncurry_term.simps by auto
    from lhs_var_imp_rstep_not_SN[OF this] SN show False ..
  qed
qed

definition only_eta :: "('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> bool"
where "only_eta E R_eta \<equiv> \<forall> l r. (l,r) \<in> E \<longrightarrow> (\<exists> a t l' r'. (l',r') \<in> R_eta \<and> instance_rule (l,r) (Fun a [l',t], Fun a [r',t]))"


lemma only_eta: assumes eta: "only_eta E (E \<union> R)" shows "rstep E \<subseteq> rstep R"
proof(rule trs_inclusion[OF one_step_ctxt_closed_rstep subst_closed_rstep], rule)
  fix l r
  assume "(l,r) \<in> E"
  thus "(l,r) \<in> rstep R"
  proof (induct l arbitrary: r rule: wf_induct[OF wf_measures[of "[size]"]])
    case (1 l r)
    show ?case
    proof (cases "(l,r) \<in> R")
      case True
      thus ?thesis by (rule rstep.id)
    next
      case False
      show ?thesis 
      proof (cases l)
        case (Var x)
        with eta 1(2) have "(l,r) \<in> R" unfolding only_eta_def instance_rule_def by force
        thus ?thesis using False by simp
      next
        case (Fun f ss)
        from False 1(2)
        obtain a t l' r' where l'r': "(l',r') \<in> E \<union> R" and inst: "instance_rule (l,r) (Fun a [l', t], Fun a [r',t])"
          using eta[unfolded only_eta_def] by blast+
        from inst[unfolded instance_rule_def] obtain \<sigma> where l: "l = Fun a [l',t] \<cdot> \<sigma>" and r: "r = Fun a [r',t] \<cdot> \<sigma>" by auto
        hence "size l' < size l" using size_subst[of l' \<sigma>] by simp
        with 1(1)[THEN spec, of l', THEN mp] l'r' have "(l',r') \<in> rstep R" by auto 
        from rstep.subst[OF rstep.ctxt[OF this,of "More a [] \<box> [t]"],of \<sigma>] show ?thesis unfolding l r
          by simp
      qed
    qed
  qed
qed

lemma only_eta_subset:
  assumes eta: "only_eta E (E \<union> R)"
  shows "rstep R = rstep (E \<union> R)"
  using rstep_mono only_eta[OF eta] unfolding rstep_union
  by auto

definition inj_sig_map :: "'f \<Rightarrow> nat \<Rightarrow> 'f sig_map \<Rightarrow> bool"
  where "inj_sig_map a m sm \<equiv> inj_on (\<lambda> (f,n,i). (get_symbol sm f n i, n + i * (m - 1))) {(f,n,i). i \<le> aarity sm f n} \<and> sm a m = [a]"

lemma inj_sig_map: assumes inj: "inj_sig_map a m sm" and i: "i < aarity sm f n" and m: "n + (Suc i) * (m - 1) = m"
  shows "get_symbol sm f n (Suc i) \<noteq> a" 
proof -
  from inj[unfolded inj_sig_map_def, THEN conjunct1, unfolded inj_on_def get_symbol_def, THEN bspec[of _ _ "(a,m,0)"], THEN bspec[of _ _ "(f,n,Suc i)"]] 
    m i
  show ?thesis by (simp add: get_symbol_def inj[unfolded inj_sig_map_def])
qed

lemma inj_sig_map_a: assumes inj: "inj_sig_map a m sm"
  shows "get_symbol sm a m 0 = a"
  using inj[unfolded inj_sig_map_def]
  unfolding get_symbol_def by auto

definition inverse_sig_map :: "'f \<Rightarrow> nat \<Rightarrow> 'f sig_map \<Rightarrow> 'f \<Rightarrow> nat \<Rightarrow> 'f \<times> nat \<times> nat"
  where "inverse_sig_map a m sm f m' \<equiv> THE (g,n,i). get_symbol sm g n i = f \<and> n + i * (m - 1) = m' \<and> i \<le> aarity sm g n"

lemma inverse_sig_map_gen: assumes inj: "inj_sig_map a m sm" and i: "i \<le> aarity sm f n" shows "inverse_sig_map a m sm (get_symbol sm f n i) (n+i * (m - 1)) = (f,n,i)" 
  unfolding inverse_sig_map_def
proof (rule, simp add: i, clarify)
  fix ff nn ii
  assume "get_symbol sm ff nn ii = get_symbol sm f n i" and "nn + ii * (m - 1) = n + i * (m - 1)" and "ii \<le> aarity sm ff nn"
  with i inj[unfolded inj_sig_map_def inj_on_def, THEN conjunct1, THEN bspec[of _ _ "(f,n,i)"], THEN bspec[of _ _ "(ff,nn,ii)"]]
  show "ff = f \<and> (nn,ii) = (n,i)" by auto
qed

lemma inverse_sig_map: assumes inj: "inj_sig_map a 2 sm" and i: "i \<le> aarity sm f n" shows "inverse_sig_map a 2 sm (get_symbol sm f n i) (n+i) = (f,n,i)"
  using inverse_sig_map_gen[OF inj i] by simp


lemma inverse_sig_map_a: assumes inj: "inj_sig_map a m sm" shows  "inverse_sig_map a m sm a m = (a,m,0)" 
proof -
  from inj[unfolded inj_sig_map_def, THEN conjunct2] 
  have "inverse_sig_map a m sm a m = inverse_sig_map a m sm (get_symbol sm a m 0) (m + 0 * (m - 1))" unfolding get_symbol_def by (auto)
  also have "\<dots> = (a,m,0)" by (rule inverse_sig_map_gen[OF inj], simp)
  finally show ?thesis . 
qed

fun curry :: "'f \<Rightarrow> 'f sig_map \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term"
where "curry a _ (Var x) = (Var x)"
  |   "curry a sm (Fun f ts) = (let (f,n,i) = inverse_sig_map a 2 sm f (length ts);
                                    ss = map (curry a sm) ts
                                 in apply_args a (Fun f (take n ss)) (drop n ss))"

fun curry_subst :: "'f \<Rightarrow> 'f sig_map \<Rightarrow> ('f,'v)subst \<Rightarrow> ('f,'v)subst"
where "curry_subst a sm (sig) = (\<lambda>x. curry a sm (sig x))"

lemma curry_subst: "curry a sm (t \<cdot> \<sigma>) = curry a sm t \<cdot> curry_subst a sm \<sigma>"
proof (induct t)
  case (Var x)
  thus ?case by simp
next
  case (Fun f ss)
  hence id: "map (\<lambda>x. curry a sm (x \<cdot> \<sigma>)) ss = map (\<lambda> x. curry a sm x \<cdot> curry_subst a sm \<sigma>) ss" by auto
  show ?case by (cases "inverse_sig_map a 2 sm f (length ss)", auto simp: o_def Let_def apply_args_subst take_map[symmetric] drop_map[symmetric] id)
qed

fun curry_ctxt :: "'f \<Rightarrow> 'f sig_map \<Rightarrow> ('f,'v)ctxt \<Rightarrow> ('f,'v)ctxt"
where "curry_ctxt a _ \<box> = \<box>"
  |   "curry_ctxt a sm (More f bef C aft) = (let (g,n,i) = inverse_sig_map a 2 sm f (Suc (length bef + length aft));
                                    cbef = map (curry a sm) bef;
                                    caft = map (curry a sm) aft;   
                                    D = curry_ctxt a sm C;
                                    m = n - Suc (length bef)                                   
                                 in (if length bef < n then apply_args_ctxt a (More g cbef D (take m caft)) (drop m caft)
                                                       else apply_args_ctxt a (More a [apply_args a (Fun g (take n cbef)) (drop n cbef)] D []) caft))"

lemma curry_ctxt: "curry a sm (C \<langle> t \<rangle> ) = (curry_ctxt a sm C) \<langle> curry a sm t \<rangle> "
proof (induct C)
  case Hole
  thus ?case by simp
next
  case (More f bef C aft)
  show ?case
  proof (cases "inverse_sig_map a 2 sm f (Suc (length bef + length aft))")
    case (fields g n i)
    {
      assume "length bef < n"
      hence "n - length bef = Suc (n - Suc (length bef))" by arith
    } note helper = this
    show ?thesis
      by (simp add: fields Let_def More apply_args_ctxt_insert take_map[symmetric] drop_map[symmetric], auto simp: apply_args_append helper)
  qed
qed


definition curry_trs :: "'f \<Rightarrow> 'f sig_map \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs"
where "curry_trs a sm R = {(curry a sm l, curry a sm r) | l r. (l,r) \<in> R}"


lemma curry_rstep: 
  assumes step: "(s,t) \<in> rstep R"
  shows "(curry a sm s, curry a sm t) \<in> rstep (curry_trs a sm R)"
using step
proof (induct)
  case (IH C \<sigma> l r)
  show "(curry a sm C\<langle>l \<cdot> \<sigma>\<rangle>,curry a sm C\<langle>r \<cdot> \<sigma>\<rangle>) \<in> rstep (curry_trs a sm R)"
    unfolding curry_ctxt curry_subst  
    by (rule rstep.ctxt, rule rstep.subst, rule rstep.id, unfold curry_trs_def, insert IH, auto)
qed

lemma curry_apply_args: assumes inj: "inj_sig_map a 2 sm"
  shows "curry a sm (apply_args a t ts) = apply_args a (curry a sm t) (map (curry a sm) ts)"
proof (induct ts rule: List.rev_induct)
  case Nil thus ?case by simp
next
  case (snoc s ss)
  have two: "Suc (Suc 0) = 2" by simp
  show ?case
    by (simp add: apply_args_last snoc[symmetric] inverse_sig_map_a[OF inj] two)
qed

lemma curry_uncurry: assumes inj: "inj_sig_map a 2 sm"
  shows "curry a sm (uncurry_term a sm t) = t"
proof -
  let ?c = "curry a sm"
  let ?u = "uncurry_term a sm"
  show ?thesis
  proof (induct t rule: applicative_term_induct[where a = a])
    case (Var x ts)
    have "?c (?u (apply_args a (Var x) ts)) = ?c (apply_args a (Var x) (map ?u ts))" 
      unfolding uncurry_apply_args_var ..
    also have "\<dots> = apply_args a (Var x) (map ?c (map ?u ts))"
      unfolding curry_apply_args[OF inj] by simp
    also have "\<dots> = apply_args a (Var x) ts"
      by (rule arg_cong[where f = "apply_args a (Var x)"], insert Var, induct ts, auto)
    finally show ?case . 
  next
    case (Fun f ss ts)
    let ?n = "min (length ts) (aarity sm f (length ss))"
    let ?f = "get_symbol sm f (length ss) ?n"
    from Fun(1) have id_ts: "map (curry a sm) (map (uncurry_term a sm) ts) = ts" by (induct ts, auto)
    from Fun(2) have id_ss: "map (curry a sm) (map (uncurry_term a sm) ss) = ss" by (induct ss, auto)
    from Fun(3) have napp: "(f,length ss) \<noteq> (a,2)" by simp
    have inv: "inverse_sig_map a 2 sm ?f (length ss + ?n) = (f,length ss,?n)" 
      by (rule inverse_sig_map[OF inj], simp)
    have "?c (?u (apply_args a (Fun f ss) ts)) = apply_args a (?c (Fun ?f (map ?u ss @ take ?n (map ?u ts)))) (drop ?n ts)" 
      unfolding uncurry_apply_args_fun[OF Fun(3)]
      unfolding curry_apply_args[OF inj]
      unfolding drop_map[symmetric]
      unfolding id_ts ..
    also have "\<dots> = apply_args a (apply_args a (Fun f ss) (take ?n ts)) (drop ?n ts)" 
      using id_ts id_ss by (simp add: o_def take_map[symmetric] inv)
    also have "\<dots> = apply_args a (Fun f ss) ts" 
      unfolding apply_args_append[symmetric] by simp
    finally show ?case . 
  qed
qed

lemma curry_uncurry_trs: assumes "inj_sig_map a 2 sm"
  shows "curry_trs a sm (uncurry_trs a sm R) = R"
proof (rule set_eqI, clarify)
  fix l r
  let ?c = "curry a sm"
  let ?u = "uncurry_term a sm"
  let ?cu = "\<lambda> t. ?c (?u t)"
  let ?CU = "curry_trs a sm (uncurry_trs a sm R)"
  note curry_uncurry = curry_uncurry[OF assms]
  show "((l,r) \<in> ?CU) = ((l,r) \<in> R)"
  proof
    assume "(l,r) \<in> R"
    hence "(?cu l,?cu r) \<in> ?CU"
      unfolding curry_trs_def uncurry_trs_def by force
    thus "(l,r) \<in> ?CU" unfolding curry_uncurry .
  next
    assume "(l,r) \<in> ?CU"
    then obtain ll rr where lr: "(ll,rr) \<in> R" and l: "l = ?cu ll" and r: "r = ?cu rr"
      unfolding curry_trs_def uncurry_trs_def by force
    from lr show "(l,r) \<in> R" unfolding l r curry_uncurry .
  qed
qed


lemma curry_trs_uncurry_of_sig: assumes inj: "inj_sig_map a 2 sm"
  shows "curry_trs a sm (uncurry_of_sig a sm) \<subseteq> Id" 
proof (clarify)
  fix cl cr
  assume clcr: "(cl,cr) \<in> curry_trs a sm (uncurry_of_sig a sm)"
  let ?c = "curry a sm"
  from clcr obtain l r where lr: "(l,r) \<in> uncurry_of_sig a sm" and cl: "cl = ?c l" and cr: "cr = ?c r"
    unfolding curry_trs_def by auto
  then obtain f n i where 
    l: "l = Fun a [generate_f_xs (get_symbol sm f n i) (n + i), Var (generate_var (n+i))]"
    and r: "r = generate_f_xs (get_symbol sm f n (Suc i)) (n + Suc i)"
    and i: "i < aarity sm f n"
    unfolding uncurry_of_sig_def by auto
  hence i: "i \<le> aarity sm f n" and si: "Suc i \<le> aarity sm f n" by auto
  note invi = inverse_sig_map[OF inj i]
  note invsi = inverse_sig_map[OF inj si, simplified]
  have id: "Fun a [?c (generate_f_xs (get_symbol sm f n i) (n + i)), Var (generate_var (n+i))] = ?c (generate_f_xs (get_symbol sm f n (Suc i)) (n + Suc i))" 
    unfolding generate_f_xs_def
    by (simp add: o_def invi invsi Let_def apply_args_last)
  have two: "Suc (Suc 0) = 2" by simp
  show "cl = cr" unfolding cl cr l r  
    by (simp add: inverse_sig_map_a[OF inj] two id)
qed
  
  
lemma curry_rstep_uncurry_of_sig: assumes inj: "inj_sig_map a 2 sm"
  and step: "(s,t) \<in> rstep (uncurry_of_sig a sm)"
  shows "curry a sm s = curry a sm t"
proof -
  from rstep_mono[OF curry_trs_uncurry_of_sig[OF inj], unfolded rstep_id] curry_rstep[OF step]
  show ?thesis by auto
qed

lemma uncurry_of_sig_SN: assumes inj: "inj_sig_map a 2 sm"
  shows "SN (rstep (uncurry_of_sig a sm))" (is "SN (rstep ?R)")
proof(rule ccontr)
  assume nSN: "\<not> ?thesis"
  have wf: "wf_trs ?R"
  proof -
    { 
      fix l r
      assume lr: "(l,r) \<in> ?R"
      have "vars_term r \<subseteq> vars_term l"
      proof -
        from lr[unfolded uncurry_of_sig_def]
        obtain f n i where l: "l = Fun a [generate_f_xs (get_symbol sm f n i) (n+ i), Var (generate_var (n+ i))]" and r: "r = generate_f_xs (get_symbol sm f n (Suc i)) (n + Suc i)"
          by auto
        {
          fix f i
          have "vars_term (generate_f_xs f i) = generate_var ` {0 ..< i}"
            unfolding generate_f_xs_def by auto
        } note vars_gen = this
        have id: "{0..<Suc (n+i)} = insert (n+i) {0..<(n+i)}" by auto
        show ?thesis unfolding l r 
          by (simp add: vars_gen id)
      qed
    }
    thus ?thesis
      unfolding wf_trs_def uncurry_of_sig_def by auto
  qed
  {
    fix f n
    have "defined ?R (f,n) \<Longrightarrow> ((f,n) = (a,2))"
      unfolding defined_def uncurry_of_sig_def
      by (auto)
  } note defined = this
  from not_SN_imp_ichain[OF wf nSN]
  obtain s t \<sigma> where "ichain (DP id ?R,?R) s t \<sigma>" by blast
  then obtain s t where "(s,t) \<in> DP id ?R" by auto
  from this[unfolded DP_def] 
  obtain l r h us
    where lr: "(l,r) \<in> ?R" and u: "r \<unrhd> Fun h us" and d: "defined ?R (h, length us)"  by auto
  from defined[OF d] have h: "h = a" and us: "length us = 2" by auto
  from lr[unfolded uncurry_of_sig_def]
  obtain f n i where i: "i < aarity sm f n" and r: "r = generate_f_xs (get_symbol sm f n (Suc i)) (n + Suc i)" (is "_ = ?r") by auto
  from u[unfolded r generate_f_xs_def] have "?r = Fun h us" 
  proof (cases)
    case refl
    thus ?thesis unfolding generate_f_xs_def by auto
  next
    case (subt u)
    then obtain x where "u = Var x" by auto
    from subt(2)[unfolded this] have False by cases
    thus ?thesis by simp
  qed
  from this[unfolded h] have get: "get_symbol sm f n (Suc i) = a" and 2: "n + Suc i = 2" unfolding generate_f_xs_def us[symmetric] by auto
  from inj_sig_map[OF inj i] 2 get 
  show False by simp 
qed

lemma uncurry_preserves_SN: assumes inj: "inj_sig_map a 2 sm"
  and SN: "SN_elt (rstep R) t"
  and curry: "t = curry a sm s"
  shows "SN_elt (rstep (uncurry_trs a sm R \<union> uncurry_of_sig a sm)) s" (is "SN_elt (rstep (?U \<union> ?S)) _")
proof 
  fix f
  assume f0: "f 0 = s" and fsteps: "\<forall> i. (f i, f (Suc i)) \<in> rstep (?U \<union> ?S)"
  obtain g where g: "g = (\<lambda> i. curry a sm (f i))" by auto
  let ?C = "curry_trs a sm"
  from f0 curry have t: "t = g 0" unfolding g by auto
  { 
    fix i
    from curry_rstep[OF fsteps[THEN spec, of i], of a sm]
    have "(g i, g (Suc i)) \<in> rstep (?C (?U \<union> ?S))" unfolding g .
    hence "(g i, g (Suc i)) \<in> rstep (?C ?U \<union> ?C ?S)" unfolding curry_trs_def by force
    hence "(g i, g (Suc i)) \<in> rstep R \<union> rstep (?C ?S)" unfolding curry_uncurry_trs[OF inj] rstep_union .
    hence "(g i, g (Suc i)) \<in> rstep R \<union> Id" 
      using rstep_mono[OF curry_trs_uncurry_of_sig[OF inj], unfolded rstep_id] by auto
  } note gsteps = this
  hence "\<forall> i. (g i, g (Suc i)) \<in> Id \<union> rstep R" by auto
  from non_strict_ending[OF this _, simplified, OF SN[unfolded t]]
  obtain j where j: "\<And> i. i \<ge> j \<Longrightarrow> (g i, g (Suc i)) \<notin> rstep R" by auto
  obtain h where h: "h = (\<lambda> i. f (i + j))" by auto
  {
    fix i
    from fsteps have step: "(h i, h (Suc i)) \<in> rstep ?U \<union> rstep ?S" unfolding h rstep_union by auto
    have "(h i, h (Suc i)) \<in> rstep ?S"      
    proof (rule ccontr)
      assume "(h i, h (Suc i)) \<notin> rstep ?S"
      with step have "(h i, h (Suc i)) \<in> rstep ?U" by auto
      from curry_rstep[OF this, of a sm, unfolded curry_uncurry_trs[OF inj]]
      have "(g (i+j), g (Suc (i+j))) \<in> rstep R" unfolding g h by simp
      with j[of "i + j"] show False by simp
    qed
  }
  hence "\<not> SN (rstep ?S)" by auto
  with uncurry_of_sig_SN[OF inj]
  show False by simp
qed
  
lemma uncurrying_sound_dp: assumes hvfPR: "\<And> l r. (l,r) \<in> R_eta \<union> P \<Longrightarrow> hvf_term a l"
  and eta1: "eta_closed a sm R_eta R_eta"
  and eta2: "R_eta = E \<union> R"
  and eta3: "only_eta E R_eta"
  and inj: "inj_sig_map a 2 sm"
  and finite: "finite_dpp (uncurry_trs a sm P,uncurry_trs a sm R_eta \<union> uncurry_of_sig a sm)" (is "finite_dpp (?P,?R)")
  shows "finite_dpp (P,R)"
unfolding finite_dpp_def
proof (clarify)
  fix s t \<sigma>
  assume  "min_ichain (P,R) s t \<sigma>"
  hence chain: "min_ichain (P,R_eta) s t \<sigma>"
    by (auto simp: only_eta_subset[OF eta3[unfolded eta2]] eta2 minimal_cond_def)
  let ?u = "uncurry_term a sm"
  let ?s = "\<lambda> i. ?u (s i)"
  let ?t = "\<lambda> i. ?u (t i)"
  let ?us = "\<lambda> i. uncurry_subst a sm (\<sigma> i)"
  {
    fix i
    from chain have "(s i, t i) \<in> P" by auto
    hence P: "(?s i, ?t i) \<in> ?P" unfolding uncurry_trs_def by auto
    from chain have "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (rstep R_eta)^*" by auto
    hence R: "(?u (t i \<cdot> \<sigma> i), ?u (s (Suc i) \<cdot> \<sigma> (Suc i))) \<in> (rstep ?R)^*"
    proof (induct)
      case base
      thus ?case by simp
    next
      case (step y z)
      have "(?u y, ?u z) \<in> (rstep ?R)^+"
      proof (rule uncurry_step_trancl[OF _ eta1 _ step(2)])
        fix l r
        assume "(l,r) \<in> R_eta"
        with hvfPR show "hvf_term a l" by blast
      next
        fix x r
        show "(Var x,r) \<notin> R_eta"
        proof
          assume "(Var x,r) \<in> R_eta"
          from left_var_imp_not_SN[OF this] chain 
          show False by (auto simp: minimal_cond_def)
        qed
      qed
      from rtrancl_trancl_trancl[OF step(3) this] show ?case ..
    qed
    have t: "(?u (t i) \<cdot> ?us i, ?u (t i \<cdot> \<sigma> i)) \<in> (rstep ?R)^*"
      using set_mp[OF rtrancl_mono[OF rstep_mono[of _ ?R]] uncurry_subst]
      by auto
    have hvfs: "hvf_term a (s (Suc i))" 
      by (rule hvfPR, insert chain, auto)
    have s: "?u (s (Suc i)) \<cdot> ?us (Suc i) = ?u (s (Suc i) \<cdot> \<sigma> (Suc i))"
      by (rule uncurry_subst_hvf[OF hvfs, symmetric])
    have R: "(?u (t i) \<cdot> ?us i, ?u (s (Suc i)) \<cdot> ?us (Suc i)) \<in> (rstep ?R)^*"
      unfolding s
      by (rule rtrancl_trans[OF t R])
    from chain have SN: "SN_elt (rstep R_eta) (t i \<cdot> \<sigma> i)" by (auto simp: minimal_cond_def)
    have min: "SN_elt (rstep ?R) (?u (t i) \<cdot> ?us i)" 
    proof (rule uncurry_preserves_SN[OF inj SN])
      let ?c = "curry a sm"
      have "t i \<cdot> \<sigma> i = ?c (?u (t i \<cdot> \<sigma> i))" unfolding curry_uncurry[OF inj] ..
      also have "\<dots> = ?c (?u (t i) \<cdot> ?us i)" 
        using uncurry_subst[of a sm "t i" "\<sigma> i"]
      proof (induct)
        case base thus ?case by simp
      next
        case (step y z)
        from curry_rstep_uncurry_of_sig[OF inj step(2), symmetric] step(3)
        show ?case by simp
      qed        
      finally 
      show "t i \<cdot> \<sigma> i = ?c (?u (t i) \<cdot> ?us i)" .
    qed
    note P and R and min
  }
  hence chain: "min_ichain (?P, ?R) ?s ?t ?us" by (auto simp: minimal_cond_def)
  from finite have "\<not> (\<exists> s t sig. min_ichain (?P, ?R) s t sig)" unfolding finite_dpp_def by auto
  with chain show False by blast
qed


section {* uncurrying of top symbols *}

abbreviation (input) get_default_sym :: "'f sig_map \<Rightarrow> 'f \<Rightarrow> nat \<Rightarrow> 'f"
where "get_default_sym sm \<equiv> \<lambda> f n. get_symbol sm f n 0"

abbreviation map_sym :: "'f sig_map \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term"
where "map_sym sm \<equiv> map_term_wa (get_default_sym sm)"


definition get_inv_default_sym where "get_inv_default_sym a m sm \<equiv> \<lambda> f n. fst (inverse_sig_map a m sm f n)"


lemma inj_sig_map_below: assumes inj: "inj_sig_map a m sm"
  shows "get_inv_default_sym a m sm (get_default_sym sm f n) n = f"
  unfolding get_inv_default_sym_def
  using inverse_sig_map_gen[OF inj, of 0 f n, simplified]
  by auto

definition map_inv_sym :: "'f \<Rightarrow> nat \<Rightarrow> 'f sig_map \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term"
  where "map_inv_sym a m sm \<equiv> map_term_wa (get_inv_default_sym a m sm)"


lemma map_inv_sym: assumes inj: "inj_sig_map a m sm"
  shows "map_inv_sym a m sm (map_sym sm t) = t"
  by (unfold map_inv_sym_def map_term_wa_compose, rule map_term_wa_id[OF inj_sig_map_below[OF inj]])

definition map_inv_trs :: "'f \<Rightarrow> nat \<Rightarrow> 'f sig_map \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs"
  where "map_inv_trs a m sm \<equiv> map_trs_wa (get_inv_default_sym a m sm)"


abbreviation uncurry_below_trs :: "'f sig_map \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs"
where "uncurry_below_trs sm \<equiv> map_trs_wa (get_default_sym sm)"  

lemma map_inv_trs: assumes inj: "inj_sig_map a m sm"
  shows "map_inv_trs a m sm (uncurry_below_trs sm R) = R"
  by (unfold map_inv_trs_def map_trs_wa_compose, rule map_trs_wa_id[OF inj_sig_map_below[OF inj]])


lemma map_sym_preserves_SN: assumes inj: "inj_sig_map a m sm"
  and SN: "SN_elt (rstep R) t"
  shows "SN_elt (rstep (uncurry_below_trs sm R)) (map_sym sm t)"
proof
  fix f
  assume f0: "f 0 = map_sym sm t" and steps: "\<forall> i. (f i, f (Suc i)) \<in> rstep (uncurry_below_trs sm R)"
  obtain g where g: "g = (\<lambda> i. map_inv_sym a m sm (f i))" by auto
  let ?R = "map_inv_trs a m sm (uncurry_below_trs sm R)"
  {
    fix i
    from map_trs_wa_rstep[OF steps[THEN spec, of i]]
    have "(g i, g (Suc i)) \<in> rstep ?R"
      unfolding g map_inv_sym_def map_inv_trs_def .
    hence "(g i, g (Suc i)) \<in> rstep R" 
      unfolding map_inv_trs[OF inj] .
  } note steps = this
  have "g 0 = t" unfolding g f0 by (rule map_inv_sym[OF inj])
  with steps SN show False by auto
qed


fun uncurry_top :: "'f \<Rightarrow> nat \<Rightarrow> 'f sig_map \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term"
  where "uncurry_top a n sm (Fun f ts) = (let mt = map (map_sym sm) in if f = a \<and> length ts = n \<and> \<not> (is_var (hd ts)) \<and> aarity sm (the (root (hd ts))) (num_args (hd ts)) \<noteq> 0 then (case hd ts of Fun g ss \<Rightarrow> Fun (get_symbol sm g (length ss) 1) (mt (ss @ tl ts)))
               else Fun (get_default_sym sm f (length ts)) (mt ts))"
      | "uncurry_top a n sm (Var x) = Var x"

fun hvf_top :: "'f \<Rightarrow> nat \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
where "hvf_top a n (Fun f ts) = (f = a \<and> length ts = n \<longrightarrow> \<not> (is_var (hd ts)))"
    | "hvf_top a n (Var _) = False"

abbreviation uncurry_top_subst where "uncurry_top_subst sm \<equiv> map_subst_wa (get_default_sym sm)"

lemma uncurry_top_subst_hvf: assumes hvf: "hvf_top a (Suc n) t"
   shows "uncurry_top a (Suc n) sm (t \<cdot> \<sigma>) = uncurry_top a (Suc n) sm t \<cdot> uncurry_top_subst sm \<sigma>"
proof -
  from hvf obtain f ts where t: "t = Fun f ts" by (cases t, auto)
  show ?thesis
  proof (cases "f = a \<and> length ts = Suc n")
    case True
    then obtain s tss where ts: "ts = s # tss" by (cases ts, auto)
    with hvf True show ?thesis unfolding t by (cases s, auto simp: Let_def)
  next
    case False
    thus ?thesis unfolding t by auto
  qed
qed

definition uncurry_of_top_sig :: "'f \<Rightarrow> nat \<Rightarrow> 'f sig_map \<Rightarrow> ('f,string)trs"
where "uncurry_of_top_sig a m sm \<equiv> {(Fun a (generate_f_xs (get_symbol sm f n i) (n + i) # map (\<lambda> i. Var (generate_var i)) [n+i..<n+i+(m - 1)]),generate_f_xs (get_symbol sm f n (Suc i)) (n + i + (m - 1))) | f n i. i < aarity sm f n}"


lemma uncurry_top_sig: assumes a: "get_default_sym sm a (Suc n) = a"
  and ts: "length ts = n"
  shows "(Fun a (map_sym sm s # map (map_sym sm) ts), uncurry_top a (Suc n) sm (Fun a (s # ts))) \<in> (CS (uncurry_of_top_sig a (Suc n) sm))^=" (is "(Fun a (?m s #_ ), ?u (Fun a _)) \<in> ?C")
proof (cases s)
  case (Var x)
  thus ?thesis by (simp add: a ts)
next
  case (Fun f ss)
  show ?thesis
  proof (cases "aarity sm f (length ss) = 0")
    case True
    thus ?thesis by (auto simp: Fun a ts)
  next
    case False
    let ?m = "length ss"
    let ?l = "Fun a (generate_f_xs (get_symbol sm f ?m 0) ?m # map (\<lambda>i. Var (generate_var i)) [?m..<?m + n])"
    let ?r = "generate_f_xs (get_symbol sm f ?m (Suc 0)) (?m + n)"
    let ?sig = "(\<lambda>i. map_sym sm ((ss @ ts) ! the_inv generate_var i))"
    have lr: "(?l,?r) \<in> uncurry_of_top_sig a (Suc n) sm"
      unfolding uncurry_of_top_sig_def
      by (rule, rule exI[of _ f], rule exI[of _ ?m], rule exI[of _ 0], insert False, simp)
    show ?thesis       
      by  (simp add: Let_def ts Fun False, rule disjI1, rule subst_closureI2[where \<sigma> = ?sig,OF lr], simp_all add: o_def generate_f_xs_def
        the_inv_f_f[OF inj_generate_var] ts[symmetric], rule conjI, (rule nth_equalityI, simp, simp add: nth_append)+) 
  qed
qed


lemma uncurry_top_subst: assumes a: "get_default_sym sm a (Suc n) = a"
   shows "(uncurry_top a (Suc n) sm (Fun f ts) \<cdot> uncurry_top_subst sm \<sigma>, uncurry_top a (Suc n) sm (Fun f ts \<cdot> \<sigma>)) \<in> (CS (uncurry_of_top_sig a (Suc n) sm))^=" (is "_ \<in> ?C^=")
proof (cases "hvf_top a (Suc n) (Fun f ts)")
  case True
  from uncurry_top_subst_hvf[OF this] show ?thesis by auto
next
  case False
  let ?u = "uncurry_top a (Suc n) sm"
  let ?us = "uncurry_top_subst sm \<sigma>"
  let ?m = "map_sym sm"
  from False have f: "f = a" and l: "length ts = Suc n" and v: "is_var (hd ts)" by auto
  from l obtain s tts where ts: "ts = s # tts" and n: "n = length tts" by (cases ts, auto)
  from v ts obtain x where ts: "ts = Var x # tts" by (cases s, auto)
  obtain s u where s: "\<sigma> x = s" by auto
  with a f ts have l: "?u (Fun f ts) \<cdot> ?us  = Fun a (?m s # map ?m (map (\<lambda> w. w \<cdot> \<sigma>) tts))" (is "_ = ?l")  
    and r: "Fun f ts \<cdot> \<sigma> = Fun a (s # map (\<lambda>w. w \<cdot> \<sigma>) tts)" (is "_ = ?r") unfolding n by (auto) 
  show ?thesis unfolding l r by (rule uncurry_top_sig[OF a], auto simp: n)
qed

definition uncurry_top_trs :: "'f \<Rightarrow> nat \<Rightarrow> 'f sig_map \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs"
where "uncurry_top_trs a n sm R \<equiv> (\<lambda>(l,r). (uncurry_top a n sm l, uncurry_top a n sm r)) ` R"


lemma uncurry_top_preserves_SN: assumes inj: "inj_sig_map a (Suc n) sm"
  and SN: "SN_elt (rstep R) t"
  shows "SN_elt (rstep (uncurry_below_trs sm R)) (uncurry_top a (Suc n) sm t)"
proof (cases "uncurry_top a (Suc n) sm t = map_sym sm t")
  case True
  with map_sym_preserves_SN[OF inj SN] show ?thesis by simp
next 
  case False
  let ?n = "Suc n"
  from inj_sig_map_a[OF inj] have a: "get_symbol sm a ?n 0 = a" .
  from False obtain f ts where t: "t = Fun f ts" by (cases t, auto)
  from False have "f = a \<and> length ts = ?n" unfolding t by (cases "f = a \<and> length ts = ?n", auto)
  with t obtain l rs where t: "t = Fun a (l # rs)" and n: "n = length rs" by (cases ts, auto)
  from False[unfolded t n, unfolded uncurry_top.simps, simplified Let_def, simplified]
  have cond: "\<not> is_var l \<and> 0 < aarity sm (the (root l)) (num_args l)"
    by (cases "\<not> is_var l \<and> 0 < aarity sm (the (root l)) (num_args l)", auto simp: a)
  from cond t obtain g ls where t: "t = Fun a (Fun g ls # rs)" and aa: "aarity sm g (length ls) > 0"
    by (cases l, auto)
  let ?U = "uncurry_below_trs sm R"
  note SN_pres = map_sym_preserves_SN[OF inj]
  have R: "\<forall> lr \<in> ?U. \<forall> x. fst lr \<noteq> Var x"
  proof (clarsimp)
    fix r x
    assume "(Var x,r) \<in> ?U" 
    with left_var_imp_not_SN[of _ _ ?U] SN_pres[OF SN]
    show False by auto
  qed
  note SN = SN[unfolded t]
  let ?m = "map_sym sm"
  let ?g = "get_symbol sm g (length ls) (Suc 0)"
  have "SN_elt (rstep ?U) (Fun ?g (map ?m (ls @ rs)))"
  proof (rule SN_args_imp_SN[OF _ R])
    fix s
    from SN_pres[OF SN_imp_SN_arg[OF SN]] have rs: "\<And> r. r \<in> set rs \<Longrightarrow> SN_elt (rstep ?U) (?m r)" by simp
    {
      fix s
      assume "s \<in>  set (map ?m ls)"
      then obtain l where l: "l \<in> set ls" and s: "s = ?m l" by auto
      from SN_pres[OF SN_imp_SN_arg[OF SN_imp_SN_arg[OF SN, of "Fun g ls"] l]]
      have "SN_elt (rstep ?U) s" unfolding s by auto
    } note ls = this
    assume "s \<in>  set (map ?m (ls @ rs))"
    with ls rs show "SN_elt (rstep ?U) s"
      by auto
  next
    show "(?g, length (map ?m (ls @ rs))) \<notin> defs ?U"
    proof (simp add: defs_def)
      show "\<not> defined ?U (?g,length ls + length rs)"
      proof
        assume "defined ?U (?g, length ls + length rs)"
        from this[unfolded defined_def] obtain ll rr where
        llrr: " (ll, rr) \<in> ?U" and  "root ll = Some ?g \<and> num_args ll = length ls + length rs" by auto
        then obtain lls where ll: "ll = Fun ?g lls" and lls: "length lls = length ls + length rs" by (cases ll, auto)
        from llrr obtain lll rrr where lllrrr: "(lll,rrr) \<in> R" and ll2: "ll = ?m lll" unfolding map_trs_wa_def by auto
        from ll ll2 obtain h llls where lll3: "lll = Fun h llls" by (cases lll, auto)
        from ll2[unfolded ll lll3] lls have len: "length llls = length ls + length rs" and g: "?g = get_default_sym sm h (length ls + length rs)" by auto
        from inverse_sig_map_gen[OF inj, of 0 h "length ls + length rs"] have 
          one: "inverse_sig_map a ?n sm ?g (length ls + length rs) = (h,length ls + length rs,0)" unfolding g by simp
        with aa inverse_sig_map_gen[OF inj, of "Suc 0" g "length ls"] show False by (auto simp: n)
      qed
    qed
  qed
  thus ?thesis using cond aa unfolding t n
    by (simp add: Let_def)
qed

lemma uncurry_top_subst_SN: assumes 
     inj: "inj_sig_map a (Suc n) sm"
  and ndef: "\<not> defined R (a,Suc n)"
  and SN: "SN_elt (rstep R) (t \<cdot> \<sigma>)"
  shows "SN_elt (rstep (uncurry_below_trs sm R)) (uncurry_top a (Suc n) sm t \<cdot> uncurry_top_subst sm \<sigma>)"
    (is "SN_elt (rstep ?U) (?u t \<cdot> ?us)")
proof (cases t)
  case (Var x)
  thus ?thesis using map_sym_preserves_SN[OF inj SN] by auto
next
  case (Fun f ts)
  let ?n = "Suc n"
  show ?thesis
  proof (cases "hvf_top a ?n t")
    case True    
    show ?thesis unfolding uncurry_top_subst_hvf[OF True, symmetric]
      by (rule uncurry_top_preserves_SN[OF inj SN])
  next
    case False    
    let ?m = "map_sym sm"
    note a = inj_sig_map_a[OF inj]
    from False have f: "f = a" and l: "length ts = ?n" and v: "is_var (hd ts)" unfolding Fun using a by auto
    then obtain s tts where ts: "ts = s # tts" and n: "n = length tts" by (cases ts, auto)
    from v ts obtain x where ts: "ts = Var x # tts" by (cases s, auto)
    obtain s u where s: "\<sigma> x = s" by auto
    with f ts have l: "?u (Fun f ts) \<cdot> ?us  = Fun a (?m s # map ?m (map (\<lambda> w. w \<cdot> \<sigma>) tts))" (is "_ = ?l") 
      and r: "Fun f ts \<cdot> \<sigma> = Fun a (s # map (\<lambda> w. w \<cdot> \<sigma>) tts)" (is "_ = ?r") by (auto simp: a[unfolded n])
    let ?args = "?m s # map ?m (map (\<lambda> w. w \<cdot> \<sigma>) tts)"
    show ?thesis unfolding Fun l
    proof (rule SN_args_imp_SN[unfolded defs_def, simplified])
      show "\<not> defined ?U (a, length ?args)"
      proof (simp, rule)
        assume "defined ?U (a,Suc (length tts))"
        then obtain l r where lr: "(l,r) \<in> ?U" and "root l = Some a" and "num_args l = Suc (length tts)" unfolding defined_def by auto
        then obtain ls where l: "l = Fun a ls" and len: "length ls = Suc (length tts)" by (cases l, auto)
        from lr[unfolded map_trs_wa_def]
        obtain ll rr where llrr: "(ll,rr) \<in> R" and l2: "l = map_sym sm ll" by auto
        from l2 len obtain g lls where ll: "ll = Fun g lls" and len2: "length lls = ?n" and "get_default_sym sm g ?n = a" unfolding n l by (cases ll, auto)
        with inverse_sig_map_a[OF inj] have "inverse_sig_map a ?n sm (get_default_sym sm g ?n) ?n = (a,?n,0)" by simp
        with inverse_sig_map_gen[OF inj, of 0 g ?n] have g: "g = a" by auto
        from llrr[unfolded ll g] len2 have d: "defined R (a,?n)" unfolding defined_def by auto
        from ndef d show False by auto
      qed
    next
      fix u
      assume "u \<in> set ?args"
      hence dis: "u = ?m s \<or> (\<exists> t \<in> set tts. u = ?m (t \<cdot> \<sigma>))" by auto
      note SN = map_sym_preserves_SN[OF inj SN_imp_SN_arg[OF SN[unfolded Fun ts, simplified]]]
      from dis show "SN_elt (rstep ?U) u"
      proof
        assume u: "u = ?m s"
        show ?thesis unfolding u by (rule SN, simp add: s)
      next
        assume "\<exists> t \<in> set tts. u = ?m (t \<cdot> \<sigma>)"
        then obtain t where t: "t \<in> set tts" and u: "u = ?m (t \<cdot> \<sigma>)" by auto
        show ?thesis unfolding u by (rule SN, insert t, auto)
      qed
    next
      {
        fix r x
        assume "(Var x,r) \<in> ?U"
        then obtain ll rr where "(ll,rr) \<in> R" and "Var x = ?m ll"  unfolding map_trs_wa_def by auto
        then have "(Var x, rr) \<in> R" by (cases ll, auto)
        from left_var_imp_not_SN[OF this] SN have False by simp
      }
      thus "\<forall> lr \<in> ?U. \<forall> x. fst lr \<noteq> Var x" by auto
    qed
  qed
qed

definition eta_closed_top :: "'f \<Rightarrow> nat \<Rightarrow> 'f sig_map \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> bool"
  where "eta_closed_top a n sm R P \<equiv> \<forall> f ls r. (Fun f ls,r) \<in> R \<longrightarrow> aarity sm f (length ls) \<noteq> 0 \<longrightarrow> (\<exists> ll rr ys. length ys = n - 1 \<and> distinct ys \<and> 
           set ys \<inter> vars_rule (ll,rr) = {} \<and> instance_rule (Fun f ls,r) (ll,rr) \<and>
           (Fun a (ll # map Var ys), Fun a (rr # map Var ys)) \<in> P)"

lemma uncurry_top_steps: assumes 
      inj: "inj_sig_map a (Suc n) sm"
  and ndef: "\<not> defined R (f,length ts)"
  and ndefa: "\<not> defined R (a,(Suc n))"
  and steps: "(Fun f ts,s) \<in> (rstep R)^*"
  and P: "\<forall> (l,r) \<in> P. hvf_top a (Suc n) l"
  and eta: "f \<noteq> a \<or> length ts \<noteq> Suc n \<or> cap R (hd ts) \<noteq> HHole \<or> (eta_closed_top a (Suc n) sm R P \<and> CS (uncurry_of_top_sig a (Suc n) sm) \<subseteq> CS P')"
  and SN: "SN_elt (rstep R) (Fun f ts)"
  shows "(uncurry_top a (Suc n) sm (Fun f ts), uncurry_top a (Suc n) sm s) \<in> (rstep (uncurry_below_trs sm R) \<union> CS (uncurry_top_trs a (Suc n) sm P \<union> P') \<inter> {(s,t). SN_elt (rstep (uncurry_below_trs sm R)) t})^*" 
  (is "(?u (Fun f ts),_) \<in> ?Rel")
  using steps
proof (induct rule: rtrancl_induct)
  case base thus ?case by simp
next
  case (step s u)
  let ?n = "Suc n"
  from inj_sig_map_a[OF inj] have a: "get_symbol sm a ?n 0 = a" .
  let ?UP = "uncurry_top_trs a ?n sm P"
  let ?UR = "uncurry_below_trs sm R"
  let ?US = "uncurry_of_top_sig a ?n sm"
  let ?CS = "CS (?UP \<union> P')"
  let ?CSm = "?CS \<inter> {(s,t). SN_elt (rstep ?UR) t}"
  have R: "\<forall> (l,r) \<in> R. \<not> (is_var l)" 
  proof (clarify)
    fix l r
    assume "(l,r) \<in> R" and "is_var l"
    with left_var_imp_not_SN[of _ r R "Fun f ts"] SN
    show False by (cases l, auto)
  qed
  note SN = steps_preserve_SN_elt[OF step(1) SN]
  from rsteps_imp_nrrsteps[OF _ _ R step(1)] ndef
  have some_steps: "(Fun f ts, s) \<in> (nrrstep R)^*"  by auto
  from nrrsteps_equiv_num_args[OF this] nrrsteps_preserve_root[OF this]
  obtain ss where s: "s = Fun f ss" and lts: "length ts = length ss" by auto
  from nrrstep_seq_imp_arg_rstep_seq[OF some_steps[unfolded s]]
  have first_steps: "ss \<noteq> [] \<Longrightarrow> (ts ! 0, ss ! 0) \<in> (rstep R)^*" by auto
  note ndef = ndef[unfolded lts]
  note steps = step(3)[unfolded s]
  note step = step(2)[unfolded s]
  note SNs = SN[unfolded s]
  note SNu = step_preserves_SN_elt[OF step SNs]
  from step obtain C l r \<sigma> where l: "Fun f ss = C\<langle>l \<cdot> \<sigma>\<rangle>" and r: "u = C\<langle>r \<cdot> \<sigma>\<rangle>" and lr: "(l,r) \<in> R" by auto
  from lr R obtain g ls where lhs: "l = Fun g ls" by (cases l, auto)
  have "(?u (Fun f ss),?u u) \<in> ?Rel"
  proof (cases C)
    case Hole
    from l[unfolded this] ndef lr show ?thesis unfolding lhs defined_def by force
  next
    case (More h bef D aft)
    from l r More have ss: "ss = bef @ D\<langle>l\<cdot>\<sigma>\<rangle> # aft" and u: "u = Fun f (bef @ D\<langle>r\<cdot>\<sigma>\<rangle> # aft)" by auto
    let ?C = "map_ctxt_wa (get_default_sym sm)"
    let ?m = "map_sym sm" 
    from lr have mlr: "(?m l, ?m r) \<in> ?UR" unfolding map_trs_wa_def by auto
    from lr have "(D\<langle>l\<cdot>\<sigma>\<rangle>, D\<langle>r\<cdot>\<sigma>\<rangle>) \<in> rstep R" by auto
    from map_trs_wa_rstep[OF this] have "(?m (D\<langle>l\<cdot>\<sigma>\<rangle>), ?m (D\<langle>r\<cdot>\<sigma>\<rangle>)) \<in> rstep ?UR" by simp
    hence step: "\<And> C. (C\<langle>?m (D\<langle>l\<cdot>\<sigma>\<rangle>)\<rangle>, C\<langle>?m (D\<langle>r\<cdot>\<sigma>\<rangle>)\<rangle>) \<in> rstep ?UR" by (rule rstep.ctxt)
    note no_trans_step = step[of "More (get_default_sym sm f (Suc (length bef + length aft))) (map ?m bef) \<box> (map ?m aft)", simplified]
    from lr have "(l\<cdot>\<sigma>, r\<cdot>\<sigma>) \<in> rstep R" by auto
    from map_trs_wa_rstep[OF this] have "(?m (l\<cdot>\<sigma>), ?m (r\<cdot>\<sigma>)) \<in> rstep ?UR" by simp
    hence small_step: "\<And> C. (C\<langle>?m (l\<cdot>\<sigma>)\<rangle>, C\<langle>?m (r\<cdot>\<sigma>)\<rangle>) \<in> rstep ?UR" by (rule rstep.ctxt)
    show ?thesis
    proof (cases "f = a \<and> Suc (length bef + length aft) = ?n")
      case False
      hence False: "\<And> t c. (f = a \<and> length (bef @ t # aft) = ?n \<and> c) = False" by auto
      show ?thesis 
        by (rule r_into_rtrancl, rule UnI1, unfold ss u uncurry_top.simps Let_def False, simp add: no_trans_step)
    next
      case True
      hence f: "f = a" by auto
      show ?thesis
      proof (cases bef)
        case (Cons b bbef)
        with True have n: "Suc (length bbef + length aft) = n" by simp
        show ?thesis
        proof (cases "\<not> (is_var b) \<and> aarity sm (the (root b)) (num_args b) \<noteq> 0")
          case False
          show ?thesis
            by (rule r_into_rtrancl, rule UnI1, insert True no_trans_step, unfold ss u f Cons, insert False, unfold uncurry_top.simps Let_def, auto)   
        next
          case True
          then obtain h hs where b: "b = Fun h hs" and len: "0 < aarity sm h (length hs)" by (cases b, auto)
          show ?thesis 
            by (rule r_into_rtrancl, rule UnI1, unfold ss u f Cons b, insert len step[of "More (get_symbol sm h (length hs) (Suc 0)) (map ?m hs @ map ?m bbef) \<box> (map ?m aft)"], unfold uncurry_top.simps Let_def, simp add: n)
        qed
      next
        case Nil
        with True have n: "n = length aft" by simp
        show ?thesis 
        proof (cases D)
          case (More g bef1 E aft1)
          show ?thesis
          proof (cases "aarity sm g (Suc (length bef1 + length aft1))")
            case 0
            show ?thesis
              by (rule r_into_rtrancl, rule UnI1, insert small_step[of "More a [] (More (get_symbol sm g (Suc (length bef1 + length aft1)) 0) (map ?m bef1) (?C E) (map ?m aft1)) (map ?m aft)"] 0, unfold ss u f Nil More, simp add: a[unfolded n])
          next
            case (Suc n)
            show ?thesis
              by (rule r_into_rtrancl, rule UnI1, insert insert small_step[of "More (get_symbol sm g (Suc (length bef1 + length aft1)) (Suc 0)) (map ?m bef1) (?C E) (map ?m aft1 @ map ?m aft)"] Suc, unfold ss u f Nil More, simp add: n)
          qed
        next
          case Hole
          from lhs lr have "defined R (g,length ls)" unfolding defined_def by auto
          with ndefa have "(g,length ls) \<noteq> (a,?n)" by auto
          hence napp: "(g = a \<and> length ls = ?n) = False" by simp
          have l: "Fun f ss = Fun a ((l \<cdot> \<sigma>) # aft)" and r: "u = Fun a ((r \<cdot> \<sigma>) # aft)" unfolding ss u f Nil Hole by auto          
          hence ss: "ss = (l \<cdot> \<sigma>) #  aft" by simp
          from lts[unfolded ss] n have "length ts = ?n" by simp
          with f obtain t1 tss where ts: "ts = t1 # tss" and n2: "n = length tss" by (cases ts, auto)
          with eta f have eta: "cap R t1 \<noteq> HHole \<or> (eta_closed_top a ?n sm R P \<and> CS ?US \<subseteq> CS P')" by auto
          from first_steps have "(t1, l \<cdot> \<sigma>) \<in> (rstep R)^*" unfolding  ss using ts by auto
          from cap_rewrites[OF this] have "equiv_class (cap R (l \<cdot> \<sigma>)) \<subseteq> equiv_class (cap R t1)" . 
          with cap_lhs[OF lr] have "equiv_class (cap R t1) = UNIV" by auto
          hence "Var ''foo'' \<in> equiv_class (cap R t1)" by auto
          with eta have eta: "eta_closed_top a ?n sm R P" and P': "CS ?US \<subseteq> CS P'"
            by ((cases "cap R t1", auto)+)          
          note an = a[unfolded n]
          note SN_pres = map_sym_preserves_SN[OF inj]
          note SNr = SN_pres[OF SNu[unfolded r]]
          note SNuu = uncurry_top_preserves_SN[OF inj SNu] 
          show ?thesis 
          proof (cases "aarity sm g (length ls)")
            case 0
            hence l: "?u (Fun f ss) = Fun a (?m (l \<cdot> \<sigma>) # map ?m aft)" unfolding l lhs
              by (simp add: an)
            from small_step[of "More a [] \<box> (map ?m aft)"]
            have "(?u (Fun f ss), Fun a (?m (r \<cdot> \<sigma>) # map ?m aft)) \<in> rstep ?UR" unfolding l by auto
            hence one: "(?u (Fun f ss), Fun a (?m (r \<cdot> \<sigma>) # map ?m aft)) \<in> ?Rel" by auto 
            have "(Fun a (?m (r \<cdot> \<sigma>) # map ?m aft), ?u u) \<in> (CS ?US)^="
              unfolding r by (rule uncurry_top_sig[OF a], simp add: n)
            with P' have "(Fun a (?m (r \<cdot> \<sigma>) # map ?m aft), ?u u) \<in> (CS P')^=" by auto
            with SNuu
            have two: "(Fun a (?m (r \<cdot> \<sigma>) # map ?m aft), ?u u) \<in> ?Rel" unfolding cs_union_distr by auto
            from one two show ?thesis by auto
          next            
            case (Suc m)
            hence aa: "aarity sm g (length ls) \<noteq> 0" by simp
            from eta[unfolded eta_closed_top_def, THEN spec, THEN spec, THEN spec, THEN mp[OF _ lr[unfolded lhs]], THEN mp[OF _ aa]]
            obtain ll rr yy where yy1: "length yy = n" and yy2: "distinct yy" 
              and yy3: "set yy \<inter> vars_rule (ll,rr) = {}" and  inst: "instance_rule (l,r) (ll,rr)" and llrr: "(Fun a (ll # map Var yy), Fun a (rr # map Var yy)) \<in> P" unfolding lhs[symmetric] 
              by auto
            from P llrr have hvf: "hvf_top a ?n (Fun a (ll # map Var yy))" by auto
            from yy3 have yl: "set yy \<inter> vars_term ll = {}" and yr: "set yy \<inter> vars_term rr = {}" unfolding vars_rule_def by auto
            from inst[unfolded instance_rule_def] obtain \<delta> where ll: "l = ll \<cdot> \<delta>" and rr: "r = rr \<cdot> \<delta>" by auto
            obtain \<gamma> where gamma: "\<gamma> = \<delta> \<circ>s \<sigma>" by simp
            let ?sig = "subst_extend \<gamma> (zip yy aft)"
            obtain \<tau> where tau: "?sig = \<tau>" by auto
            {
              fix t :: "('a,string)term"
              assume y: "set yy \<inter> vars_term t = {}"              
              hence one: "t \<cdot> ?sig = t \<cdot> \<gamma>" 
                using subst_extend_id[of "UNIV - set yy" "yy" t \<gamma> "aft"] by auto
              from yy1 n have n: "length aft = length yy" by simp
              have two: "map ((subst_extend \<gamma> (zip yy aft))) yy = aft"
              proof (rule nth_equalityI, simp add: n, simp del: subst_extend.simps, intro allI impI)
                fix i
                assume i: "i < length yy"
                from n set_zip[of yy aft] have zip: "set (zip yy aft) = {(yy ! i, aft ! i) | i. i < length yy}" by simp
                with i have yi: "(yy ! i, aft ! i) \<in> set (zip yy aft)" by auto
                show "(subst_extend \<gamma> (zip yy aft)) (yy ! i) = aft ! i"
                proof -
                  show ?thesis 
                  proof (cases "lookup (yy ! i) (zip yy aft)")
                    case None
                    from lookup_complete[OF None] yi show ?thesis by simp
                  next
                    case (Some ai)
                    from lookup_sound[OF Some] have "(yy ! i, ai) \<in> set (zip yy aft)" .
                    with zip obtain j where j: "j < length yy" and yij: "yy ! i = yy ! j" and ai: "ai = aft ! j" by auto
                    from nth_eq_iff_index_eq[OF yy2 i j] yij ai have "ai = aft ! i" by simp
                    thus ?thesis
                      by (simp add: Some)
                  qed
                qed
              qed
              from one two have "Fun a (t # map Var yy) \<cdot> ?sig = Fun a ((t \<cdot> \<gamma>) # aft)"
                by (simp add: o_def)
            } note id = this
            let ?tau = "map_subst_wa (\<lambda>f n. get_symbol sm f n 0) \<tau>"
            from id[OF yl] have l: "Fun f ss = Fun a (ll # map Var yy) \<cdot> ?sig" unfolding l ll gamma by auto
            from id[OF yr] have r: "u = Fun a (rr # map Var yy) \<cdot> ?sig" unfolding r rr gamma by auto            
            from llrr have llrr: "(?u (Fun a (ll # map Var yy)), ?u (Fun a (rr # map Var yy))) \<in> ?UP" 
              unfolding uncurry_top_trs_def by force
            have "(?u (Fun a (rr # map Var yy)) \<cdot> ?tau, ?u u) \<in> (CS ?US)^="
              unfolding r tau by (rule uncurry_top_subst[OF a])
            hence two: "(?u (Fun a (rr # map Var yy)) \<cdot> ?tau, ?u u) \<in> ?Rel" unfolding cs_union_distr using SNuu P' by auto              
            have SN: "SN_elt (rstep ?UR) (?u (Fun a (rr # map Var yy)) \<cdot> ?tau)"
              by (rule uncurry_top_subst_SN[OF inj ndefa SNu[unfolded r tau]])
            have "(?u (Fun f ss), ?u (Fun a (rr # map Var yy)) \<cdot> ?tau) \<in> CS ?UP"
              unfolding l tau uncurry_top_subst_hvf[OF hvf] 
              by (rule subst_closureI2, rule llrr, auto)
            with SN have one: "(?u (Fun f ss), ?u (Fun a (rr # map Var yy)) \<cdot> ?tau) \<in> ?Rel" unfolding cs_union_distr by auto
            from one two show ?thesis by auto
          qed
        qed
      qed
    qed
  qed
  with steps
  show ?case by auto
qed

lemma uncurrying_top_sound_dp: assumes hvflP: "\<And> l r. (l,r) \<in> P_eta \<Longrightarrow> hvf_top a (Suc n) l \<and> \<not> is_var r"
  and P: "P \<subseteq> P_eta"
  and P_eta: "(\<forall> (l,r) \<in> P. the (root r) \<noteq> a \<or> num_args r \<noteq> (Suc n) \<or> cap R (hd (args r)) \<noteq> HHole) \<or> (eta_closed_top a (Suc n) sm R P_eta \<and> CS (uncurry_of_top_sig a (Suc n) sm) \<subseteq> CS P')"
  and ndefa: "\<not> defined R (a,Suc n)"
  and ndef: "\<And> s f ts. (s,Fun f ts) \<in> P \<Longrightarrow> \<not> (defined R (f,length ts))"
  and inj: "inj_sig_map a (Suc n) sm"
  and finite: "finite_dpp (uncurry_top_trs a (Suc n) sm P_eta \<union> P',uncurry_below_trs sm R)" (is "finite_dpp (?P,?R)")
  shows "finite_dpp (P,R)"
unfolding finite_dpp_def
proof (clarify)
  fix s t \<sigma>
  assume chain: "min_ichain (P,R) s t \<sigma>"
  let ?n = "Suc n"
  let ?u = "uncurry_top a ?n sm"
  let ?s = "\<lambda> i. ?u (s i)"
  let ?t = "\<lambda> i. ?u (t i)"
  let ?us = "\<lambda> i. uncurry_top_subst sm (\<sigma> i)"
  let ?min = "{(s, t) . SN_elt (rstep (map_trs_wa (\<lambda>f n. get_symbol sm f n 0) R)) t}"
  let ?Pm = "CS ?P \<inter> ?min"
  let ?US = "uncurry_of_top_sig a (Suc n) sm"
  let ?Both = "?Pm \<union> rstep ?R"
  let ?BothRev = "rstep ?R \<union> ?Pm"
  let ?Rel = "?Both^* O ?Pm O ?Both^*"
  let ?RelRev = "?BothRev^* O ?Pm O ?BothRev^*"
  from inj_sig_map_a[OF inj] have a: "get_symbol sm a ?n 0 = a" by auto
  obtain S where S: "S = (\<lambda> i. (?s i \<cdot> ?us i))" by auto
  {
    fix i
    from chain P have inP: "(s i, t i) \<in> P_eta" and iP: "(s i, t i) \<in> P" by auto
    from chain P have inPs: "(s (Suc i), t (Suc i)) \<in> P_eta" by auto
    from inP have P: "(?s i, ?t i) \<in> ?P" unfolding uncurry_top_trs_def by auto
    from hvflP[OF inP] obtain f ts where ti: "t i = Fun f ts"  by (cases "t i", auto)
    from ndef[OF iP[unfolded ti]] have ndef: "\<not> defined R (f, length ts)" .
    from hvflP[OF inPs] obtain g ss where  ssi: "s (Suc i) = Fun g ss" by (cases "s (Suc i)", auto)
    from hvflP[OF inPs] ssi have hvf: "hvf_top a ?n (Fun g ss)" by auto
    have s_subst: "?u (s (Suc i) \<cdot> \<sigma> (Suc i)) = ?s (Suc i) \<cdot> ?us (Suc i)"  unfolding ssi by (rule uncurry_top_subst_hvf[OF hvf])
    from chain have "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (rstep R)^*"  and min: "SN_elt (rstep R) (t i \<cdot> \<sigma> i)" by (auto simp: minimal_cond_def)
    hence steps: "(Fun f (map (\<lambda> t. t \<cdot> \<sigma> i) ts), s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (rstep R)^*" 
      and minf: "SN_elt (rstep R) (Fun f (map (\<lambda> t. t \<cdot> \<sigma> i) ts))" unfolding ti by auto
    have t_subst_pre: "(?t i \<cdot> ?us i, ?u (t i \<cdot> \<sigma> i)) \<in> (CS ?P)^=" 
    proof (cases "hvf_top a ?n (t i)")
      case True
      show ?thesis unfolding uncurry_top_subst_hvf[OF True] by auto
    next
      case False
      hence "the (root (t i)) = a \<and> num_args (t i) = ?n \<and> cap R (hd (args (t i))) = HHole" unfolding ti
        by (cases "hd ts", auto)
      with iP P_eta have "CS ?US \<subseteq> CS P'" by blast
      hence subset: "(CS (uncurry_of_top_sig a ?n sm))^= \<subseteq> (CS ?P)^=" unfolding cs_union_distr by auto
      from uncurry_top_subst[OF a, of f ts "\<sigma> i"] subset 
      show ?thesis unfolding ti by auto
    qed
    with uncurry_top_preserves_SN[OF inj min] have t_subst: "(?t i \<cdot> ?us i, ?u (t i \<cdot> \<sigma> i)) \<in> (?Pm)^="  by auto
    hence t_subst: "(?t i \<cdot> ?us i, ?u (t i \<cdot> \<sigma> i)) \<in> (?Pm)^*"  by auto
    have t_subst: "(?t i \<cdot> ?us i, ?u (t i \<cdot> \<sigma> i)) \<in> (rstep ?R \<union> ?Pm)^*" 
      by (rule set_mp[OF rtrancl_mono t_subst], auto)
    have R: "(?u (t i \<cdot> \<sigma> i), ?u (s (Suc i) \<cdot> \<sigma> (Suc i))) \<in> (rstep ?R \<union> ?Pm)^*"
      unfolding ti subst_apply_term.simps
    proof (rule uncurry_top_steps[OF inj _ ndefa steps _ _ minf], simp add: ndef, force simp: hvflP)      
      show "f \<noteq> a \<or> length (map (\<lambda> t. t \<cdot> \<sigma> i) ts) \<noteq> ?n \<or> cap R (hd (map (\<lambda> t. t \<cdot> \<sigma> i) ts)) \<noteq> HHole \<or> (eta_closed_top a ?n sm R P_eta \<and> CS ?US \<subseteq> CS P')"
      proof (cases "f \<noteq> a \<or> length ts \<noteq> ?n \<or> (eta_closed_top a ?n sm R P_eta \<and> CS ?US \<subseteq> CS P')")
        case True
        thus ?thesis by simp
      next
        case False
        from False have "length ts = ?n" by simp
        then obtain t1 tts where ts: "ts = t1 # tts" and n: "n = length tts" by (cases ts, auto)
        from False P_eta iP[unfolded ti ts] n have nhol: "cap R t1 \<noteq> HHole" by auto
        have "cap R (t1 \<cdot> \<sigma> i) \<noteq> HHole"
        proof 
          assume "cap R (t1 \<cdot> \<sigma> i) = HHole"
          with capInstanceSubset[of R t1 "\<sigma> i"] nhol show False
            by (cases "cap R t1", auto)
        qed
        thus ?thesis unfolding ts by simp
      qed
    qed
    from R t_subst have steps: "(?t i \<cdot> ?us i, ?s (Suc i) \<cdot> ?us (Suc i)) \<in> (rstep ?R \<union> ?Pm)^*"
      unfolding s_subst by auto
    from P have P: "(?s i \<cdot> ?us i, ?t i \<cdot> ?us i) \<in> CS ?P"
      by (rule subst_closureI2, auto)
    have SN: "SN_elt (rstep ?R) (?t i \<cdot> ?us i)"
      by (rule uncurry_top_subst_SN[OF inj ndefa min])
    with P have P: "(?s i \<cdot> ?us i, ?t i \<cdot> ?us i) \<in> ?Pm" by auto 
    with steps have "(?s i \<cdot> ?us i, ?s (Suc i) \<cdot> ?us (Suc i)) \<in> ?RelRev" by auto
    hence "(S i, S (Suc i)) \<in> ?Rel" unfolding S Un_commute[of "?Pm" "rstep ?R"] .
  } note main = this
  hence "\<not> SN ?Rel" unfolding SN_defs by blast
  with SN_imp_SN_trancl[OF finite[unfolded DP_Framework.finite_dpp_iff_rel_SN rel_SN_def], unfolded rel_step_trancl_conv]
  show False by simp
qed


section {* automation *}

definition eta_closed_rules :: "'f \<Rightarrow> 'f sig_map \<Rightarrow> ('f :: show,'v :: show)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> shows check"
where "eta_closed_rules a sm R R' \<equiv> check_allm (\<lambda> (l,r). 
        case aarity_term a sm l of 
            Some (Suc aa) \<Rightarrow> check (list_ex 
                  (\<lambda> (lll,rrr). case (lll,rrr) of (Fun f [ll, Var x],Fun g [rr, Var y]) \<Rightarrow> f = a \<and> g = a \<and> x = y \<and> x \<notin> set (vars_rule_list (ll,rr)) \<and> instance_rule (l,r) (ll,rr) | _ \<Rightarrow> False)
                                   R') (shows ''eta expansion of '' +@+ shows_rule (l,r) +@+ shows '' missing'') 
          | _             \<Rightarrow> OK) 
      R"


lemma eta_closed_rules: "isOK(eta_closed_rules a sm R R') = eta_closed a sm (set R) (set R')"
proof -
  let ?Q = "\<lambda> l r. (\<lambda> (lll,rrr). case (lll,rrr) of (Fun f [ll, Var x],Fun g [rr, Var y]) \<Rightarrow> f = a \<and> g = a \<and> x = y \<and> x \<notin> set (vars_rule_list (ll,rr)) \<and> instance_rule (l,r) (ll,rr) | _ \<Rightarrow> False)"
  show ?thesis
  proof
    assume ok: "isOK(eta_closed_rules a sm R R')"
    show "eta_closed a sm (set R) (set R')"
      unfolding eta_closed_def
    proof (intro allI impI)
      fix l r aa
      assume lr: "(l,r) \<in> set R" and "aarity_term a sm l = Some aa" and "0 < aa"
      then obtain aa where aa: "aarity_term a sm l = Some (Suc aa)" by (cases aa, auto)
      let ?P = "?Q l r"
      from ok[unfolded eta_closed_rules_def, simplified, THEN bspec[OF _ lr], simplified, unfolded aa, simplified, unfolded list_ex_iff]
      obtain lll rrr where lllrrr: "(lll,rrr) \<in> set R'" and P: "?P (lll,rrr)" by auto
      from P obtain f ls where lll: "lll = Fun f ls" by (cases lll, auto)
      from P[unfolded lll] obtain ll ls where lll: "lll = Fun f (ll # ls)" unfolding lll by (cases ls, auto)
      from P[unfolded lll] obtain lt ls where lll: "lll = Fun f (ll # lt # ls)" unfolding lll by (cases ls, auto)
      from P[unfolded lll] obtain x where lll: "lll = Fun f (ll # Var x # ls)" unfolding lll by (cases lt, auto)  
      from P[unfolded lll] have lll: "lll = Fun f [ll, Var x]" unfolding lll by (cases ls, auto)  
      note P = P[unfolded lll]
      from P obtain g ls where rrr: "rrr = Fun g ls" by (cases rrr, auto)
      from P[unfolded rrr] obtain rr ls where rrr: "rrr = Fun g (rr # ls)" unfolding rrr by (cases ls, auto)
      from P[unfolded rrr] obtain lt ls where rrr: "rrr = Fun g (rr # lt # ls)" unfolding rrr by (cases ls, auto)
      from P[unfolded rrr] obtain y where rrr: "rrr = Fun g (rr # Var y # ls)" unfolding rrr by (cases lt, auto)  
      from P[unfolded rrr] have rrr: "rrr = Fun g [rr, Var y]" unfolding rrr by (cases ls, auto)
      from P[unfolded rrr] lll rrr have lll: "lll = Fun a [ll, Var x]" and rrr: "rrr = Fun a [rr, Var x]" and x: "x \<notin> set (vars_rule_list (ll,rr))" and inst: "instance_rule (l,r) (ll,rr)" by auto
      show "\<exists> ll rr y. instance_rule (l,r) (ll,rr) \<and> y \<notin> vars_rule (ll,rr) \<and> (Fun a [ll,Var y], Fun a [rr, Var y]) \<in> set R'"
        by (rule exI[of _ ll], rule exI[of _ rr], rule exI[of _ x], intro conjI, rule inst, rule x[unfolded vars_rule_list_sound],
          rule lllrrr[unfolded lll rrr])
    qed
  next
    assume eta: "eta_closed a sm (set R) (set R')"
    show "isOK (eta_closed_rules a sm R R')"
      unfolding eta_closed_rules_def isOK_check_allm
    proof (intro ballI)
      fix lr
      assume mem: "lr \<in> set R"
      obtain l r where lr: "lr = (l,r)" by (cases lr, auto)
      let ?P = "?Q l r"
      let ?R = " (\<lambda> (l,r). 
        case aarity_term a sm l of 
        Some (Suc aa) \<Rightarrow> check (list_ex (?Q l r) R') (shows ''eta expansion of '' +@+ shows_rule (l,r) +@+ shows '' missing'') 
        | _             \<Rightarrow> OK) "
      show "isOK (?R lr)"
      proof (cases "aarity_term a sm l")
        case None
        thus ?thesis unfolding lr by auto
      next
        case (Some aa)
        show ?thesis 
        proof (cases aa)
          case 0
          thus ?thesis using Some unfolding lr by auto
        next
          case (Suc aa)
          from eta[unfolded eta_closed_def, THEN spec, of l, THEN spec, of r, THEN spec, THEN mp[OF _ mem[unfolded lr]], THEN mp[OF _ Some],
          THEN mp, unfolded Suc]
          obtain ll rr y where inst: "instance_rule (l,r) (ll,rr)" and y:"y \<notin> vars_rule (ll,rr)" 
            and mem: "(Fun a [ll, Var y], Fun a [rr, Var y]) \<in> set R'" (is "?lr \<in> set R'")
            by auto
          have "(?Q l r) ?lr" 
            by (simp add: y inst mem)
          with mem have "list_ex (?Q l r) R'"
            unfolding list_ex_iff by blast
          thus ?thesis using Some Suc unfolding lr by auto
        qed
      qed
    qed
  qed
qed

definition eta_closed_top_rules :: "'f \<Rightarrow> nat \<Rightarrow> 'f sig_map \<Rightarrow> ('f :: show,'v :: show)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> shows check"
where "eta_closed_top_rules a n sm R P \<equiv> check_allm (\<lambda> (l,r). 
        case l of 
            Fun ff ls \<Rightarrow> check (aarity sm ff (length ls) = 0 \<or> list_ex 
                  (\<lambda> (lll,rrr). case (lll,rrr) of (Fun f (ll # yy),Fun g (rr # zz)) \<Rightarrow> f = a \<and> g = a \<and> zz = yy \<and> length yy = n - 1 \<and> distinct yy \<and> list_all is_var yy \<and> list_inter (map the_var yy) (vars_rule_list (ll,rr)) = [] \<and> instance_rule (l,r) (ll,rr) | _ \<Rightarrow> False)
                                   P) (shows ''eta expansion of '' +@+ shows_rule (l,r) +@+ shows '' missing'') 
          | _             \<Rightarrow> OK) 
      R"


lemma eta_closed_top_rules: "isOK(eta_closed_top_rules a n sm R P) = eta_closed_top a n sm (set R) (set P)"
proof -
  let ?Q = "\<lambda> l r. (\<lambda> (lll,rrr). case (lll,rrr) of (Fun f (ll # yy),Fun g (rr # zz)) \<Rightarrow> f = a \<and> g = a \<and> zz = yy \<and> length yy = n - 1 \<and> distinct yy \<and> list_all is_var yy \<and> list_inter (map the_var yy) (vars_rule_list (ll,rr)) = [] \<and> instance_rule (l,r) (ll,rr) | _ \<Rightarrow> False)"
  let ?R = "\<lambda> l r.  isOK(case l of 
            Fun ff ls \<Rightarrow> check (aarity sm ff (length ls) = 0 \<or> list_ex (?Q l r) P) (shows ''eta expansion of '' +@+ shows_rule (l,r) +@+ shows '' missing'') 
          | _             \<Rightarrow> OK)" 
  let ?Var = "Var :: 'b \<Rightarrow> ('a,'b)term"
  show ?thesis
  proof 
    assume ok: "isOK(eta_closed_top_rules a n sm R P)" 
    show "eta_closed_top a n sm (set R) (set P)"
      unfolding eta_closed_top_def
    proof (intro allI impI)
      fix ff l r
      assume lr: "(Fun ff l,r) \<in> set R" and aa: "aarity sm ff (length l) \<noteq> 0"
      let ?l = "Fun ff l"
      let ?P = "?Q ?l r"
      from ok[unfolded eta_closed_top_rules_def, simplified, THEN bspec[OF _ lr], unfolded list_ex_iff list_inter] aa
      obtain lll rrr where lllrrr: "(lll,rrr) \<in> set P" and P: "?P (lll,rrr)" by auto
      from P obtain f ls where lll: "lll = Fun f ls" by (cases lll, auto)
      from P[unfolded lll] obtain ll yy where lll: "lll = Fun f (ll # yy)" unfolding lll by (cases ls, auto)
      note P = P[unfolded lll]
      from P obtain g rs where rrr: "rrr = Fun g rs" by (cases rrr, auto)
      from P[unfolded rrr] obtain rr rs where rrr: "rrr = Fun g (rr # rs)" unfolding rrr by (cases rs, auto)
      from P[unfolded rrr] lll rrr have lll: "lll = Fun a (ll # yy)" and rrr: "rrr = Fun a (rr # yy)" 
        and len: "length yy = n - 1" and yy1: "distinct yy" and yy2: "list_all is_var yy" and yy3: "list_inter (map the_var yy) (vars_rule_list (ll,rr)) = []"
        and inst: "instance_rule (?l,r) (ll,rr)" by auto
      have id: "map ?Var (map the_var yy) = yy" by (rule Var_the_var_id[OF yy2])
      have "distinct (map ?Var (map the_var yy))" unfolding id  by (rule yy1)
      from this[unfolded distinct_map[of ?Var]] have yy1: "distinct (map the_var yy)" ..
      from arg_cong[OF yy3, of set] have yy3: "set (map the_var yy) \<inter> vars_rule (ll,rr) = {}" unfolding list_inter vars_rule_list_sound by simp      
      show "\<exists> ll rr ys. length ys = n - 1 \<and> distinct ys \<and> set ys \<inter> vars_rule (ll,rr) = {} \<and> instance_rule (?l,r) (ll,rr) \<and> (Fun a (ll # map Var ys), Fun a (rr # map Var ys)) \<in> set P"
        by (rule exI[of _ ll], rule exI[of _ rr], rule exI[of _ "map the_var yy"], simp only: inst yy3 yy1 id, simp add: len lllrrr[unfolded lll rrr]) 
    qed
  next
    assume eta: "eta_closed_top a n sm (set R) (set P)"
    {
      fix l r
      assume mem: "(l,r) \<in> set R"
      have "?R l r"
      proof (cases l)	
        case (Var x) 
        thus ?thesis by auto
      next
        case (Fun ff ls)
        show ?thesis 
        proof (cases "aarity sm ff (length ls)")
          case 0
          show ?thesis unfolding Fun by (simp add: 0)
        next
          case (Suc m)
          from eta[unfolded eta_closed_top_def, THEN spec, THEN spec, THEN spec, THEN mp[OF _ mem[unfolded Fun]], THEN mp, unfolded Suc]
          obtain ll rr ys where len: "length ys = n - 1" and dist: "distinct ys" and ys: "set ys \<inter> vars_rule (ll,rr) = {}" and 
            inst: "instance_rule (l,r) (ll,rr)" 
            and mem: "(Fun a (ll # map ?Var ys), Fun a (rr # map ?Var ys)) \<in> set P" (is "?lr \<in> set P")
            unfolding Fun
            by auto
          from ys have "set ys \<inter> set (vars_rule_list (ll,rr)) = {}" unfolding vars_rule_list_sound .
          hence ys: "list_inter ys (vars_rule_list (ll,rr)) = []" unfolding list_inter[symmetric] 
            by (cases "list_inter ys (vars_rule_list (ll,rr))", auto)
          from dist have dist: "distinct (map Var ys)"
            unfolding distinct_map inj_on_def by auto
          have "(?Q l r) ?lr" 
            by (simp add: ys inst mem len o_def dist list_all_iff) 
          with mem have "list_ex (?Q l r) P"
            unfolding list_ex_iff by blast
          thus ?thesis unfolding Fun
            by (simp add: Suc list_ex_iff)
        qed
      qed
    } note main = this
    show "isOK (eta_closed_top_rules a n sm R P)"
      unfolding eta_closed_top_rules_def isOK_check_allm
      by (intro ballI, insert main, auto)
  qed
qed




definition only_eta_rules :: "('f :: show,'v :: show)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> shows check"
where "only_eta_rules E R_eta \<equiv> check_allm (\<lambda> (l,r). 
              check ( 
                  (case (l,r) of (Fun f ls, Fun g rs) \<Rightarrow> f = g \<and> length ls = 2 \<and> length rs = 2 \<and> ls ! 1 = rs ! 1 \<and> list_ex (\<lambda> (l',r'). instance_rule (hd ls, hd rs) (l',r')) R_eta | _ \<Rightarrow> False))
                   (shows ''rule '' +@+ shows_rule (l,r) +@+ shows '' is not an (eta-expanded) original rule'')
        ) E"

lemma only_eta_rules: assumes inf: "infinite (UNIV :: ('v :: show) set)"
  shows "isOK(only_eta_rules (E :: ('f :: show,'v :: show)rules) R_eta) = only_eta (set E) (set R_eta)"
proof -
  let ?P = "\<lambda> l r. (
                  (case (l,r) of (Fun f ls, Fun g rs) \<Rightarrow> f = g \<and> length ls = 2 \<and> length rs = 2 \<and> ls ! 1 = rs ! 1 \<and> list_ex (\<lambda> (l',r'). instance_rule (hd ls, hd rs) (l',r')) R_eta | _ \<Rightarrow> False))"
  show ?thesis
  proof
    assume ok: "isOK(only_eta_rules E R_eta)"
    show "only_eta (set E) (set R_eta)"
      unfolding only_eta_def
    proof(intro allI impI)
      fix l r :: "('f,'v)term"
      assume lr: "(l,r) \<in> set E"      
      with ok[unfolded only_eta_rules_def] have P: "?P l r" by auto
      from P obtain f ls where l: "l = Fun f ls" by (cases l, auto)
      from P obtain g rs where r: "r = Fun g rs" unfolding l by (cases r, auto)
      from P l r obtain l' r' where l: "l = Fun f ls" and r: "r = Fun f rs" and ls: "length ls = 2" and rs: "length rs = 2" and ls1: "ls ! 1 = rs ! 1" and l'r': "(l',r') \<in> set R_eta" and inst: "instance_rule (hd ls, hd rs) (l',r')" unfolding list_ex_iff by auto
      obtain l1 l2 where ls: "ls = [l1,l2]" by (rule obtain_length2[OF ls])
      obtain r1 r2 where rs: "rs = [r1,r2]" by (rule obtain_length2[OF rs])
      from ls rs l r ls1 have l: "l = Fun f [l1,l2]" and r: "r = Fun f [r1,l2]" by auto      
      from inst[unfolded ls rs, simplified] obtain \<sigma> where l1: "l1 = l' \<cdot> \<sigma>" and r1: "r1 = r' \<cdot> \<sigma>" unfolding instance_rule_def by auto
      from infinite_imp_elem[OF Diff_infinite_finite[OF finite_set[of "vars_rule_list (l',r')"] inf]]
      obtain x where xl': "x \<notin> vars_term l'" and xr': "x \<notin> vars_term r'" unfolding vars_rule_list_sound vars_rule_def by auto
      let ?sig = "subst_extend \<sigma> (zip [x] [l2])"
      note subst_extend_id = subst_extend_id[of "UNIV - {x}" "[x]" _ \<sigma> "[l2]", simplified]
      from xl' subst_extend_id[of l'] have l': "l' \<cdot> ?sig = l' \<cdot> \<sigma>" by auto
      from xr' subst_extend_id[of r'] have r': "r' \<cdot> ?sig = r' \<cdot> \<sigma>" by auto
      show "\<exists> a t l' r'. (l',r') \<in> set R_eta \<and> instance_rule (l,r) (Fun a [l',t], Fun a [r',t])"
        by (rule exI[of _ f], rule exI[of _ "Var x"], intro exI, intro conjI, rule l'r', unfold instance_rule_def l r l1 r1, rule exI[of _ ?sig], simp
           add: l'[symmetric] r'[symmetric])
    qed
  next
    assume eta: "only_eta (set E) (set R_eta)"
    show "isOK(only_eta_rules E R_eta)"
      unfolding only_eta_rules_def isOK_check_allm
    proof
      fix lr
      assume mem: "lr \<in> set E"
      obtain l r where lr: "lr = (l,r)" by (cases lr, auto)
      let ?Q = "\<lambda> (l,r). check (?P l r)
                   (shows ''rule '' +@+ shows_rule (l,r) +@+ shows '' is not an (eta-expanded) original rule'')"
      have "?P l r"
      proof -
        from eta[unfolded only_eta_def, THEN spec, THEN spec, THEN mp[OF _ mem[unfolded lr]]]
        obtain a t l' r' where l'r': "(l',r') \<in> set R_eta" and inst: "instance_rule (l,r) (Fun a [l',t], Fun a [r', t])" 
          by auto
        then obtain \<sigma> where l: "l =  Fun a [l',t] \<cdot> \<sigma>" and r: "r = Fun a [r', t] \<cdot> \<sigma>"
          unfolding instance_rule_def by auto
        show "?P l r"
          by (unfold l r list_ex_iff, simp, rule bexI[OF _ l'r'], unfold instance_rule_def, rule exI[of _ \<sigma>], simp)
      qed
      thus "isOK(?Q lr)" unfolding lr by auto
    qed
  qed
qed


definition uncurry_rules :: "'f \<Rightarrow> 'f sig_map \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rules"
  where "uncurry_rules a sm \<equiv> map (\<lambda> (l,r). (uncurry_term a sm l, uncurry_term a sm r))"

lemma uncurry_rules: "set (uncurry_rules a sm R) = uncurry_trs a sm (set R)"
  unfolding uncurry_rules_def uncurry_trs_def by auto

type_synonym 'f sig_map_list = "(('f \<times> nat) \<times> 'f list) list"

definition sig_list_to_sig_map :: "'f \<Rightarrow> 'f sig_map_list \<Rightarrow> ('f sig_map_list \<Rightarrow> 'f \<Rightarrow> nat \<Rightarrow> 'f)  \<Rightarrow> 'f sig_map"
where "sig_list_to_sig_map a sml fmap \<equiv> let fm = fmap sml in (\<lambda> f n. case lookup (f,n) sml of None \<Rightarrow> [fm f n] | Some xs \<Rightarrow> if xs = [] then [fm f n] else xs)"


definition uncurry_of_sig_list :: "'f \<Rightarrow> 'f sig_map_list \<Rightarrow> 'f sig_map \<Rightarrow> ('f,string)rules"
  where "uncurry_of_sig_list a sml sm \<equiv> concat (map (\<lambda> ((f,n),_). (
                  let g = get_symbol sm f n in
                    map (\<lambda> i. (Fun a [generate_f_xs (g i) (n+i), Var (generate_var (n+i))], generate_f_xs (g (Suc i)) (n + Suc i))) [0 ..< aarity sm f n]
           )) sml)"

lemma uncurry_of_sig_list: "set (uncurry_of_sig_list a sml (sig_list_to_sig_map a sml fmap)) = uncurry_of_sig a (sig_list_to_sig_map a sml fmap)" (is "?l = ?r")
proof -
  let ?sm = "sig_list_to_sig_map a sml fmap"
  {
    fix l r
    assume "(l,r) \<in> ?l"
    then obtain f n list where mem: "((f,n),list) \<in> set sml" 
      and mem2: "(l,r) \<in>  set (map (\<lambda>i. (Fun a
                                   [generate_f_xs (get_symbol ?sm f n i) (n + i),
                                    Var (generate_var (n + i))],
                                  generate_f_xs  (get_symbol ?sm f n (Suc i)) (n + Suc i)))
                         [0..<aarity ?sm f n])"
      unfolding uncurry_of_sig_list_def Let_def by auto
    from mem2
    obtain i where i: "i < aarity ?sm f n" and l: "l = Fun a [generate_f_xs (get_symbol ?sm f n i) (n + i),Var (generate_var (n + i))]"
      and r: "r = generate_f_xs  (get_symbol ?sm f n (Suc i)) (n + Suc i)" by auto
    have "(l,r) \<in> ?r" unfolding l r uncurry_of_sig_def using i by auto
  } note l_imp_r = this
  {
    fix l r
    assume "(l,r) \<in> ?r"
    then obtain f n i where i: "i < aarity ?sm f n" and l: "l = Fun a [generate_f_xs (get_symbol ?sm f n i) (n + i),Var (generate_var (n + i))]"
      and r: "r = generate_f_xs  (get_symbol ?sm f n (Suc i)) (n + Suc i)" unfolding uncurry_of_sig_def by auto
    from i[unfolded aarity_def sig_list_to_sig_map_def] obtain list where lookup: "lookup (f,n) sml = Some list" 
      by (cases "lookup (f,n) sml", auto simp: Let_def)
    from lookup_sound[OF lookup] have mem: "((f,n),list) \<in> set sml" by auto
    have "(l,r) \<in> ?l" unfolding uncurry_of_sig_list_def
      by (simp, rule bexI[OF _ mem], simp add: Let_def, unfold l r, insert i, auto)
  }
  with l_imp_r show ?thesis by auto
qed  

(* uncurry info consists of
   1. uncurry symbol
   2. symbol mapping
   3. uncurrying rules
   4. additional rules in eta expansion
*)
type_synonym ('f,'v) uncurry_info = "'f \<times> 'f sig_map_list \<times> ('f,'v)rules \<times> ('f,'v)rules"    
      
fun uncurry_tt :: "('f,string)uncurry_info \<Rightarrow> ('f,string)rules \<Rightarrow> ('f :: show,string)rules \<Rightarrow> shows check"
where "uncurry_tt (a,sml,U,E) R R' = do {
           let sm = sig_list_to_sig_map a sml (\<lambda> _ f _. f);
           let R_eta = E @ R;
           check_allm (\<lambda> (l,r). check (hvf_term a l) (shows ''head variable in lhs '' +@+ shows_term l +@+ shows '' not allowed'')) R_eta;
           eta_closed_rules a sm R_eta R_eta;
           check_subseteq (uncurry_rules a sm R_eta) R'
              <+? (\<lambda> lr. shows ''uncurried rule '' +@+ shows_rule lr +@+ shows '' is missing'');
           check_CS_subseteq (uncurry_of_sig_list a sml sm) R'
              <+? (\<lambda> lr. shows ''uncurry rule '' +@+ shows_rule lr +@+ shows '' is missing'')
   }"

definition uncurry_eta_split :: "('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rules \<times> ('f,'v)rules"
  where "uncurry_eta_split Eboth Rtest \<equiv> let 
       test = (\<lambda> (l,r). num_args l > 0 \<and> num_args r > 0 \<and> list_ex (eq_rule_mod_vars (hd (args l), hd (args r))) Rtest)
     in partition test Eboth"

fun uncurry_rel_tt :: "('f,string)uncurry_info \<Rightarrow> ('f,string)rules \<Rightarrow> ('f,string)rules \<Rightarrow> ('f,string)rules \<Rightarrow> ('f :: show,string)rules \<Rightarrow> shows check"
where "uncurry_rel_tt (a,sml,U,Eb) R Rw R' Rw' = do {
           let (E,Ew) = uncurry_eta_split Eb R;
           let sm = sig_list_to_sig_map a sml (\<lambda> _ f _. f);
           let R_eta = E @ R;
           let Rw_eta = Ew @ Rw;
           let Rb_eta = R_eta @ Rw_eta;
           check_allm (\<lambda> (l,r). check (hvf_term a l) (shows ''head variable in lhs '' +@+ shows_term l +@+ shows '' not allowed'')) R_eta;
           check_allm (\<lambda> (l,r). check (hvf_term a l) (shows ''head variable in lhs '' +@+ shows_term l +@+ shows '' not allowed'')) Rw_eta;
           eta_closed_rules a sm R_eta R_eta;
           eta_closed_rules a sm Rb_eta Rb_eta;
           check_allm (\<lambda> (l,r). check (\<not> is_var l) (shows ''lhs must not be a variable in rule '' +@+ shows_rule (l,r))) Rw_eta;
           check_subseteq (uncurry_rules a sm R_eta) R'
              <+? (\<lambda> lr. shows ''uncurried rule '' +@+ shows_rule lr +@+ shows '' is missing'');
           check_subseteq (uncurry_rules a sm Rw_eta) Rw'
              <+? (\<lambda> lr. shows ''uncurried rule '' +@+ shows_rule lr +@+ shows '' is missing'');
           check_CS_subseteq (uncurry_of_sig_list a sml sm) Rw'
              <+? (\<lambda> lr. shows ''uncurry rule '' +@+ shows_rule lr +@+ shows '' is missing'')
   }"

declare hvf_term.simps[simp del]

lemma uncurry_tt: "sound_tt_impl (\<lambda> R R'. isOK(uncurry_tt info R R'))"
proof (cases info)
  case (fields a sml U E)
  show ?thesis unfolding fields sound_tt_impl_def
  proof (intro allI impI)
    fix R R'
    assume ok: "isOK(uncurry_tt (a,sml,U,E) R R') \<and> SN (rstep (set R'))"
    let ?sm = "sig_list_to_sig_map a sml (\<lambda> _ f _. f)"
    let ?R_eta = "E @ R"
    let ?U = "uncurry_trs a ?sm (set ?R_eta)"
    let ?S = "uncurry_of_sig a ?sm"
    obtain R_eta where R_eta: "?R_eta = R_eta" by simp
    from ok[unfolded uncurry_tt.simps Let_def]
    have hvf: "\<And> l r. (l,r) \<in> set ?R_eta \<Longrightarrow> hvf_term a l" 
      and eta: "eta_closed a ?sm (set ?R_eta) (set ?R_eta)"
      and U: "?U \<subseteq> set R'"
      and S: "CS ?S \<subseteq> CS (set R')"
      by (auto simp: R_eta eta_closed_rules Let_def  uncurry_rules uncurry_of_sig_list)
    from R_eta have R_eta: "set R \<subseteq> set ?R_eta" by auto 
    from CC_monotone[OF S, unfolded rstep_CC_CS[symmetric]] rstep_mono[OF U]
    have R': "rstep (?U \<union> ?S) \<subseteq> rstep (set R')" unfolding rstep_union by auto
    from SN_subset[OF ok[THEN conjunct2] R'] have
      SN: "SN (rstep (?U \<union> ?S))" by auto
    show "SN (rstep (set R))"
      by (rule uncurrying_sound[OF hvf R_eta eta SN])
  qed
qed

lemma uncurry_rel_tt: "sound_rel_tt_impl (\<lambda> (R,Rw) (R',Rw'). isOK(uncurry_rel_tt info R Rw R' Rw'))"
proof (cases info)
  case (fields a sml U Eb)
  show ?thesis unfolding sound_rel_tt_impl_def fields
  proof (intro allI impI, unfold split_def fst_conv snd_conv rel_SN_conv[symmetric])
    fix R Rw R' Rw'
    assume ok: "isOK(uncurry_rel_tt (a,sml,U,Eb) R Rw R' Rw')" and SN: "rel_SN (rstep (set R'), rstep (set Rw'))"
    obtain E Ew where "uncurry_eta_split Eb R = (E,Ew)" by (cases "uncurry_eta_split Eb R", auto)
    note ok =  ok[unfolded uncurry_rel_tt.simps split Let_def this]
    let ?sm = "sig_list_to_sig_map a sml (\<lambda> _ f _. f)"
    let ?R_eta = "E @ R"
    let ?Rw_eta = "Ew @ Rw"
    let ?U = "uncurry_trs a ?sm (set ?R_eta)"
    let ?Uw = "uncurry_trs a ?sm (set ?Rw_eta)"
    let ?S = "uncurry_of_sig a ?sm"
    obtain R_eta where R_eta: "?R_eta = R_eta" by simp
    obtain Rw_eta where Rw_eta: "?Rw_eta = Rw_eta" by simp
    from ok
    have hvf1: "\<And> l r. (l,r) \<in> set ?R_eta \<Longrightarrow> hvf_term a l" 
      and hvf2: "\<And> l r. (l,r) \<in> set ?Rw_eta \<Longrightarrow> hvf_term a l" 
      and eta1: "eta_closed a ?sm (set ?R_eta) (set ?R_eta)"
      and eta2: "eta_closed a ?sm (set ?R_eta \<union> set ?Rw_eta) (set ?R_eta \<union> set ?Rw_eta)"
      and var: "\<And> l r. (l,r) \<in> set ?Rw_eta \<Longrightarrow> \<not> is_var l"
      and U: "?U \<subseteq> set R'"
      and Uw: "?Uw \<subseteq> set Rw'"
      and S: "CS ?S \<subseteq> CS (set Rw')"
      by (auto simp: R_eta Rw_eta eta_closed_rules Let_def  uncurry_rules uncurry_of_sig_list)
    {
      fix x r
      have "(Var x,r) \<notin> set ?Rw_eta"
        using var[of "Var x" r] by auto
    } note var = this
    from R_eta have R_eta: "set R \<subseteq> set ?R_eta" by auto 
    from Rw_eta have Rw_eta: "set Rw \<subseteq> set ?R_eta \<union> set ?Rw_eta" by auto
    from CC_monotone[OF S, unfolded rstep_CC_CS[symmetric]] rstep_mono[OF Uw]
    have Rw': "rstep (?Uw \<union> ?S) \<subseteq> rstep (set Rw')" unfolding rstep_union by auto
    from rstep_mono[OF U] have R': "rstep ?U \<subseteq> rstep (set R')" .
    show "rel_SN (rstep (set R), rstep (set Rw))"
      by (rule uncurrying_rel_sound[OF hvf1 hvf2 R_eta eta1 Rw_eta eta2 var rel_SN_mono[OF R' Rw' SN]])
  qed
qed


fun uncurry_proc :: "('f,string)uncurry_info \<Rightarrow> ('f sig_map_list \<Rightarrow> 'f \<Rightarrow> nat \<Rightarrow> 'f) \<Rightarrow> ('f \<Rightarrow> nat \<Rightarrow> 'f sig_map_list \<Rightarrow> shows check) \<Rightarrow> ('f,string)rules \<Rightarrow> ('f,string)rules \<Rightarrow> ('f,string)rules \<Rightarrow> ('f :: show,string)rules \<Rightarrow> ('f,string)rule_map \<Rightarrow> shows check"
where "uncurry_proc (a,sml,_,E) fmap check_inj P R P' R' rm = do {
           let sm = sig_list_to_sig_map a sml fmap;
           let R_eta = E @ R;
           let U = uncurry_rules a sm R_eta;
           let S = uncurry_of_sig_list a sml sm;
           only_eta_rules E R_eta;
           check_inj a 2 sml;
           check_allm (\<lambda> (l,r). check (hvf_term a l) (shows ''head variable in lhs '' +@+ shows_term l +@+ shows '' not allowed'')) P;
           check_allm (\<lambda> (l,r). check (hvf_term a l) (shows ''head variable in lhs '' +@+ shows_term l +@+ shows '' not allowed'')) R_eta;
           eta_closed_rules a sm R_eta R_eta;
           check_subseteq (uncurry_rules a sm P) P'
              <+? (\<lambda> lr. shows ''uncurried pair '' +@+ shows_rule lr +@+ shows '' is missing'');
           check_subseteq U R'
              <+? (\<lambda> lr. shows ''uncurried rule '' +@+ shows_rule lr +@+ shows '' is missing'');
           check_CS_subseteq S R'
              <+? (\<lambda> lr. shows ''uncurry rule '' +@+ shows_rule lr +@+ shows '' is missing'');
           check_CS_subseteq R' (U @ S)
              <+? (\<lambda> lr. shows ''additional rule '' +@+ shows_rule lr +@+ shows '' is not allowed'')
   }"

lemma uncurry_proc: assumes check_inj: "isOK(check_inj a 2 sml) \<Longrightarrow> inj_sig_map a 2 (sig_list_to_sig_map a sml fmap)"
  shows "sound_proc_impl (\<lambda> P R rm P' R'. isOK(uncurry_proc (a,sml,U,E) fmap check_inj P R P' R' rm))"
  unfolding sound_proc_impl_def
proof (intro allI impI)
  fix P R P' R'
  assume ok: "isOK(uncurry_proc (a,sml,U,E) fmap check_inj P R P' R' (computeRuleMap R))" and finite: "finite_dpp (set P', set R')"
  let ?sm = "sig_list_to_sig_map a sml fmap"
  let ?R_eta = "E @ R"
  obtain R_eta where R_eta: "?R_eta = R_eta" by simp
  let ?UP = "uncurry_trs a ?sm (set P)"
  let ?U = "uncurry_trs a ?sm (set ?R_eta)"
  let ?S = "uncurry_of_sig a ?sm"
  from ok[unfolded uncurry_proc.simps Let_def]
  have 
     hvf: "\<And> l r. (l,r) \<in> set ?R_eta \<Longrightarrow> hvf_term a l" 
    and hvfP: "\<And> l r. (l,r) \<in> set P \<Longrightarrow> hvf_term a l" 
    and eta: "eta_closed a ?sm (set ?R_eta) (set ?R_eta)"
    and only: "isOK(only_eta_rules E ?R_eta)"
    and U: "?U \<subseteq> set R'"
    and UP: "?UP \<subseteq> set P'"
    and S: "CS ?S \<subseteq> CS (set R')"
    and inj: "isOK(check_inj a 2 sml)"
    and R': "CS (set R') \<subseteq> CS (?U \<union> ?S)"
    by (auto simp: R_eta eta_closed_rules uncurry_rules uncurry_of_sig_list)
  have only: "only_eta (set E) (set ?R_eta)" 
  proof -
    have inf: "infinite (UNIV :: string set)" by (rule infinite_UNIV_listI)
    show ?thesis using only[unfolded only_eta_rules[OF inf]] .
  qed
  from hvf hvfP have hvf: "\<And> l r. (l,r) \<in> set ?R_eta \<union> set P \<Longrightarrow> hvf_term a l" by blast
  from CC_monotone[OF S, unfolded rstep_CC_CS[symmetric]] rstep_mono[OF U]
       CC_monotone[OF R', unfolded rstep_CC_CS[symmetric]] 
  have R': "rstep (?U \<union> ?S) = rstep (set R')" unfolding rstep_union by auto
  from finite have finite: "finite_dpp (?UP, ?U \<union> ?S)" using R' UP unfolding finite_dpp_def
    by (simp add: minimal_cond_def, blast)
  show "finite_dpp (set P, set R)"
    by (rule uncurrying_sound_dp[OF hvf eta _ only check_inj[OF inj] finite], auto)
qed

abbreviation uncurry_below_rules where "uncurry_below_rules sm \<equiv> map_rules_wa (get_default_sym sm)"

definition uncurry_top_rules :: "'f \<Rightarrow> nat \<Rightarrow> 'f sig_map \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rules"
  where "uncurry_top_rules a n sm \<equiv> map (\<lambda> (l,r). (uncurry_top a n sm l, uncurry_top a n sm r))"

lemma uncurry_top_rules: "set (uncurry_top_rules a n sm R) = uncurry_top_trs a n sm (set R)"
  unfolding uncurry_top_rules_def uncurry_top_trs_def by auto

definition uncurry_of_top_sig_list :: "'f \<Rightarrow> nat \<Rightarrow> 'f sig_map_list \<Rightarrow> 'f sig_map \<Rightarrow> ('f,string)rules"
  where "uncurry_of_top_sig_list a m sml sm \<equiv> concat (map (\<lambda> ((f,n),_). (
                  let g = get_symbol sm f n in
                    map (\<lambda> i. (Fun a (generate_f_xs (g i) (n+i)  # map (\<lambda>i. Var (generate_var i)) [n + i..<n + i + (m - 1)]), generate_f_xs (g (Suc i)) (n + i + (m - 1)))) [0 ..< aarity sm f n]
           )) sml)"

lemma uncurry_of_top_sig_list: "set (uncurry_of_top_sig_list a m sml (sig_list_to_sig_map a sml fmap)) = uncurry_of_top_sig a m (sig_list_to_sig_map a sml fmap)" (is "?l = ?r")
proof -
  let ?sm = "sig_list_to_sig_map a sml fmap"
  let ?g = "get_symbol ?sm"
  {
    fix l r
    assume "(l,r) \<in> ?l"
    then obtain f n list where mem: "((f,n),list) \<in> set sml" 
      and mem2: "(l,r) \<in>  set (map (\<lambda>i. (Fun a
                              (generate_f_xs (?g f n i) (n+i)  # map (\<lambda>i. Var (generate_var i)) [n + i..<n + i + (m - 1)]), 
                              generate_f_xs (?g f n (Suc i)) (n + i + (m - 1)))) [0 ..< aarity ?sm f n])"
      unfolding uncurry_of_top_sig_list_def Let_def by auto
    from mem2
    obtain i where i: "i < aarity ?sm f n" and l: "l = Fun a (generate_f_xs (?g f n i) (n + i) # map (\<lambda> i. Var (generate_var i)) [n+ i..< n + i + (m - 1)])" 
      and r: "r = generate_f_xs  (?g f n (Suc i)) (n + i + (m - 1))" by auto
    have "(l,r) \<in> ?r" unfolding l r uncurry_of_top_sig_def using i by auto
  } note l_imp_r = this
  {
    fix l r
    assume "(l,r) \<in> ?r"
    then obtain f n i where i: "i < aarity ?sm f n" and l: "l = Fun a (generate_f_xs (?g f n i) (n + i) # map (\<lambda> i. Var (generate_var i)) [n+i..< n + i + (m - 1)])" 
      and r: "r = generate_f_xs  (?g f n (Suc i)) (n + i + (m - 1))" unfolding uncurry_of_top_sig_def by auto 
    from i[unfolded aarity_def sig_list_to_sig_map_def] obtain list where lookup: "lookup (f,n) sml = Some list" 
      by (cases "lookup (f,n) sml", auto simp: Let_def)
    from lookup_sound[OF lookup] have mem: "((f,n),list) \<in> set sml" by auto
    have "(l,r) \<in> ?l" unfolding uncurry_of_top_sig_list_def
      by (simp, rule bexI[OF _ mem], simp add: Let_def, unfold l r, insert i, auto)
  }
  with l_imp_r show ?thesis by auto
qed  

fun uncurry_top_proc :: "('f,string)uncurry_info \<Rightarrow> nat \<Rightarrow> ('f sig_map_list \<Rightarrow> 'f \<Rightarrow> nat \<Rightarrow> 'f) \<Rightarrow> ('f \<Rightarrow> nat \<Rightarrow> 'f sig_map_list \<Rightarrow> shows check) \<Rightarrow> ('f,string)rules \<Rightarrow> ('f,string)rules \<Rightarrow> ('f,string)rules \<Rightarrow> ('f :: show,string)rules \<Rightarrow> ('f,string)rule_map \<Rightarrow> shows check"
where "uncurry_top_proc (a,sml,_,E) n fmap check_inj P R P' R' (Some rm) = do {
           check (n \<noteq> 0) (shows ''the arity of the uncurried symbol must be at least 1'');
           let sm = sig_list_to_sig_map a sml fmap;
           let P_eta = E @ P;
           let U = uncurry_below_rules sm R;
           check_inj a n sml;           
           check_allm (\<lambda> (l,r). do {
                 check (hvf_top a n l) (shows ''head variable in lhs '' +@+ shows_term l +@+ shows '' not allowed'');
                 check_no_var r;
                 check_no_defined_root rm r
           }) P_eta;
           check (rm (a,n) = []) (shows ''application symbol '' +@+ shows a +@+ shows '' must not be defined in R'');
           (if list_ex (\<lambda> (l,r). the (root r) = a \<and> num_args r = n \<and> capRM2 rm (hd (args r)) = HHole) P then
               do {
                  check_CS_subseteq (uncurry_of_top_sig_list a n sml sm) P'
                       <+? (\<lambda> lr. shows ''uncurry rule '' +@+ shows_rule lr +@+ shows '' is missing in P'');
                  eta_closed_top_rules a n sm R P_eta
               } else OK);
           check_subseteq (uncurry_top_rules a n sm P_eta) P'
              <+? (\<lambda> lr. shows ''uncurried pair '' +@+ shows_rule lr +@+ shows '' is missing'');
           check_subseteq U R'
              <+? (\<lambda> lr. shows ''uncurried rule '' +@+ shows_rule lr +@+ shows '' is missing'');
           check_subseteq R' U
              <+? (\<lambda> lr. shows ''additional rule '' +@+ shows_rule lr +@+ shows '' in R is not allowed'')
   }" |
   "uncurry_top_proc _ _ _ _ _ _ _ _ None = OK"

lemma uncurry_top_proc: assumes check_inj: "n \<noteq> 0 \<Longrightarrow> isOK(check_inj a n sml) \<Longrightarrow> inj_sig_map a n (sig_list_to_sig_map a sml fmap)"
  shows "sound_proc_impl (\<lambda> P R rm P' R'. isOK(uncurry_top_proc (a,sml,U,E) n fmap check_inj P R P' R' rm))"
  unfolding sound_proc_impl_def
proof (intro allI impI)
  fix P R P' R'
  assume ok: "isOK(uncurry_top_proc (a,sml,U,E) n fmap check_inj P R P' R' (computeRuleMap R))" and finite: "finite_dpp (set P', set R')"
  let ?sm = "sig_list_to_sig_map a sml fmap"
  let ?P_eta = "E @ P"
  obtain P_eta where P_eta: "?P_eta = P_eta" by simp
  let ?UP = "uncurry_top_trs a n ?sm (set ?P_eta)"
  let ?U = "uncurry_below_trs ?sm (set R)"
  let ?S = "uncurry_of_top_sig a n ?sm"
  show "finite_dpp (set P, set R)"
  proof (cases "computeRuleMap R")
    case (Some rm)
    from ok[unfolded Some uncurry_top_proc.simps Let_def]
    have 
      hvf: "\<And> l r. (l,r) \<in> set ?P_eta \<Longrightarrow> hvf_top a n l \<and> \<not> is_var r"
      and ndef: "\<And> l r. (l,r) \<in> set ?P_eta \<Longrightarrow> \<not> defined (set R) (the (root r), num_args r)" 
      and eta: "isOK (if list_ex (\<lambda> (l,r). the (root r) = a \<and> num_args r = n \<and> cap (set R) (hd (args r)) = HHole) P then
               do {
                  check_CS_subseteq (uncurry_of_top_sig_list a n sml ?sm) P'
                       <+? (\<lambda> lr. shows ''uncurry rule '' +@+ shows_rule lr +@+ shows '' is missing in P'');
                  eta_closed_top_rules a n ?sm R P_eta
               } else OK)"
      and R': "set R' = ?U"
      and UP: "?UP \<subseteq> set P'"
      and a: "rm (a,n) = []"
      and inj: "isOK(check_inj a n sml)"
      and n: "n \<noteq> 0"
      by (auto simp: P_eta uncurry_top_rules map_rules_wa check_no_defined_root_sound[OF Some] capRM2Sound[OF Some])
    from n obtain m where n: "n = Suc m" by (cases n, auto)
    let ?d = "\<exists> (l,r) \<in> set P. the (root r) = a \<and> num_args r = n \<and> cap (set R) (hd (args r)) = HHole"
    let ?c = "\<forall> (l,r) \<in> set P. the (root r) \<noteq> a \<or> num_args r \<noteq> n \<or> cap (set R) (hd (args r)) \<noteq> HHole"
    have eta: "?c \<or> eta_closed_top a n ?sm (set R) (set ?P_eta) \<and> CS ?S \<subseteq> CS (set P')"
    proof (cases ?d)
      case False
      thus ?thesis by blast
    next
      case True
      hence "?d = True" by simp
      note eta = eta[unfolded this list_ex_iff, simplified, unfolded uncurry_of_top_sig_list] 
      from eta[unfolded eta_closed_top_rules] P_eta show ?thesis by simp
    qed
    from computeRuleMap_defs[OF Some] a have a: "\<not> defined (set R) (a,n)" unfolding defs_def by auto
    from ndef have ndef: "\<And> s f ts. (s, Fun f ts) \<in> set ?P_eta \<Longrightarrow> \<not> defined (set R) (f, length ts)" by force
    from CS_monotone[OF UP] have P': "CS (?UP \<union> set P') \<subseteq> CS (set P')" unfolding cs_union_distr by auto 
    let ?min = "{(s,t) | s t. SN_elt (rstep ?U) t}"
    have finite: "finite_dpp (?UP \<union> set P', ?U)" unfolding DP_Framework.finite_dpp_iff_rel_SN
      by (rule rel_SN_mono[OF _ subset_refl finite[unfolded R' DP_Framework.finite_dpp_iff_rel_SN]], insert P', auto simp: P_eta)
    show ?thesis
      by (rule uncurrying_top_sound_dp[where n = m, unfolded n[symmetric], OF hvf _ eta a ndef check_inj[OF _ inj] finite], insert n, auto)
  next
    case None
    from computeRuleMap_None_not_SN[OF None]
    show ?thesis unfolding finite_dpp_def by (auto simp: minimal_cond_def)
  qed
qed


subsection {* generating injective symbol mappings from lists *}

fun label_depth :: "('f,'l)lab \<Rightarrow> nat"
where "label_depth (UnLab _) = 0"
   |  "label_depth (Lab f _) = Suc (label_depth f)"
   |  "label_depth (FunLab f _) = Suc (label_depth f)"
   |  "label_depth (Sharp f) = Suc (label_depth f)"

fun gen_label :: "('f,'l)lab \<Rightarrow> nat \<Rightarrow> ('f,'l)lab"
where "gen_label f 0 = f"
   |  "gen_label f (Suc n) = FunLab (gen_label f n) []"

lemma label_depth_gen_label: "label_depth (gen_label f n) = n + label_depth f"
  by (induct n, auto)


lemma gen_label_inj_i: "inj (gen_label f)"
  unfolding inj_on_def
proof (intro ballI, simp, rule)
  fix n m
  assume "gen_label f n = gen_label f m"
  hence "label_depth (gen_label f n) = label_depth (gen_label f m)" 
    by simp  
  thus "n = m" unfolding label_depth_gen_label by auto
qed

lemma gen_label_inj_f: "inj (\<lambda> f. gen_label f i)"
  unfolding inj_on_def
proof (intro ballI, simp, rule)
  fix f g
  assume "gen_label f i = gen_label g i"
  thus "f = g"
    by (induct i arbitrary: f g, auto)
qed

definition fmap :: "('f,'l)lab \<Rightarrow> nat \<Rightarrow> ('f,'l)lab sig_map_list \<Rightarrow> ('f,'l)lab \<Rightarrow> nat \<Rightarrow> ('f,'l)lab"
where "fmap a nn sml \<equiv> let m = Suc (max_list (map label_depth (a # concat (map snd sml))))
                   in (\<lambda> f n. if (f,n) = (a,nn) then a else gen_label f m)"

definition check_inj :: "('f :: show,'l :: show)lab \<Rightarrow> nat \<Rightarrow> ('f,'l)lab sig_map_list \<Rightarrow> shows check"
  where "check_inj a nn sml \<equiv> let symbols = map (\<lambda>((f,n),fs). map (\<lambda> (g,i). (g,n+i * (nn - 1))) (zip fs [0..<length fs])) sml;
                               fsymbols = concat symbols
                           in do {
                             check_partition symbols 
                                  <+? (\<lambda> f. (shows ''symbol '' +@+ shows f +@+ shows '' occurs twice)''));
                             check ((a,nn) \<notin> set fsymbols) (shows ''application symbol'' +@+ shows a +@+ shows '' must not occur as new symbol'');
                             check ((a,nn) \<notin> set (map fst sml)) (shows ''application symbol'' +@+ shows a +@+ shows '' must not be uncurried'');
                             (if nn \<le> 1 then check_allm (check_pairwise (\<lambda> gn1 gn2. check (gn1 \<noteq> gn2) (shows ''symbol '' +@+ shows gn1 +@+ shows '' occurs twice''))) symbols else OK)
                           }"

lemma check_inj: assumes ok: "isOK(check_inj a nn sml)"
  shows "inj_sig_map a nn (sig_list_to_sig_map a sml (fmap a nn))"
proof -
  note ok = ok[unfolded check_inj_def Let_def, simplified]
  {
    fix v
    assume "lookup (a,nn) sml = Some v"
    from lookup_sound[OF this] have "(a,nn) \<in> set (map fst sml)" by force
    with ok have False by auto
  } note Some = this
  have a: "sig_list_to_sig_map a sml (fmap a nn) a nn = [a]"
    unfolding sig_list_to_sig_map_def Let_def
    by (cases "lookup (a,nn) sml", simp add: fmap_def, insert Some, auto)
  show ?thesis 
    unfolding inj_sig_map_def get_symbol_def aarity_def
  proof (rule conjI[OF _ a])
    let ?s = "sig_list_to_sig_map a sml (fmap a nn)"
    let ?g = "\<lambda>(f,n,i). (?s f n ! i, n + i * (nn - 1))"
    let ?U = "{(f,n,i). i \<le> length (?s f n) - 1}"
    let ?L = "a # concat (map snd sml)"
    let ?M = "set ?L"
    let ?m = "Suc (max_list (map label_depth ?L))"
    let ?S = "{(fs ! i,n + i * (nn - 1)) | f n i fs. ((f,n),fs) \<in> set sml \<and> i < length fs}"
    {
      fix f n i
      have "?s f n \<noteq> []"  unfolding sig_list_to_sig_map_def
        by (cases "lookup (f,n) sml", auto simp: Let_def)
      hence "(i \<le> length (?s f n) - 1) = (i < length (?s f n))"
        by (cases "?s f n", auto)
    } note len = this
    show "inj_on ?g ?U"
      unfolding inj_on_def
    proof (intro ballI, unfold len, clarify)
      fix f g n m i j
      assume i: "i < length (?s f n)"
      and    j: "j < length (?s g m)"
      and    idg: "?s f n ! i = ?s g m ! j"
      and    ids: "n + i * (nn - 1) = m + j * (nn - 1)" 
      from idg have lid: "label_depth (?s f n ! i) = label_depth (?s g m ! j)" by auto
      {
        fix f
        assume "f \<in> ?M"
        hence "label_depth f \<in> set (map label_depth ?L)" by auto 
        from max_list[OF this] have "label_depth f < ?m" by simp
      } note small = this
      {
        fix f
        have "label_depth (gen_label f ?m) \<ge> ?m" unfolding label_depth_gen_label by auto
      } note large = this
      {
        fix f n i
        assume napp: "(f,n) \<noteq> (a,nn)" and fmap: "?s f n = [fmap a nn sml f n]" and i: "i < length (?s f n)"
        hence "?s f n ! i = gen_label f ?m \<and> i = 0" and "label_depth (?s f n ! i) \<ge> ?m" using large unfolding fmap_def Let_def by auto
      } note generated = this
      {
        fix f n i
        assume app: "(f,n) = (a,nn)" and i: "i < length (?s f n)"
        hence "?s f n ! i = a \<and> i = 0" and "label_depth (?s f n ! i) < ?m" using a small[of a] by auto
      } note app = this
      {
        fix f n i
        assume napp: "(f,n) \<noteq> (a,nn)" and fmap: "?s f n \<noteq> [fmap a nn sml f n]" and i: "i < length (?s f n)"
        note fmap =  fmap[unfolded sig_list_to_sig_map_def Let_def]
        from fmap obtain fs where Some: "lookup (f,n) sml = Some fs" by (cases "lookup (f,n) sml", auto)
        note mem = lookup_sound[OF Some]
        from fmap Some have fs: "fs \<noteq> []" by (cases fs, auto)
        from Some fs have sfn: "?s f n = fs" unfolding sig_list_to_sig_map_def Let_def by auto
        from i have mem2: "?s f n ! i \<in> set fs" unfolding sfn by auto
        from mem mem2 i have inS: "(?s f n ! i,n+i * (nn - 1)) \<in> ?S" unfolding sfn by auto
        have ld: "label_depth (?s f n ! i) < ?m"
          by (rule small, insert mem mem2, force)
        from fmap mem2 Some mem i sfn have "\<exists> fs. ?s f n = fs \<and> i < length fs \<and> ((f,n),fs) \<in> set sml \<and> lookup (f,n) sml = Some fs" by force
        note inS and ld and this
      } note sml = this        
      have a2: "(a,nn) \<notin> ?S"
      proof
        assume "(a,nn) \<in> ?S"
        then obtain f n i fs mm where a: "a = fs ! i" and nn: "nn = n + i * (mm - 1)" and mm: "mm = nn" and mem: "((f,n),fs) \<in> set sml" and i: "i < length fs" by auto        
        have id: "(fs ! i, nn) = (fs ! i, n + ([0..<length fs] ! i) * (mm - 1))" using i unfolding nn by auto 
        from ok[THEN conjunct2, THEN conjunct1, THEN bspec[OF _ mem], simplified] show False unfolding set_zip a
          by (simp add: id, insert mm, insert i, best)
      qed
      show "f = g \<and> (n,i) = (m,j)"
      proof (cases "label_depth (?s f n ! i) < ?m")
        case False
        from app(2)[OF _ i] False have nappf: "(f,n) \<noteq> (a,nn)" by blast
        from sml(2)[OF nappf _ i] False have genf: "?s f n = [fmap a nn sml f n]" by blast
        from generated(1)[OF nappf genf i] have sfn: "?s f n ! i = gen_label f ?m" and i: "i = 0" by auto
        note False = False[unfolded lid]
        from app(2)[OF _ j] False have nappg: "(g,m) \<noteq> (a,nn)" by blast
        from sml(2)[OF nappg _ j] False have geng: "?s g m = [fmap a nn sml g m]" by blast
        from generated(1)[OF nappg geng j] have sgm: "?s g m ! j = gen_label g ?m" and j: "j = 0" by auto
        from idg[unfolded sfn sgm] gen_label_inj_f[unfolded inj_on_def] have "f = g" by auto
        thus ?thesis using ids i j by auto
      next
        case True
        from generated(2)[OF _ _ i] True have sfn: "(f,n) = (a,nn) \<or> ((f,n) \<noteq> (a,nn) \<and> ?s f n \<noteq> [fmap a nn sml f n])" by force
        from generated(2)[OF _ _ j] True[unfolded lid] have sgm: "(g,m) = (a,nn) \<or> ((g,m) \<noteq> (a,nn) \<and> ?s g m \<noteq> [fmap a nn sml g m])" by force
        from sfn
        show ?thesis
        proof
          assume fn: "(f,n) = (a,nn)"
          from app(1)[OF this i] a2 fn have nmem: "(?s f n ! i,n + i * (nn - 1)) \<notin> ?S" and i: "i = 0" by auto
          from sgm show ?thesis
          proof
            assume "(g,m) = (a,nn)"
            with fn app(1)[OF this j] i ids show ?thesis by auto
          next
            assume "((g,m) \<noteq> (a,nn) \<and> ?s g m \<noteq> [fmap a nn sml g m])"
            from sml(1)[OF _ _ j] this nmem[unfolded idg ids] show ?thesis 
              by auto
          qed
        next
          assume "(f,n) \<noteq> (a,nn) \<and> ?s f n \<noteq> [fmap a nn sml f n]"
          hence nappf: "(f,n) \<noteq> (a,nn)" and smlf: "?s f n \<noteq> [fmap a nn sml f n]" by auto
          from sml(1)[OF nappf smlf i] have finS: "(?s f n ! i, n+ i * (nn - 1)) \<in> ?S" .
          from sgm show ?thesis
          proof
            assume "(g,m) = (a,nn)"
            with finS[unfolded idg ids] app(1)[OF this j] a2 show ?thesis by auto
          next
            assume "((g,m) \<noteq> (a,nn) \<and> ?s g m \<noteq> [fmap a nn sml g m])"
            hence nappg: "(g,m) \<noteq> (a,nn)" and smlg: "?s g m \<noteq> [fmap a nn sml g m]" by auto
            from sml(1)[OF nappg smlg j] have ginS: "(?s g m ! j, m + j * (nn - 1 )) \<in> ?S" .
            from sml(3)[OF nappf smlf i]
            obtain fs where sfn: "?s f n = fs" and i: "i < length fs" and fnfs: "((f,n),fs) \<in> set sml" by auto
            from fnfs obtain I where fnfsI: "sml ! I = ((f,n),fs)" and I: "I < length sml" unfolding set_conv_nth by auto
            show ?thesis
            proof (cases "(f,n) = (g,m)")
              case True note oTrue = this
              show ?thesis
              proof (cases "nn \<le> Suc 0")
                case False
                with True ids show ?thesis by auto
              next
                case True
                hence nn1: "(nn \<le> Suc 0) = True" and nn2: "nn - 1 = 0" by auto                
                from ok[unfolded nn1 nn2, THEN conjunct2, THEN conjunct2, THEN conjunct2, simplified, THEN bspec[OF _ fnfs], simplified]
                have dist: "\<And> j. j < length fs \<Longrightarrow> \<forall> i < j. fs ! i \<noteq> fs ! j" by simp
                from sfn sgm oTrue idg i j have idfs: "fs ! i = fs ! j" and j: "j < length fs" by auto
                have "i = j"
                proof (cases "i < j")
                  case True
                  from dist[OF j] idfs True show ?thesis by simp
                next
                  case False note oFalse = this
                  show ?thesis
                  proof (cases "j < i")
                    case True
                    from dist[OF i] idfs True show ?thesis by auto
                  next
                    case False
                    with oFalse show ?thesis by simp 
                  qed
                qed
                with oTrue show ?thesis by simp
              qed                
            next
              case False
              from sml(3)[OF nappg smlg j]
              obtain gs where sgm: "?s g m = gs" and j: "j < length gs" and gmgs: "((g,m),gs) \<in> set sml" by auto
              from gmgs obtain J where gmgsJ: "sml ! J = ((g,m),gs)" and J: "J < length sml" unfolding set_conv_nth by auto
              show ?thesis
              proof (cases "I = J")
                case False
                obtain nnn where nnn: "nnn = nn - Suc 0" by simp
                let ?f = "\<lambda>x. set ((\<lambda>((f, n), fs). map (\<lambda>(g, i). (g, n + i * nnn)) (zip fs [0..<length fs])) x)"
                let ?innA = "\<lambda> n fs. set (map (\<lambda>(g, i). (g, n + i * nnn)) (zip fs [0..<length fs]))"
                let ?innB = "\<lambda> n fs. {(fs ! i,n+ i * nnn) | i. i < length fs}"
                let ?g = "\<lambda>((f,n),fs). ?innB n fs"
                have inn: "?innA = ?innB"
                proof (rule ext, rule ext)
                  fix n fs
                  show "?innA n fs = ?innB n fs"
                  proof (simp add: set_zip, rule, (auto)[1])
                    from nth_upt[of 0 _ "length fs"]
                    have upt: "\<And> i. i < length fs \<Longrightarrow> [0..<length fs] ! i = i" by auto
                    show "{(fs ! i, n + i * nnn) | i. i < length fs} \<subseteq> (\<lambda>(g, i). (g, n + i * nnn)) ` {(fs ! i, [0..<length fs] ! i) |i. i < length fs}"
                      by (rule, clarify, simp, insert upt, force)
                  qed
                qed 
                have fg: "?f = ?g" unfolding inn[symmetric]  by (intro ext, clarify)
                from fnfs fnfsI I i have mem1: "(?s f n ! i, n + i * nnn) \<in> map ?g sml ! I" unfolding sfn nnn by auto
                from gmgs gmgsJ J j have mem2: "(?s g m ! j, m + j * nnn) \<in> map ?g sml ! J" unfolding sgm nnn by auto
                from ok[THEN conjunct1, unfolded o_def is_partition_alt is_partition_alt_def, THEN spec[of _ I], THEN spec[of _ J], unfolded nnn[symmetric] fg] I J False
                have disjoint: "map ?g sml ! I \<inter> map ?g sml ! J = {}" by auto
                from disjoint mem1 mem2 have False using idg ids nnn by auto
                thus ?thesis ..
              next
                case True
                with False fnfsI gmgsJ show ?thesis by simp 
              qed
            qed
          qed
        qed
      qed
    qed
  qed
qed

fun uncurry_proc_both where 
  "uncurry_proc_both None (a,sml,U,E) = uncurry_proc (a,sml,U,E) (fmap a 2) check_inj"
| "uncurry_proc_both (Some n)  (a,sml,U,E) = uncurry_top_proc (a,sml,U,E) n (fmap a n) check_inj"



lemma uncurry_proc_both: 
  shows "sound_proc_impl (\<lambda> P R rm P' R'. isOK(uncurry_proc_both mode info P R P' R' rm))"
proof -
  obtain a sml U E where info: "info = (a,sml,U,E)" by (cases info, auto)
  show ?thesis 
  proof (cases mode)
    case None
    show ?thesis unfolding info None uncurry_proc_both.simps
      by (rule uncurry_proc, rule check_inj)
  next
    case (Some n)
    show ?thesis unfolding info Some uncurry_proc_both.simps
      by (rule uncurry_top_proc, rule check_inj)
  qed
qed

hide_const(open) fmap
hide_const(open) gen_label
hide_const(open) label_depth

end
